/*****************************************************************************
* PROJECT:	Sangoma Frame Relay Network Device Driver
* FILE:		flip.c		Main Module's C Source Code
* TARGET OS:	Linux
* AUTHOR:	Gene Kozin
* ----------------------------------------------------------------------------
* Copyright (c) 1995-1996 Sangoma Technologies Inc.  All Rights Reserved.
* ============================================================================
* Rev.#	Date		Who	Details
* -----	-----------	---	----------------------------------------------
*  1.40	Dec.18,1996	GK	o support for configurable link management
*  1.30	Nov.11,1996	GK	o Updated for SDLA driver v4.00
*  1.20	Jul.15,1996	GK	o Changes for Linux 2.0 compatibility.
*  1.10	Jun.18,1996	GK	o Updated for SDLA Driver 3.0 compatibility.
*				o Use adapter mailbox access arbitration.
*  1.00	Mar.21,1996	GK	o Beta release.
*  0.21	Feb.09.1996	GK	o Changes to allow FLIP work with Linux 1.3.X
*				  kernel. (for 1.2.X #define _1_2_X_).
*  0.20	Dec.14.1995	GK	o All firmware implementation-dependent stuff
*				  has been moved to separate modules.
*  0.11	Nov.29.1995	GK	o flip_ifsend() always returns 0.
*  0.10	Nov.27.1995	GK	o Driver is implemented as a network device.
*  0.03	Nov.09,1995	GK	o Firmware ID changed.
*  0.03	Nov.03,1995	GK	o Increment/decrement mod_use_count properly
*				  to inhibit removing of the module while
*				  there are still opened interfaces.
*				o Allow to open an interface even if DLCI is
*				  not active.  Just print a warning.
*				o Added retry limit to all FR calls to
*				  eliminate a danger of looping endlessly.
*  0.02	Oct.27,1995	GK	o Added minimal support for Access Node mode.
*  0.01	Oct.12,1995	GK	o Alpha version.
*****************************************************************************/

/*****************************************************************************
* Notes:
*  1) This module is not portable.  Linux only!
*  2) Compilation options:
*	GNU C (Linux):	
*	MSC (SCO Unix):	-K -M3 -Zp4
*****************************************************************************/

#include <linux/config.h>	/* OS configuration options */
#include <linux/stddef.h>	/* offsetof(), etc. */
#include <linux/errno.h>	/* return codes */
#include <linux/kernel.h>
#include <linux/malloc.h>	/* kmalloc(), kfree() */
#include <linux/interrupt.h>
#include <linux/version.h>	/* for UTS_RELEASE */
#include <linux/module.h>	/* support for loadable modules */
#include <linux/inet.h>		/* in_aton(), in_ntoa() prototypes */
#include <linux/netdevice.h>	/* support for network drivers */
#include <linux/if.h>		/* network interface stuff */
#include <linux/if_ether.h>	/* for ethertypes */
#include <linux/skbuff.h>	/* for struct sk_buff */
#include <asm/string.h>		/* inline mem*, str* functions */
#include <asm/byteorder.h>	/* htons(), etc. */

#include "sdladrv.h"		/* SDLA driver kernel interface definitions */
#include "flipio.h"		/* this driver API definitions */
#include "flip.h"		/* common defines */

/****** Defines *************************************************************/

#define	FLIP_RELEASE	30		/* driver minor version number */
#define	FLIP_MAX_PORTS	SDLA_MAX_UNITS	/* max number of physical ports */
#define	IF_NAME_LEN	16		/* max length of interface name */

#if (_OS_VER_ == 102)
#define	DRV_FULLNAME	"WANPIPE(tm) Frame Relay Router Module for Linux 1.2.X"
#elif (_OS_VER_ == 103)
#define	DRV_FULLNAME	"WANPIPE(tm) Frame Relay Router Module for Linux 1.3.X"
#elif (_OS_VER_ == 200)
#define	DRV_FULLNAME	"WANPIPE(tm) Frame Relay Router Module for Linux 2.0.X"
#else
#error _OS_VER_ is not defined or invalid!
#endif

/****** Data Types **********************************************************/

typedef struct	enet_statistics	IfStats_t;	/* just for shortness */

typedef struct	LookUp			/* Look-Up Table Entry */
{
  int		token;			/* look-up code */
  void		*ptr;			/* pointer */
} LookUp_t;

typedef struct	IfData			/* Interface Data Space (IDS) */
{
  struct device	dev;			/* Linux network device. Must be 1st */
  IfStats_t	stats;			/* statistics */
  struct IfData	*next;			/* link forward */
  struct IfData	*prev;			/* link backward */
  char		name[IF_NAME_LEN];	/* interface name */
  int		ifno;			/* interface number */
  int		port;			/* port number */
  int		dlci;			/* DLCI number for this interface */
  int		flags;			/* status flags */
} IfData_t;

typedef int	Handler	(flipUsr_t *usr);

/****** Function Prototypes *************************************************/

/*
 * Module Wrappers.
 * These are called by insmod & rmmod and must be declared as public.
 */
int	init_module	(void);
void	cleanup_module	(void);

/*
 * SDLA Module Entry Points. (Called by the SDLA Driver).
 */
static int mod_bind	(sdlaData_t *ads);
static int mod_unbind	(sdlaData_t *ads);
static int mod_intr	(sdlaData_t *ads);

/*
 * Master Device Driver Entry Points.
 */
static int flip_init	(struct device *dev);
static int flip_open	(struct device *dev);
static int flip_close	(struct device *dev);
static int flip_ioctl	(struct device *dev, struct ifreq *ifr, int cmd);

/*
 * Master Device IOCTL Handlers.
 */
static int flipio_dummy		(void *ptr);
static int flipio_config	(flipPortCfg_t *portcfg);
static int flipio_reset		(int *port_p);
static int flipio_status	(int *port_p);
static int flipio_add		(flipIfCfg_t *ifcfg);
static int flipio_delete	(int *ifno_p);
static int flipio_ifstatus	(int *ifno_p);

/*
 * Network Interface Device Driver Entry Points.
 */
static int flip_ifinit	(struct device *dev);
static int flip_ifopen	(struct device *dev);
static int flip_ifclose	(struct device *dev);
static int flip_ifsend	(struct sk_buff *skb, struct device *dev);
static IfStats_t *flip_ifstats (struct device *dev);
static int flip_header	(
#if (_OS_VER_ < 200)
	unsigned char *buff, struct device *dev, unsigned short type,
	void *daddr, void *saddr, unsigned len, struct sk_buff *skb
#else
	struct sk_buff *skb, struct device *dev, unsigned short type,
	void *daddr, void *saddr, unsigned len
#endif
);
static int flip_rebuild_h (void *buff, struct device *dev,
		 	 unsigned long dst, struct sk_buff *skb);
#if (_OS_VER_ < 103)
static unsigned short flip_type_trans (struct sk_buff *skb, struct device *dev);
#endif

/*
 * Support Functions.
 */
static int	port_down	(Port_t *pds_p);
static int	if_delete	(IfData_t *ids_p);
static IfData_t *if_find	(int port, int dlci);
static void	*lookup		(LookUp_t *table, int token);
static void	dlci_to_Q922	(unsigned int dlci, char *str);
static unsigned short decapsulate (Encaps_t *frm_p, unsigned *hdrlen_p);

/****** Global Data **********************************************************
* Note:	All data must be explicitly initialized!!!
*/

#if (_OS_VER_ < 200)
char		kernel_version[]= UTS_RELEASE;	/* must be public */
#endif

static char	drv_name[]	= DRV_NAME;	/* short module name */
static char	drv_fullname[]	= DRV_FULLNAME;
static char	mod_copyright[]	= "(c) 1995-1996 Sangoma Technologies Inc.";

static IfData_t	*iflist	= NULL;		/* list of configured interfaces */
static Port_t	*ports[FLIP_MAX_PORTS];	/* adapter data space pointers */
struct device	mdev;			/* FLIP master control device */

static sdlaMod_t modreg =		/* Module registration info */
{
  DRV_NAME,		/* .name */
  NULL,			/* .next */
  &mod_bind,		/* .bind */
};

static LookUp_t	ioctl_table[] =		/* IOCTL Handler Look-up Table */
{
  { FLIPIO_CONFIG,	flipio_config	},
  { FLIPIO_RESET,	flipio_reset	},
  { FLIPIO_STATUS,	flipio_status	},
  { FLIPIO_ADD,		flipio_add	},
  { FLIPIO_DELETE,	flipio_delete	},
  { FLIPIO_IFSTATUS,	flipio_ifstatus	},
  { 0,			flipio_dummy	},	/* End */
};

static LookUp_t	frfunc_table[] =	/* FR Function Look-up Table */
{
  { SFID_FR502,	&fr502_func	},
  { SFID_FR508,	&fr508_func	},
  { 0,		NULL		},	/* End */
};

/*
 * SNAP Organizationally Unique Identifiers.
 */
static unsigned char OUI_Ether[] = {0x00, 0x00, 0x00};	/* routed packets */
static unsigned char OUI_802_2[] = {0x00, 0x80, 0xC2};	/* bridged packets */

/****** Module Wrappers *****************************************************/

/*============================================================================
 * init_module()	Module Load Entry Point
 * ---------------------------------------------------------------------------
 *  o Announce ourselves.
 *  o Initialize static data.
 *  o Register ourselves with SDLA driver.
 * Return 0 if O.k. to install.
 */
int init_module	(void)
{
  int	err = 0;

  printk("%s v%u.%u %s\n",
         drv_fullname, FLIP_VERSION, FLIP_RELEASE, mod_copyright)
  ;

  /*
   * Initialize global data.
   */
  memclr(ports, sizeof(ports));
  memclr(&mdev, sizeof(mdev));

  /*
   * Register with SDLA driver.
   */
  if (sdla_register(SDLA_SIGNATURE, SDLA_VERSION, &modreg))
  {
    if (SDLA_DEBUG)
      printk("%s: failed to register with SDLA driver!\n", drv_name)
    ;
    err = -1;
  }
  else
  {
    /*
     * Register master control device.
     */
    mdev.name = drv_name;
    mdev.init = flip_init;
    err = register_netdev(&mdev);
    if (err && SDLA_DEBUG)
      printk("%s: failed to register network interface! Error %d\n",
             drv_name, err)
    ;
  }
  return err;
} /* init_module() */

/*============================================================================
 * cleanup_module()	Module Unload Entry Point
 * ---------------------------------------------------------------------------
 * o Shut down all ports.
 * o De-register master control device.
 * o De-register from SDLA driver.
 */
void cleanup_module (void)
{
  int		i;
  Port_t	*pds_p;

  if (SDLA_DEBUG > 1)
    printk("%s: unloading ...\n", drv_name)
  ;

  /*
   * Shut down and delete all configured ports.
   */
  for (i = 0; i < FLIP_MAX_PORTS; ++i)
  {
    pds_p = ports[i];
    if (pds_p)
    {
      if (pds_p->ifcnt)
        printk ("%s: Warning: Port %d is still active!\n", drv_name, i)
      ;
      port_down(pds_p);
      kfree(pds_p);		/* release port data space */
      ports[i] = NULL;		/* indicates that port is removed */
    }
  }

  /*
   * De-register master control device.
   */
  unregister_netdev(&mdev);

  /*
   * De-register from SDLA driver.
   */
  sdla_unregister(drv_name);
} /* cleanup_module() */

/******* SDLA Module Entry Points *******************************************/

/*============================================================================
 * mod_bind()		Bind module.
 * ---------------------------------------------------------------------------
 * o Allocate and initialize port data space.
 * o Check adapter state and firmware compatibility.
 * Return 0 if done or SDLAER_* if error.
 */
static int mod_bind (sdlaData_t *ads)
{
  int		err = 0;
  Port_t	*pds_p;		/* -> port data space */

  if (ads->state != SDLA_RUNNING)
    err = SDLAER_BADSTATE		/* invalid adapter state */
  ;
  else if (ports[ads->cardno] != NULL)
    err = SDLAER_BADSTATE		/* port data space already in use */
  ;
  else if ((pds_p = kmalloc(sizeof(Port_t), GFP_KERNEL)) == NULL)
    err = SDLAER_FAILURE		/* memory allocation failure */
  ;
  else
  {
    char	str[80];

    /*
     * Initialize port data space.
     */
    memclr(pds_p, sizeof(Port_t));
    pds_p->ads   = ads;
    pds_p->wbase   = ads->vbase;
    pds_p->fr_func = lookup(frfunc_table, ads->firmware_id);
    if (pds_p->fr_func == NULL)
      err = SDLAER_BADTYPE
    ;
    else if (ads->irq == 0)
    {
      printk("%s: Non interrupt-driven mode is not supported!\n", drv_name);
      err = SDLAER_BADTYPE;		/* non interrupt-driven port */
    }
    else if (pds_p->fr_func->version(pds_p, str))
    {
      printk("%s: Firmware failure!\n", drv_name);
      err = SDLAER_FAILURE;		/* adapter is dead */
    } 

    /*
     * Finish initialization.
     */
    if (err) kfree(pds_p);
    else
    {
      printk("%s: Port %d running Frame Relay v%s.\n",
             drv_name, ads->cardno, str);
      pds_p->flags |= PORT_READY;	/* port is ready now */

      /*
       * Setup module entry points and private data.
       */
      ads->read	  = NULL;
      ads->write  = NULL;
      ads->ioctl  = NULL;
      ads->intr   = &mod_intr;
      ads->halt   = NULL;
      ads->unbind = &mod_unbind;
      ads->priv = pds_p;		/* link ADS to PDS */
      ports[ads->cardno] = pds_p;
      MOD_INC_USE_COUNT;		/* prevent module from unloading */
    }
  }
  return err;
} /* mod_bind() */

/*============================================================================
 * mod_unbind()		Unbind module.
 * ---------------------------------------------------------------------------
 * o Release resources.
 */
static int mod_unbind (sdlaData_t *ads)
{
  int		err	= 0;
  Port_t	*pds_p	= ads->priv;

  if ((pds_p == NULL) || pds_p->ifcnt)
    err = SDLAER_BADSTATE
  ;
  else
  {
    Port_t	*pds_p	= ads->priv;	/* -> port data space */

    port_down(pds_p);			/* shut down port */
    kfree(pds_p);			/* release port data space */
    ports[ads->cardno] = NULL;	/* indicates that port is removed */
    ads->priv = NULL;
    MOD_DEC_USE_COUNT;			/* allow unloading the module */
  }
  return err;
} /* mod_unbind() */

/*============================================================================
 * Interrupt Service Routine.
 * ---------------------------------------------------------------------------
 */
static int mod_intr (sdlaData_t *ads)
{
  Port_t *pds_p	 = ads->priv;

  if (pds_p->flags & PORT_READY)
  {
    pds_p->flags |= PORT_INTERRUPT;
    pds_p->fr_func->interrupt(pds_p);
    pds_p->flags &= ~PORT_INTERRUPT;
  }
  return 0;
} /* mod_intr() */

/******* Master Device Driver Entry Points **********************************/

/*============================================================================
 * Initialize Master Device.
 * ---------------------------------------------------------------------------
 * Called only once, during device registration.  Returning anything but zero
 * will fail the registration.
 */
static int flip_init (struct device *dev)
{
  int	i;

  dev->open	= flip_open;
  dev->stop	= flip_close;
  dev->do_ioctl	= flip_ioctl;
/*
  dev->get_stats= flip_stats;
*/
  dev->family	= AF_UNSPEC;		/* address family */

  /*
   * Initialize socket buffers.
   */
  for (i = 0; i < DEV_NUMBUFFS; ++i) skb_queue_head_init(&dev->buffs[i]);
  return 0;

#undef	ids_p
} /* flip_init() */

/*============================================================================
 * Open Master Device.
 * ---------------------------------------------------------------------------
 * Return 0 if O.k.
 */
static int flip_open (struct device *dev)
{
  int	err	= 0;

  if (dev->start) err = -EBUSY;		/* device is busy */
  else
  {
    dev->tbusy = 0;
    dev->interrupt = 0;
    dev->start = 1;
  }
  return err;
} /* flip_open() */

/*============================================================================
 * Close Master Device.
 * ---------------------------------------------------------------------------
 */
static int flip_close (struct device *dev)
{
  dev->start = 0;
  return 0;
} /* flip_close() */

/*============================================================================
 * Master Device IOCTL.
 * ---------------------------------------------------------------------------
 * Return 0 if O.k.
 */
static int flip_ioctl (struct device *dev, struct ifreq *ifr, int cmd)
{
  int		err	= 0;		/* system return code (errno) */
  int		ioerr	= 0;		/* FLIP I/O result */ 
  int		len;			/* data size to pass back to user */
  flipUsr_t	usr;			/* user call-specific data */
  flipCtl_t	*ioctl_p;

  /*
   * Get a pointer to FLIP control block.
   */
  ioctl_p = (void *)ifr->ifr_ifru.ifru_data;

  /*
   * Verify IOCTL request arguments.
   */
  if (!ioctl_p || verify_area(VERIFY_WRITE, ioctl_p, sizeof(flipCtl_t)))
    err = -EFAULT
  ;
#if (_OS_VER_ < 200)
  else if ((get_user_long((void *)&ioctl_p->signature) != FLIP_SIGNATURE))
#else
  else if ((get_user(&ioctl_p->signature) != FLIP_SIGNATURE))
#endif
    err = -EINVAL
  ;
  else
  {
#if (_OS_VER_ < 200)
    if (get_user_word(&ioctl_p->version) != FLIP_VERSION)
#else
    if (get_user(&ioctl_p->version) != FLIP_VERSION)
#endif
      ioerr = FLIPER_BADVER
    ;
    else
    {
      /*
       * Copy user data to the kernel address space.
       */
      memcpy_fromfs((void *)&usr, (void *)&ioctl_p->u, sizeof(flipUsr_t));

      /*
       * Process user request.
       * IOCTL handler returns size of data to be passed back to user
       * or error code (< 0)
       */
      len = ((Handler *)lookup(ioctl_table, cmd))((void *)&usr);
      if (len > 0) memcpy_tofs((void *)&ioctl_p->u, (void *)&usr, len);
      else ioerr = len;
    }

    /*
     * If error occured, place an error code into control block
     * and return EIO to the system.
     */
    if (ioerr)
    {
#if (_OS_VER_ < 200)
      put_user_word(ioerr, &ioctl_p->result);
#else
      put_user(ioerr, &ioctl_p->result);
#endif
      err = -EIO;
    }
  }
#if	_DEBUG_ >= 2
  if (err)
    printk("%s: ioctl %d error %d (i/o error %d).\n",
           drv_name, cmd - SIOCDEVPRIVATE, err, ioerr)
  ;
#endif
  return err;
} /* flip_ioctl() */

/****** IOCTL Handlers ******************************************************/

/*============================================================================
 * Dummy Handler.
 * ---------------------------------------------------------------------------
 */
static int flipio_dummy (void *ptr)
{
  return FLIPER_BADCMD;
} /* flipio_dummy() */

/*============================================================================
 * FLIPIO_CONFIG Handler. Configure Port.
 * ---------------------------------------------------------------------------
 * o Configure port.
 *
 * On error return:
 *   FLIPER_BADPORT	- invalid port number
 *   FLIPER_BADSTATE	- port in use
 *   FLIPER_NEXIST	- port does not exist (not bound to SDLA)
 *   FLIPER_	- firmware failure
 *   FLIPER_	- port configuration failed
 */
static int flipio_config (flipPortCfg_t *portcfg)
{
  int	err	= 0;

  if (portcfg->port >= FLIP_MAX_PORTS)
    err = FLIPER_BADPORT		/* invalid port number */
  ;
  else
  {
    Port_t *pds_p = ports[portcfg->port];	/* -> port data space */

    if (pds_p == NULL)
      err = FLIPER_NEXIST		/* port does not exist */
    ;
    else if (pds_p->ifcnt)
      err = FLIPER_BADSTATE		/* port is active */
    ;
    else
    {
      /*
       * Configure port.
       */
      memcpy(&pds_p->cfg, portcfg, sizeof(flipPortCfg_t));
      err = pds_p->fr_func->configure(pds_p);
    }
  }
  return err;
} /* flipio_config() */

/*============================================================================
 * FLIPIO_RESET Handler.
 * ---------------------------------------------------------------------------
 * o Shut down the port.
 * o Release kernel memory.
 * On error return:
 *   FLIPER_BADPORT	- invalid port number
 *   FLIPER_NEXIST	- port does not exist
 *   FLIPER_BADSTATE	- there are active interfaces on this port
 */
static int flipio_reset (int *port_p)
{
  int	err	= 0;

  if (*port_p >= FLIP_MAX_PORTS)
    err = FLIPER_BADPORT		/* invalid port number */
  ;
  else
  {
    Port_t *pds_p = ports[*port_p];	/* -> port data space */

    if (pds_p == NULL)
      err = FLIPER_NEXIST		/* port does not exist */
    ;
    else if (!port_down(pds_p))
      err = FLIPER_BADSTATE		/* port is active */
    ;
  }
  return err;
} /* flipio_reset() */

/*============================================================================
 * FLIPIO_STATUS Handler. Read Port Status.
 * ---------------------------------------------------------------------------
 * Return size of the port status block.
 * On error return:
 *   FLIPER_BADPORT	- invalid port number or port does not exist
 */
static int flipio_status (int *port_p)
{
  int	err	= 0;

  if (*port_p >= FLIP_MAX_PORTS)
    err = FLIPER_BADPORT		/* invalid port number */
  ;
  else
  {
    Port_t *pds_p = ports[*port_p];	/* -> port data space */

    if (pds_p == NULL)
      err = FLIPER_NEXIST		/* port does not exist */
    ;
    else
    {
      /*
       * Update port status & statistics.
       */
      pds_p->fr_func->status(pds_p);
      pds_p->fr_func->statistics(pds_p);
      memcpy((void *)port_p, &pds_p->ifcnt, sizeof(flipPortSts_t));
    }
  }
  return err ? err : sizeof(flipPortSts_t);
} /* flipio_status() */

/*============================================================================
 * FLIPIO_ADD Handler. Add Network Interface.
 * ---------------------------------------------------------------------------
 * o Allocate and initialize data space for the interface.
 * o Register network device.
 * o Add new interface to the iterface list.
 * o Increment port reference count.
 * On error return:
 *   FLIPER_BADPORT	- invalid port number or port does not exist
 *   FLIPER_BADDLCI	- DLCI is not available
 *   FLIPER_DLCIBUSY	- DLCI is in use
 *   FLIPER_MALLOC	- cannot allocate kernel memory for interface data
 *   FLIPER_REGDEV	- cannot register network device
 */
static int flipio_add (flipIfCfg_t *ifcfg)
{
  int		err	= 0;
  int		port	= ifcfg->port;
  Port_t	*pds_p;			/* -> port data space */
  IfData_t	*ifds;			/* -> interface data space */

  /*
   * Verify arguments.
   */
  if ((port >= FLIP_MAX_PORTS) || (ports[port] == NULL))
    err = FLIPER_BADPORT		/* invalid port */
  ;
  else
  {
    pds_p = ports[port];
    if (pds_p->ifcnt >= DLCI_NUM)
      err = FLIPER_BADDLCI		/* no DLCI available */
    ;
    else if (if_find(port, ifcfg->dlci))
      err = FLIPER_DLCIBUSY		/* DLCI is in use */
    ;
    else
    {
      /*
       * Allocate and initialize interface data space.
       */
      ifds = kmalloc(sizeof(IfData_t), GFP_KERNEL);
      if (!ifds) err = FLIPER_MALLOC;
      else
      {
        memset(ifds, 0, sizeof(IfData_t));
        sprintf(ifds->name, "%s%u", drv_name, ifcfg->ifno);
        ifds->ifno = ifcfg->ifno;
        ifds->dlci = ifcfg->dlci;
        ifds->port = port;
        ifds->dev.name = ifds->name;
        ifds->dev.init = flip_ifinit;

        /*
         * Register network driver.
         */
        err = register_netdev(&ifds->dev);

        if (err)
        {
#if _DEBUG_ >= 2
          printk("%s: register_netdev() returned %d\n", drv_name, err);
#endif
          err = FLIPER_REGDEV;		/* device registration failure */
          kfree(ifds);
        }
        else
        {
          /*
           * Link IDS to interface list.
           */
          if (iflist)
          {
            ifds->next = iflist;
            iflist->prev = ifds;
          }
          iflist = ifds;
          ++pds_p->ifcnt;		/* increment port reference count */
        }
      }
    }
  }
  return err;
} /* flipio_add() */

/*============================================================================
 * FLIPIO_DELETE Handler. Delete Network Interface
 * ---------------------------------------------------------------------------
 * o Scan interface list and delete entry with matching interface number
 * On error return:
 *   FLIPER_IFNEXIST	- interface not found
 *   FLIPER_BADSTATE	- interface is active
 */
static int flipio_delete (int *ifno_p)
{
  int		err	= FLIPER_IFNEXIST;
  IfData_t	*iftmp	= iflist;

  while (iftmp)
  {
    if (iftmp->ifno == *ifno_p)
    {
      if (if_delete(iftmp)) err = 0;
      else err = FLIPER_BADSTATE;
      break;
    }
    else iftmp = iftmp->next;
  }
  return err;
} /* flipio_delete() */

/*============================================================================
 * FLIPIO_IFSTATUS Handler. Get Interface Status.
 * ---------------------------------------------------------------------------
 * Return size of the interface status block.
 * On error return:
 *   FLIPER_IFNEXIST	- interface not found
 */
static int flipio_ifstatus (int *ifno_p)
{
  int		err	= FLIPER_IFNEXIST;
  IfData_t	*iftmp;

  for (iftmp = iflist; iftmp != NULL; iftmp = iftmp->next)
    if (iftmp->ifno == *ifno_p)
    {

      err = 0;
      break;
    }
  ;
  return err ? err : sizeof(flipIfSts_t);
} /* flipio_ifstatus() */

/******* Network Device Driver Entry Points *********************************/

/*============================================================================
 * Initialize Network Interface.
 * ---------------------------------------------------------------------------
 */
static int flip_ifinit (struct device *dev)
{
#define	ids_p	((IfData_t *)dev)

  Port_t	*pds_p	= ports[ids_p->port];	/* -> port data space */
  int		i;

  /*
   * Initialize device structure.
   */
  dev->open		= flip_ifopen;
  dev->stop		= flip_ifclose;
  dev->hard_start_xmit	= flip_ifsend;
  dev->get_stats	= flip_ifstats;
  dev->hard_header	= flip_header;
  dev->rebuild_header	= flip_rebuild_h;
#if (_OS_VER_ < 103)
  dev->type_trans	= flip_type_trans;
#endif
  dev->family		= AF_INET;		/* address family */
  dev->type		= ARPHRD_FR;		/* ARP interface h/w type */
  dev->mtu		= pds_p->cfg.mtu - FLIP_HDR_SZ;
  dev->pa_alen		= sizeof(unsigned long); /* protocol address length */
  dev->hard_header_len	= FLIP_HDR_SZ;		/* media header length */
  dev->addr_len		= FR_HWADDR_LEN;	/* hardware address length */
  dlci_to_Q922(ids_p->dlci, dev->dev_addr);	/* hardware address */

  /*
   * Initialize socket buffers.
   */
  for (i = 0; i < DEV_NUMBUFFS; ++i)
    skb_queue_head_init(&dev->buffs[i])
  ;
  return 0;

#undef	ids_p
} /* flip_ifinit() */

/*============================================================================
 * Open Network Interface.
 * ---------------------------------------------------------------------------
 * o Check if DLCI is active. If not, then print a warning, but do not return
 *   an error in order not to break ifconfig.
 * o Initialize some flags.
 * o Increment module use count to prevemt the module from unloading while
 *   there are still active interfaces.
 */
static int flip_ifopen (struct device *dev)
{
#define	ids_p	((IfData_t *)dev)

  int		err	= 0;
  Port_t	*pds_p	= ports[ids_p->port];	/* -> port data space */
  int		i;

  if (dev->start)
    err = -EBUSY
  ;
  else if (pds_p->fr_func->status(pds_p))
    err = -EIO
  ;
  else
  {
    if (!(pds_p->status & FLIP_LINK_UP))
      printk ("%s: Warning: frame relay link on port %d is down!\n",
              drv_name, ids_p->port)
    ;
    else
    {
      for (i = 0; i < pds_p->dlci_num; ++i)
        if (pds_p->dlci[i] == ids_p->dlci)
        {
          i = 0;
          break;
        }
      ;
      if (i)
        printk ("%s: Warning: DLCI %d on port %d is not active!\n",
                drv_name, ids_p->dlci, ids_p->port)
      ;
    }
    dev->tbusy = 0;
    dev->interrupt = 0;
    dev->start = 1;
  }
  return err;

#undef	ids_p
} /* flip_ifopen() */

/*============================================================================
 * Close Network Interface.
 * ---------------------------------------------------------------------------
 * o Reset flags.
 * o Decrement module use count.
 */
static int flip_ifclose (struct device *dev)
{
#define	ids_p	((IfData_t *)dev)

  ids_p->flags = 0;
  dev->start = 0;
  return 0;

#undef	ids_p
} /* flip_ifclose() */

/*============================================================================
 * Get Interface Statistics.
 * ---------------------------------------------------------------------------
 */
static IfStats_t *flip_ifstats (struct device *dev)
{
#define	ids_p	((IfData_t *)dev)

  return &ids_p->stats;

#undef	ids_p
} /* flip_ifstats() */

/*============================================================================
 * Send a Packet on a Frame Relay Interface.
 * ---------------------------------------------------------------------------
 * Here is the trick:
 *   if the first byte is 0, then 2-byte encapsulation header was used and we
 *   have to adjust packet length and its beginning accordingly.
 */
static int flip_ifsend (struct sk_buff *skb, struct device *dev)
{
  int	err = 0;

#define	ids_p	((IfData_t *)dev)

  if (skb && dev)
  {
    Port_t	*pds_p = ports[ids_p->port];
    int		adjust;

    if (set_bit(0, (void*)&dev->tbusy))
    {
      if (SDLA_DEBUG > 1)
        printk ("%s: Transmit re-entrancy problem!\n", drv_name)
      ;
      err = 1;		/* retry later */
    }
    else
    {
      adjust = (*skb->data == 0) ? FLIP_HDR_SZ - sizeof(Encaps2_t) : 0;
      switch (pds_p->fr_func->send(
			pds_p,
			ids_p->dlci,
			skb->data + adjust,
			skb->len - adjust))
      {
        case 0:		/* O.k. */
          ++ids_p->stats.tx_packets;
          break;

        case 1:		/* Tx buffers full - drop packet */
          ++ids_p->stats.tx_dropped;
          break;

        case 2:		/* Adapter is busy - retry later */
          err = 1;
          break;

        default:	/* other error - discard packet */
          ++ids_p->stats.tx_errors;
      }
      if (!err) dev_kfree_skb(skb, FREE_WRITE);
      dev->tbusy = 0;
    }
  }
  return err;

#undef	ids_p
} /* flip_ifsend() */

/*============================================================================
 * Build media header.
 * ---------------------------------------------------------------------------
 * Form RFC-1490 -compliant header, return media header length.
 * Here we do the trick:
 *   for IP and some other protocols which have NLPID assigned to them,
 *   RFC-1490 allows 2-byte encapsulation header.  Since protocol stack
 *   assumes constant header size, we'll align header's end and place 0 at
 *   its beginning for transmit routine to know that this is a short header.
 */
static int flip_header (
#if (_OS_VER_ < 200)
	unsigned char *buff, struct device *dev, unsigned short type,
	void *daddr, void *saddr, unsigned len, struct sk_buff *skb)
#else
	struct sk_buff *skb, struct device *dev, unsigned short type,
	void *daddr, void *saddr, unsigned len)
#endif
{
  int	hdrlen = FLIP_HDR_SZ;
#if (_OS_VER_ >= 200)
  unsigned char* buff = skb_push(skb, hdrlen);
#endif

  switch (type)
  {
    case ETH_P_IP:			/* IP datagram encapsulation */
#define	hdr	((Encaps2_t*)(buff + FLIP_HDR_SZ - sizeof(Encaps2_t)))
      buff[0] = 0;
      hdr->control = Q922_UI;
      hdr->nlpid   = NLPID_IP;
#undef	hdr
      break;

    case ETH_P_IPX:			/* SNAP encapsulation */
    case ETH_P_ARP:
      /* what else ? */
#define	hdr	((Encaps_t*)buff)
      hdr->control = Q922_UI;
      hdr->pad     = 0x00;
      hdr->nlpid   = NLPID_SNAP;
      hdr->pid     = type;
      memcpy(hdr->oui, OUI_Ether, 3);
#undef	hdr
      break;

    default:				/* Unknown packet type */
      hdrlen = 0;
      printk("%s: flip_header(): Unknown packet type 0x%X\n", drv_name, type);
  }
  return hdrlen;
} /* flip_header() */

/*============================================================================
 * Rebuild media header.
 * ---------------------------------------------------------------------------
 */
static int flip_rebuild_h (void *buff, struct device *dev, unsigned long dst,
			   struct sk_buff *skb)
{
#if	_DEBUG_ >= 2
  printk("%s: fr_rebuild_header() called.\n", drv_name);
#endif
  return 0;
} /* flip_rebuild_h() */

#if (_OS_VER_ < 103)
/*============================================================================
 * Determine packet's protocol ID
 * ---------------------------------------------------------------------------
 */
static unsigned short flip_type_trans (struct sk_buff *skb, struct device *dev)
{
  return (unsigned short)skb->pkt_class;
} /* flip_type_trans() */
#endif

/*============================================================================
 * Receive call-back function.
 * ---------------------------------------------------------------------------
 * ! This function is called on interrupt context.
 */
int flip_receive (Port_t *pds_p, int dlci, char *buf, int len)
{
  IfData_t		*ids_p;
  struct sk_buff	*skb_p;
  unsigned short	pkt_type;
  int			hdr_len, adjust;

  ids_p = if_find(pds_p->cfg.port, dlci);
  if (ids_p)
  {
    ids_p->dev.interrupt = 1;		/* set 'inside ISR' flag */
    if (len <= 2)
    printk("%s: Received frame too short!!!\n", drv_name)
    ;
    else
    {
      pkt_type = decapsulate((void*)buf, &hdr_len);
      if (pkt_type)
      {
        adjust = FLIP_HDR_SZ - hdr_len;
        skb_p = alloc_skb(len + adjust, GFP_ATOMIC);
        if (skb_p == NULL)
          ++ids_p->stats.rx_dropped
        ;
        else
        {
          memcpy(skb_p->data + adjust, buf, len);
          skb_p->len = len + adjust;
          skb_p->dev = &ids_p->dev;
#if (_OS_VER_ < 103)
          skb_p->pkt_class = pkt_type;
#else
          skb_p->protocol = pkt_type;
#endif
#if (_OS_VER_ >= 200)
          skb_p->mac.raw = skb_p->data;
          skb_pull(skb_p, FLIP_HDR_SZ);
#endif
          netif_rx(skb_p);
          ++ids_p->stats.rx_packets;
        }
      }
    }
    ids_p->dev.interrupt = 0;		/* reset 'inside ISR' flag */
  }
  kfree(buf);
  return 0;
} /* flip_receive() */

/****** Support Functions ***************************************************/

/*============================================================================
 * Shutdown Port.
 * ---------------------------------------------------------------------------
 * o Delete all network interfaces on this port.
 * o Disable communications on this port.
 * o Disable polling (for non interrupt-driven ports only)
 * Return 0 if port is busy (i.e. there are active interfaces).
 */
static int port_down (Port_t *pds_p)
{
  int	port = pds_p->cfg.port;
  register IfData_t *ifptr = iflist;
  register IfData_t *iftmp;

  /*
   * Delete all interfaces for this port.
   */
  while (ifptr)
  {
    iftmp = ifptr;
    ifptr = ifptr->next;
    if (iftmp->port == port) if_delete(iftmp);
  }
  if (pds_p->ifcnt)
    return 0				/* some interfaces are in use */
  ;

  pds_p->flags &= ~PORT_READY;		/* clear 'ready' flag  */
  if (pds_p->flags & PORT_POLLED)
  {
	/* !!! disable polling !!! */
  }
  pds_p->fr_func->shutdown(pds_p);
  return 1;
} /* port_down() */

/*============================================================================
 * Delete Network Interface.
 * ---------------------------------------------------------------------------
 * o De-register network device.
 * o Un-link interface data space.
 * o Release kernel memory.
 * o Decrement port reference count.
 * Return 0 if interface is busy.
 */
static int if_delete (IfData_t *ids_p)
{
  if (ids_p->dev.start)
    return 0				/* interface in use */
  ;

  /*
   * De-register interface.
   */
  unregister_netdev(&ids_p->dev);

  /*
   * Unlink interface data space.
   */
  if (ids_p->prev)
    ids_p->prev->next = ids_p->next
  ;
  else iflist = ids_p->next;
  if (ids_p->next)
    ids_p->next->prev = ids_p->prev
  ;

  kfree(ids_p);				/* release memory */
  --ports[ids_p->port]->ifcnt;		/* update reference count */
  return 1;
} /* if_delete() */

/*============================================================================
 * Find Network Interface.
 * ---------------------------------------------------------------------------
 * Return a pointer to the interface data space or NULL if interface is not
 * found.
 */
static IfData_t *if_find (int port, int dlci)
{
  register IfData_t *if_p = iflist;

  while (if_p)
  {
    if ((if_p->port == port) && (if_p->dlci == dlci)) break;
    else if_p = if_p->next;
  }
  return if_p;
} /* if_find() */

/*============================================================================
 * Get a pointer from a look-up table.
 * ---------------------------------------------------------------------------
 */
static void *lookup (LookUp_t *table, int token)
{
  while (table->token && (table->token != token)) table++;
  return table->ptr;
} /* lookup() */

/*============================================================================
 * Covert DLCI number to Q.922 address (as a byte string).
 * ---------------------------------------------------------------------------
 */
static void dlci_to_Q922 (unsigned int dlci, char *str)
{
  *str = (dlci >> 2) & 0xFC;
  ++str;
  *str = (dlci << 4) & 0xF0;
} /* dlci_to_Q922() */

/*============================================================================
 * Decapsulate RFC-1490 frame.
 * ---------------------------------------------------------------------------
 * Return frame type in network byte order or zero, if frame type is unknown.
 * Place a media header length at *hdrlen_p.
 */
static unsigned short decapsulate (Encaps_t *frm_p, unsigned *hdrlen_p)
{
  unsigned short pkt_type = 0;

  if (frm_p->control != Q922_UI)
    printk("%s: Invalid Q.922 frame type 0x%02X.\n", drv_name, frm_p->control)
  ;
  else switch (frm_p->pad)
  {
    case NLPID_IP:		/* IP datagramm */
      pkt_type = ETH_P_IP;
      *hdrlen_p = sizeof(Encaps2_t);
      break;

	/* add other protocols, e.g. CLNP, ESIS, ISIS, if needed */

    case 0:			/* padded header */
      switch (frm_p->nlpid)
      {
        case NLPID_SNAP:
          if (memcmp(frm_p->oui, OUI_Ether, 3) == 0)
          {
            pkt_type = frm_p->pid;	/* Ethertype */
            *hdrlen_p = 8;
          }
          else printk("%s: Unsupported SNAP OUI %02X-%02X-%02X.\n", drv_name,
		frm_p->oui[0],
		frm_p->oui[0],
		frm_p->oui[0])
          ;
          break;

		/* add other protocols, if needed */

        default:
          printk("%s: Unknown encapsulation type 0x%02X.\n", drv_name,
		frm_p->nlpid)
          ;
      }
      break;

    default:
      printk("%s: Unknown NLPID encapsulation 0x%02X.\n", drv_name,
	frm_p->pad)
      ;
  }
  return htons(pkt_type);
} /* decapsulate() */

/****** End *****************************************************************/

