/*****************************************************************************
* PROJECT:	Sangoma Frame Relay Network Device Driver
* FILE:		fr508.c		Wrapper for S508 Frame Relay Implementation
* TARGET OS:	Linux
* AUTHOR:	Gene Kozin
* ----------------------------------------------------------------------------
* Copyright (c) 1995-1996 Sangoma Technologies Inc.  All Rights Reserved.
* ============================================================================
*  Rev	Date		Who	Details
* -----	-----------	---	----------------------------------------------
*  1.40	Dec.18,1996	GK	o support for configurable link management
*  1.10	Jun.18,1996	GK	o Updated for SDLA Driver 3.0 compatibility.
*				o Use adapter mailbox access arbitration.
*  1.00	Mar.11,1996	GK	Beta release.
*  0.22	Feb.12,1996	GK	Reflects changes to frame relay firmware API.
*  0.20	Dec.13,1995	GK	Initial version (Alpha release).
*****************************************************************************/

#include <linux/config.h>	/* OS configuration options */
#include <linux/stddef.h>	/* offsetof(), etc. */
#include <linux/errno.h>	/* return codes */
#include <linux/kernel.h>
#include <linux/malloc.h>	/* kmalloc(), kfree() */
#include <linux/interrupt.h>
#include <linux/inet.h>		/* in_aton(), in_ntoa() prototypes */
#include <linux/netdevice.h>	/* support for network drivers */
#include <linux/if.h>		/* network interface stuff */
#include <linux/if_ether.h>	/* for ethertypes */
#include <linux/skbuff.h>	/* for struct sk_buff */
#include <asm/string.h>		/* inline mem*, str* functions */
#include <asm/byteorder.h>	/* htons(), etc. */

#include "sdladrv.h"		/* SDLA driver kernel interface definitions */
#include "flipio.h"		/* this driver API definitions */
#include "flip.h"		/* common defines */

#define	_FR508_
#include "fr_api.h"		/* SDLA frame relay firmware defines */

/****** Defines *************************************************************/

/****** Data Types **********************************************************/

/****** Function Prototypes *************************************************/

/*
 * Implementation-independent Interface Functions.
 */
static int	configure	(Port_t *pds_p);
static int	shutdown	(Port_t *pds_p);
static int	send		(Port_t *pds_p, int dlci, char *buf, int len);
static int	interrupt	(Port_t *pds_p);
static int	status		(Port_t *pds_p);
static int	statistics	(Port_t *pds_p);

/*
 * Frame Relay Firmware-Specific Functions.
 */
static int	fr_read_version	(Port_t *pds_p, char *str);
static int	fr_set_config	(Port_t *pds_p, frConf_t *conf_p, int dlcinum,
				 short *dlcilist);
static int	fr_read_config	(Port_t *pds_p, frConf_t *conf_p, short *dlci);
static int	fr_set_intr	(Port_t *pds_p, frIntrCtl_t *ictl_p);
static int	fr_config_dlc	(Port_t *pds_p, int dlci, frDLCConf_t *conf_p);
static int	fr_comm_enable	(Port_t *pds_p);
static int	fr_comm_disable	(Port_t *pds_p);
static int	fr_issue_isf	(Port_t *pds_p, int isf_type);
static int	fr_write	(Port_t *pds_p, int dlci, int len, int attr);
static int	fr_add_dlci	(Port_t *pds_p, int dlcinum, short *dlci);
static int	fr_activate_dlci(Port_t *pds_p, int dlcinum, short *dlci);
static int	fr_get_status	(Port_t *pds_p, unsigned dlci, unsigned *ptr);
static int	fr_dlci_list	(Port_t *pds_p, unsigned *list);
static int	fr_link_stat	(Port_t *pds_p, frLinkStat_t *stat_p);
static int	fr_comm_stat	(Port_t *pds_p, frCommStat_t *stat_p);
static int	fr_event	(Port_t *pds_p, int event, int cmd);

static int	set_lock	(int *lck_p);

/****** Global Data **********************************************************
* Note:	All data must be explicitly initialized!!!
*/

frFunc_t	fr508_func =	/* Frame Relay Function Call Table */
{
  &configure,			/* configure() */
  &shutdown,			/* shutdown() */
  &send,			/* send() */
  &interrupt,			/* interrupt() */
  &status,			/* status() */
  &statistics,			/* statistics() */
  &fr_read_version,		/* version() */
};

/****** Interface Functions *************************************************/

/*============================================================================
 * Configure Frame Relay Port.
 * ---------------------------------------------------------------------------
 * On error return:
 *   FLIPER_NEXIST	- port configuration failed
 *   FLIPER_NEXIST	- firmware failure
 */
static int configure (Port_t *pds_p)
{
  int		err	= 0;
  sdlaData_t	*ads	= pds_p->ads;
  frConf_t	frcfg;		/* firmware configuration block */
  frIntrCtl_t	intrctl;
  unsigned short dlcilist[DLCI_NUM];
  int		i, dlcinum;

  pds_p->cfg.mtu = min(pds_p->cfg.mtu, FR_MAX_DATA);
  pds_p->cfg.CIR = min(pds_p->cfg.CIR, pds_p->cfg.speed);

  memset(&frcfg, 0, sizeof(frcfg));
  frcfg.conf_flags =
	FRCFG_DIRECT_RX |	/* enable direct receive */
	FRCFG_IGNORE_TX_CIR
  ;
  switch (pds_p->cfg.signalling)
  {
  case 1:	frcfg.conf_flags |= 0x0400; break;	/* LMI */
  case 2:	frcfg.conf_flags |= 0x0200; break;	/* Q.933 */
  case 3:	frcfg.conf_flags |= 0x0800; break;	/* none */
  }
  frcfg.station_type  = pds_p->cfg.station ?
	FRCFG_STATION_NODE : FRCFG_STATION_CPE
  ;
  if (pds_p->cfg.clocking)
    frcfg.port_mode |= FRCFG_MODE_INT_CLK
  ;
  if (pds_p->cfg.type)
    frcfg.port_mode |= FRCFG_MODE_RS232
  ;
  frcfg.baud_rate = min((pds_p->cfg.speed / 1000), 2048);
  frcfg.max_txdata = pds_p->cfg.mtu;
  frcfg.CIR_fwd = frcfg.CIR_bwd = min((pds_p->cfg.CIR / 1000), 512);
  frcfg.Bc_fwd  = frcfg.Bc_bwd = frcfg.CIR_fwd;
  frcfg.Be_fwd  = frcfg.Be_bwd = 0;
  frcfg.T391	= pds_p->cfg.T391;
  frcfg.T392	= pds_p->cfg.T392;
  frcfg.N391	= pds_p->cfg.N391;
  frcfg.N392	= pds_p->cfg.N392;
  frcfg.N393	= pds_p->cfg.N393;
  if (frcfg.station_type == FRCFG_STATION_NODE)
  {
    for (i = 0; i < DLCI_NUM; ++i)
/*
      dlcilist[i] = pds_p->cfg.dlci[i]
*/
      dlcilist[i] = 16 + i;
    ;
    dlcinum = DLCI_NUM;
  }
  else
  {
    frcfg.conf_flags |= FRCFG_AUTO_CONFIG;	/* autoconfigure DLCIs */
    dlcinum = 0;
  }
  if ((fr_comm_disable(pds_p) != FRRES_OK) ||
      (fr_set_config(pds_p, &frcfg, dlcinum, dlcilist) != FRRES_OK) ||
      (fr_comm_enable(pds_p) != FRRES_OK))
    err = FLIPER_NEXIST			/* port configuration failed */
  ;
  else
  {
    /*
     * Enable adapter interrupts (for interrupt-driven ports only).
     * Interrupt flag must be cleared to allow interrupts to occur!
     */
    if (!(pds_p->flags & PORT_POLLED))
    {
      memset(&intrctl, 0, sizeof(intrctl));
      intrctl.flags = FR_INTR_RXRDY;
      intrctl.irq = ads->irq;
      if (fr_set_intr(pds_p, &intrctl) != FRRES_OK)	/* enable interrupts */
        err = FLIPER_NEXIST
      ;
      /*
       * Clear interrupt status flags.
       */
      else ((frFlags_t *)(pds_p->wbase + FR_FLAGS_OFFS))->intr_flag = 0;
    }
    if (!err)
    {
      /*
       * Kick off the Frame Relay link:
       *  CPE:		issue Full Status Enquiry message
       *  Access Node:	add and then activate configured DLCs
       */
      if (frcfg.station_type == FRCFG_STATION_CPE)
      {
        fr_issue_isf(pds_p, FR_ISF_FSE);
      }
      else
      {
        fr_add_dlci(pds_p, dlcinum, dlcilist);
        fr_activate_dlci(pds_p, dlcinum, dlcilist);
      }
    }
  }
  return err;
} /* configure() */

/*============================================================================
 * Shut Down Frame Relay Port.
 * ---------------------------------------------------------------------------
 * On error return:
 *   FLIPER_NEXIST	- firmware failure
 */
static int shutdown (Port_t *pds_p)
{
  return fr_comm_disable(pds_p) ? FLIPER_NEXIST : 0;
} /* shutdown() */

/*============================================================================
 * Read Frame Relay Port Status.
 * ---------------------------------------------------------------------------
 * On error return:
 *   FLIPER_NEXIST	- firmware failure
 */
static int status (Port_t *pds_p)
{
  int		err = 0;
  unsigned	fr_status;

  if (fr_get_status(pds_p, 0, &fr_status) ||
      fr_dlci_list(pds_p, (int *)(&pds_p->dlci_num)))
    err = -1
  ;
  else
  {
    frFlags_t	*fr_flags_p = (void *)(pds_p->wbase + FR_FLAGS_OFFS);

    pds_p->status = 0;
    if (fr_status & FR_LINK_OPER)
      pds_p->status |= FLIP_LINK_UP
    ;
    if (fr_flags_p->mdm_status & FR_MDM_DCD)
      pds_p->status |= FLIP_MODEM_DCD
    ;
    if (fr_flags_p->mdm_status & FR_MDM_CTS)
      pds_p->status |= FLIP_MODEM_CTS
    ;
  }
  return err;
} /* status() */

/*============================================================================
 * Read Frame Relay Port Statistics.
 * ---------------------------------------------------------------------------
 * On error return:
 *   FLIPER_NEXIST	- firmware failure
 */
static int statistics (Port_t *pds_p)
{
  int		err = 0;
  frLinkStat_t	linkstat;
  frCommStat_t	commstat;

  if (fr_link_stat(pds_p, &linkstat) ||
      fr_comm_stat(pds_p, &commstat))
    err = -1
  ;
  else
  {
    pds_p->rx_overruns   = commstat.rx_overruns;
    pds_p->rx_aborts     = commstat.rx_aborts;
    pds_p->rx_bad_crc    = commstat.rx_bad_crc;
    pds_p->rx_too_long   = commstat.rx_too_long;
    pds_p->tx_underruns  = commstat.tx_underruns;
    pds_p->tx_aborts     = commstat.tx_aborts;
    pds_p->dcd_low_cnt   = commstat.dcd_dropped;
    pds_p->cts_low_cnt   = commstat.cts_dropped;
    pds_p->rx_bad_dlci   = linkstat.rx_bad_dlci;
    pds_p->rx_bad_format = linkstat.rx_bad_format;
    pds_p->rx_dropped    = linkstat.rx_too_long +
                           linkstat.rx_dropped +
                           linkstat.rx_dropped2;
    pds_p->tx_retries    = linkstat.retransmitted;
  }
  return err;
} /* statistics() */

/*============================================================================
 * Send a Packet on a Frame Relay Port.
 * ---------------------------------------------------------------------------
 * Return:
 *   0	O.k.
 *   1	Tx buffers full
 *   2	Adapter is busy
 *   3	Other
 */
static int send (Port_t *pds_p, int dlci, char *buf, int len)
{
  int		err	= 0;
  sdlaData_t	*ads	= pds_p->ads;
  int		frerr;

  /*
   * Get offset of the current transmit status element.
   */
  frerr = fr_write(pds_p, dlci, len, 0);
  if (frerr != FRRES_OK)
  {
    switch (frerr)
    {
      case FRRES_BUFFER_OVERFLOW:
	/* !!!!!!
	 * look into enabling Tx interrups here...
         */
        err = 1;
        break;

      case FRRES_MB_INUSE:
        err = 2;
        break;

      default:
        err = 3;
    }
  }
  else
  {
    unsigned long flags;

    frBufCtl_t	*bctl_p = (void *)(pds_p->wbase + pds_p->tscrnt);

    save_flags(flags);
    cli();
    if (sdla_poke(ads, bctl_p->offset, buf, len))
       err = -1
    ;
    else bctl_p->flag = 0x01;		/* start transmission */
    restore_flags(flags);
  }
  return err;
} /* send() */

/*============================================================================
 * Frame Relay Port Interrupt Service Routine.
 * ---------------------------------------------------------------------------
 * Receive interrupt:
 *  o Allocate buffer and copy a frame into it.
 *  o Pass a buffer to receive routine.
 */
static int interrupt (Port_t *pds_p)
{
  int		err = 0;
  sdlaData_t	*ads	= pds_p->ads;
  frFlags_t	*flags_p = (void *)(pds_p->wbase + FR_FLAGS_OFFS);

  switch (flags_p->intr_flag)
  {
    case FR_INTR_RXRDY:
      /*
       * Receive interrupt.
       */
      {
        frBufCtl_t	*bctl_p	= (void *)(pds_p->wbase + pds_p->rscrnt);
        int		len, dlci, offs, split;
        char		*buf;

        ++pds_p->rxint_cnt;
        if (bctl_p->flag != 0x01)
          printk("%s: Port %d receive buffers corrupted!\n",
                 DRV_NAME, pds_p->cfg.port)
        ;
        else
        {
          len  = bctl_p->length;
          dlci = bctl_p->dlci;
          offs = bctl_p->offset;
/*
#if	_DEBUG_ >= 2
          printk("Received %d bytes on DLCI %d:\n", len, dlci);
#endif
*/
          buf = kmalloc(len, GFP_ATOMIC);
          if (buf == NULL)
            printk("%s: Port %d receive memory allocation error!\n",
                   DRV_NAME, pds_p->cfg.port)
          ;
          else
          {
            split = ((offs + len) > (pds_p->rxtop + 1)) ?
                    len - (pds_p->rxtop - offs + 1) : 0
            ;
            if (sdla_peek(ads, bctl_p->offset, buf, len - split))
              err = -1
            ;
            else if (split)
            {
              if (sdla_peek(ads, pds_p->rxbase, buf + len - split, split))
                err = -1
              ;
            }
            if (!err) flip_receive(pds_p, dlci, buf, len);
            else
            {
              printk("%s: Port %d: receive error!\n",
                     DRV_NAME, pds_p->cfg.port)
              ;
              kfree(buf);
            }
          }
          bctl_p->flag = 0;
          pds_p->rscrnt += sizeof(frBufCtl_t);
          if (pds_p->rscrnt > pds_p->rslast)
            pds_p->rscrnt = pds_p->rsbase
          ;
        }
      }
      break;

    default:
        /*
         * Spurious interrupt.
         */
        ++pds_p->spuri_cnt;
  }
  flags_p->intr_flag = 0;		/* clear interrupt flags */
  return err;
} /* interrupt() */

/****** Frame Relay Firmware-Specific Functions *****************************/

/*============================================================================
 * Set Lock.
 * ---------------------------------------------------------------------------
 * Returns 1 if locked successfully, or 0 otherwise.
 */
static int set_lock (int *lck_p)
{
  if (*lck_p) return 0;

  if (++(*lck_p) > 1)
  {
    --(*lck_p);
    return 0;
  }
  return 1;
}

/*============================================================================
 * FR_READ_CODE_VERSION Command.
 * ---------------------------------------------------------------------------
 * Fill string str with firmware version info. 
 */
static int fr_read_version (Port_t *pds_p, char *str)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;
  int		len;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd = FR_READ_CODE_VERSION;
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      if ((frerr == FRRES_OK) && str)
      {
        len = cmd_p->length;
        memcpy(str, cmd_p->data, len);
        str[len] = '\0';
      }
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_READ_CODE_VERSION) && --retry);
  return frerr;
} /* fr_read_version() */

/*============================================================================
 * FR_SET_CONFIG Command (Global).
 * ---------------------------------------------------------------------------
 * Set Global Configuration Parameters. 
 */
static int fr_set_config (Port_t *pds_p, frConf_t *conf_p, int dlcinum,
			  short *dlcilist)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;
  frBufInfo_t	*bufi_p;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd    = FR_SET_CONFIG;
      cmd_p->length = sizeof(frConf_t) + dlcinum * sizeof(short);
      memcpy(cmd_p->data, conf_p, sizeof(frConf_t));
      if (dlcinum)
        memcpy(((frConf_t *)cmd_p->data)->dlci, dlcilist, dlcinum << 1)
      ;
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      if (frerr == FRRES_OK)
      {
        bufi_p = (void *)(pds_p->wbase + FR_RXBUF_INFO);
        pds_p->rsbase = bufi_p->rse_base - FR_CTL_WINDOW;
        pds_p->rslast = pds_p->rsbase +
                        (bufi_p->rse_num - 1) * sizeof(frBufCtl_t)
        ;
        pds_p->rscrnt = bufi_p->rse_next - FR_CTL_WINDOW;
        pds_p->rxbase = bufi_p->buf_base;
        pds_p->rxtop  = bufi_p->buf_top;
      }
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_SET_CONFIG) && --retry);
  return frerr;
} /* fr_set_config() */

/*============================================================================
 * FR_READ_CONFIG Command (Global).
 * ---------------------------------------------------------------------------
 * Read Global Configuration Parameters. 
 */
static int fr_read_config (Port_t *pds_p, frConf_t *conf_p, short *dlci)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;
  int		dlcinum;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd    = FR_READ_CONFIG;
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      if ((frerr == FRRES_OK) && conf_p)
      {
        memcpy(conf_p, cmd_p->data, sizeof(frConf_t));
        dlcinum = (cmd_p->length - sizeof(frConf_t)) >> 1;
        if (dlci && dlcinum)
          memcpy(dlci, ((frConf_t *)cmd_p->data)->dlci, dlcinum << 1)
        ;
      }
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_READ_CONFIG) && --retry);
  return frerr;
} /* fr_read_config() */

/*============================================================================
 * FR_SET_INTR_FLAGS Command.
 * ---------------------------------------------------------------------------
 * Set interrupt enable flags. 
 */
static int fr_set_intr (Port_t *pds_p, frIntrCtl_t *ictl_p)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd     = FR_SET_INTR_FLAGS;
      cmd_p->length  = sizeof(frIntrCtl_t);
      memcpy(cmd_p->data, ictl_p, sizeof(frIntrCtl_t));
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_SET_INTR_FLAGS) && --retry);
  return frerr;
} /* fr_set_intr() */

/*============================================================================
 * FR_SET_CONFIG Command (Single Channel).
 * ---------------------------------------------------------------------------
 * Set Global Configuration Parameters. 
 */
static int fr_config_dlc (Port_t *pds_p, int dlci, frDLCConf_t *conf_p)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd    = FR_SET_CONFIG;
      cmd_p->dlci   = dlci;
      cmd_p->length = sizeof(frDLCConf_t);
      memcpy(cmd_p->data, conf_p, sizeof(frDLCConf_t));
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_SET_CONFIG) && --retry);
  return frerr;
} /* fr_config_dlc() */

/*============================================================================
 * FR_COMM_ENABLE Command.
 * ---------------------------------------------------------------------------
 * Enable Communications. 
 */
static int fr_comm_enable (Port_t *pds_p)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd = FR_COMM_ENABLE;
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_COMM_ENABLE) && --retry);
  return frerr;
} /* fr_comm_enable() */

/*============================================================================
 * FR_COMM_DISABLE Command.
 * ---------------------------------------------------------------------------
 * Disable Communications. 
 */
static int fr_comm_disable (Port_t *pds_p)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd = FR_COMM_DISABLE;
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_COMM_DISABLE) && --retry);
  return frerr;
} /* fr_comm_disable() */

/*============================================================================
 * FR_WRITE Command.
 * ---------------------------------------------------------------------------
 * Send a frame. 
 */
static int fr_write (Port_t *pds_p, int dlci, int len, int attr)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd    = FR_WRITE;
      cmd_p->dlci   = dlci;
      cmd_p->attr   = attr;
      cmd_p->length = len;
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      if (frerr == FRRES_OK)
      {
        pds_p->tscrnt  = *((unsigned *)&cmd_p->data[0]) - FR_CTL_WINDOW;
      }
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_WRITE) && --retry);
  return frerr;
} /* fr_write() */

/*============================================================================
 * FR_ISSUE_IS_FRAME Command.
 * ---------------------------------------------------------------------------
 * Issue In-channel Signalling frame. 
 */
static int fr_issue_isf (Port_t *pds_p, int isf_type)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd     = FR_ISSUE_IS_FRAME;
      cmd_p->length  = 1;
      cmd_p->data[0] = isf_type;
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_ISSUE_IS_FRAME) && --retry);
  return frerr;
} /* fr_issue_isf() */

/*============================================================================
 * FR_ADD_DLCI Command.
 * ---------------------------------------------------------------------------
 * Add DLCI(s) (Access Node only!). 
 */
static int fr_add_dlci (Port_t *pds_p, int dlcinum, short *dlci)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd    = FR_ADD_DLCI;
      cmd_p->length = dlcinum << 1;
      memcpy(cmd_p->data, dlci, dlcinum << 1);
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_ADD_DLCI) && --retry);
  return frerr;
} /* fr_add_dlci() */

/*============================================================================
 * FR_ACTIVATE_DLCI Command.
 * ---------------------------------------------------------------------------
 * Activate DLCI(s) (Access Node only!). 
 */
static int fr_activate_dlci (Port_t *pds_p, int dlcinum, short *dlci)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd    = FR_ACTIVATE_DLCI;
      cmd_p->length = dlcinum << 1;
      memcpy(cmd_p->data, dlci, dlcinum << 1);
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_ACTIVATE_DLCI) && --retry);
  return frerr;
} /* fr_activate_dlci() */

/*============================================================================
 * FR_READ_STATUS Command.
 * ---------------------------------------------------------------------------
 * Get a list of all active DLCIs. 
 * First element of the list will contain number of active DLCIs.
 */
static int fr_get_status (Port_t *pds_p, unsigned dlci, unsigned *ptr)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd  = FR_READ_STATUS;
      cmd_p->dlci = dlci;
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      if (frerr == FRRES_OK)
      {
        *ptr = cmd_p->data[0];
      }
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_READ_STATUS) && --retry);
  return frerr;
} /* fr_get_status() */

/*============================================================================
 * FR_LIST_ACTIVE_DLCI Command.
 * ---------------------------------------------------------------------------
 * Get a list of all active DLCIs. 
 * First element of the list will contain number of active DLCIs.
 */
static int fr_dlci_list (Port_t *pds_p, unsigned *list)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;
  int		dlcinum, i;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd     = FR_LIST_ACTIVE_DLCI;
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      if (frerr == FRRES_OK)
      {
        dlcinum = cmd_p->length >> 1;
        if (dlcinum > DLCI_NUM) dlcinum = DLCI_NUM;
        *list++ = dlcinum;
        for (i = 0; i < dlcinum; ++i)
          *list++ = ((unsigned short *)cmd_p->data)[i]
        ;
      }
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_LIST_ACTIVE_DLCI) && --retry);
  return frerr;
} /* fr_dlci_list() */

/*============================================================================
 * Execute FR_READ_STATISTICS Command for DLCI == 0 (global).
 * ---------------------------------------------------------------------------
 */
static int fr_link_stat (Port_t *pds_p, frLinkStat_t *stat_p)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd     = FR_READ_STATISTICS;
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      if (frerr == FRRES_OK)
      {
        memcpy(stat_p, cmd_p->data, sizeof(frLinkStat_t));
      }
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_READ_STATISTICS) && --retry);
  return frerr;
} /* fr_link_stat() */

/*============================================================================
 * Execute FR_READ_ERROR_STATS Command.
 * ---------------------------------------------------------------------------
 */
static int fr_comm_stat (Port_t *pds_p, frCommStat_t *stat_p)
{
  frCmd_t	*cmd_p	= (void *)(pds_p->wbase + FR_CMD_OFFSET);
  int		retry	= SDLA_MAX_RETRY;
  int		frerr;

  do
  {
    if (!set_lock(&pds_p->mb_lock))
      frerr = FRRES_MB_INUSE
    ;
    else
    {
      memset(cmd_p, 0, sizeof(frCmd_t));
      cmd_p->cmd     = FR_READ_ERROR_STATS;
      frerr = sdla_exec(pds_p->ads, FR_CMD_OFFSET) ?
              cmd_p->result : FRRES_TIMEOUT
      ;
      if (frerr == FRRES_OK)
      {
        memcpy(stat_p, cmd_p->data, sizeof(frCommStat_t));
      }
      --pds_p->mb_lock;
    }
  }
  while (frerr && fr_event(pds_p, frerr, FR_READ_ERROR_STATS) && --retry);
  return frerr;
} /* fr_comm_stat() */

/*============================================================================
 * Handle Frame Relay asyncronous event and common errors.
 * ---------------------------------------------------------------------------
 * Return zero if previous command has to be cancelled.
 */
static int fr_event (Port_t *pds_p, int event, int cmd)
{
  int	retry	= 0;
  int	port	= pds_p->cfg.port;

  switch (event)
  {
    case FRRES_LINK_DOWN:
      printk("%s: Physical link down on port %d!\n", DRV_NAME, port);
      switch (cmd)
      {
        case FR_WRITE:
        case FR_READ:
          break;

        default:
          retry = 1;
      }
      break;

    case FRRES_CHANNEL_DOWN:
      printk("%s: Link becomes inoperative on port %d.\n", DRV_NAME, port);
      retry = 1;
      break;

    case FRRES_CHANNEL_UP:
      printk("%s: Link becomes operative on port %d.\n", DRV_NAME, port);
      retry = 1;
      break;

    case FRRES_DLCI_CHANGE:
      printk("%s: DLCI status change on port %d.\n", DRV_NAME, port);
      retry = 1;
      break;

    case FRRES_DLCI_MISMATCH:
      printk("%s: DLCI mismatch on port %d.\n", DRV_NAME, port);
      retry = 1;
      break;

    default:
      printk(
	"%s: Command 0x%02X returned 0x%02X on port %d.\n",
	DRV_NAME, cmd, event, port)
      ;
  }
  return retry;
} /* fr_event() */

/****** End *****************************************************************/

