/*****************************************************************************
* PROJECT:	Sangoma Synchronous Data Link Adapter Driver
* FILE:		sdla.c		Main Module
* TARGET OS:	Linux
* AUTHOR:	Gene Kozin
* ----------------------------------------------------------------------------
* Copyright (c) 1995-1996 Sangoma Technologies Inc.  All Rights Reserved.
* ============================================================================
* Rev.#	Date		Who	Details
* -----	-----------	---	----------------------------------------------
*  4.00	Nov.11,1996	GK	o Changed sdlaData, sdlaSetup, sdlaStatus and
*				  sdlaMod data structures.
*
*  3.01	Jul.12,1996	GK	o Changes for Linux 2.0 compatibility.
*
*  3.00	Jun.12,1996	GK	o Implemented stackable module interface.
*				o Major changes in kernel APIs.
*				o Added support for S503 card.
*				o Uses new sdlactl hardware control routines.
*
*  2.00	Apr.30,1996	GK	o SDLA hardware interrupt acknowledgement is
*				  performed before calling protocol driver's
*				  ISR.
*				o Removed dependency on _IOxx macros for
*				  portability reasons.
*				o Fail hardware initialization if memory test
*				  returns zero.
*				o Added auto memory address selection.
*				o Register I/O ports with Linux kernel.
*				o Miscellaneous bug fixes.
*				
*  1.00	Dec.20,1995	GK	Beta version.
*				o Binding routine ignores firmware ID leaving
*				  it to the binder to verify.
*				o Zero adapter memory after initialization.
*				o Added kernel versions of SDLA_PEEK &
*				  SDLA_POKE IOCTLs.
*				o Fixed a bug in interrupt routine (premature
*				  brake from the for loop).
*  0.03	Nov.13,1995	GK	o Autodetect routine now verifies I/O port
*				  before performing autodetection.
*				o Minor Setup/Shutdown IOCTL changes.
*  0.02	Oct.24,1995	GK	Fixed adapter type autodetection.
*  0.01	Oct.14,1995	GK	Alpha version.
* ============================================================================
* Notes:
*   o Compile with -O2 -fomit-frame-pointer. gcc only!
*****************************************************************************/

#include <linux/config.h>	/* OS configuration options */
#include <linux/kernel.h>
#include <linux/version.h>	/* for UTS_RELEASE */
#include <linux/module.h>	/* support for loadable modules */
#include <linux/fs.h>		/* support for device drivers */
#include <linux/mm.h>		/* verify_area(), etc. */
#include <linux/malloc.h>	/* kmalloc(), kfree() */
#include <linux/ioport.h>	/* request_region(), release_region() */
#include <linux/errno.h>	/* return codes */

#include <asm/string.h>		/* inline memset(), etc. */
#include <asm/segment.h>	/* kernel <-> user copy */
#include "sdladrv.h"		/* generic SDLA interface */

/****** Defines *************************************************************/

#define	DRV_NAME	"sdla"
#define	SDLA_RELEASE	0		/* driver minor version number */
#define	DRV_MAJOR_DEV	17		/* major device number */

/****** Data Types **********************************************************/

typedef int Handler (sdlaData_t* ads, sdlaUsr_t* usr);

typedef struct LookUp		/* Look-Up Table Entry */
{
	int code;		/* look-up code */
	void* ptr;		/* pointer */
} LookUp_t;

/****** Function Prototypes *************************************************/

/*
 * Module Wrappers.
 * These are called by insmod & rmmod and must be declared as public.
 */
int init_module (void);
void cleanup_module (void);

/*
 * Driver Entry Points.
 */
static int sdla_open (struct inode* inode_p, struct file* file_p);
static void sdla_close (struct inode* inode_p, struct file* file_p);
static int sdla_read (struct inode* inode_p, struct file* file_p,
			 char* buf, int cnt);
static int sdla_write (struct inode* inode_p, struct file* file_p,
			 const char* buf, int cnt);
static int sdla_ioctl (struct inode* inode_p, struct file* file_p,
			 unsigned int cmd, unsigned long arg);
#if (_OS_VER_ < 200)
static void sdla_intr (int irq, struct pt_regs* regs);
#else
static void sdla_intr (int irq, void* dev_id, struct pt_regs* regs);
#endif

/*
 * IOCTL Handlers.
 */
static int sdlaio_dummy (sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_setup (sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_start (sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_stop (sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_down (sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_bind (sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_unbind (sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_status (sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_mapmem (sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_read (sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_write	(sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_peek (sdlaData_t* ads, sdlaUsr_t* usr_p);
static int sdlaio_poke (sdlaData_t* ads, sdlaUsr_t* usr_p);

/*
 * Miscellaneous Functions.
 */
static void* lookup (int code, LookUp_t* table);
static int sdla_setup (sdlaData_t* ads);
static int sdla_setdpm (sdlaData_t* ads);
static int sdla_shutdown (sdlaData_t* ads);
static int verify_opt (unsigned* optlist, unsigned optval);
static unsigned	memtest (void* ptr, unsigned size);

/****** Global Data *********************************************************/

#if (_OS_VER_ < 200)
char kernel_version[] = UTS_RELEASE;	/* must be public */
#endif

int SDLA_DEBUG = _DEBUG_;	/* debug level */
static char drv_name[] = DRV_NAME;
static char fullname[] = "SDLA Driver";
static char copyright[] = "(c) 1995-1996 Sangoma Technologies Inc.";
static sdlaMod_t* modlist = NULL;		/* registered modules list */
static sdlaData_t sdla[SDLA_MAX_UNITS];		/* array of ADS's */
static unsigned	sdla_delay;			/* 20 mks loop count */

static struct file_operations fops =	/* driver entry points */
{
	NULL,				/* seek */
	sdla_read,			/* read */
	sdla_write,			/* write */
	NULL,				/* readdir */
	NULL,				/* select */
	sdla_ioctl,			/* ioctl */
	NULL,				/* mmap */
	sdla_open,			/* open */
	sdla_close,			/* release */
	NULL,				/* fsync */
	NULL,				/* check_media_change */
	NULL				/* revalidate */
};

static LookUp_t	ioctl_table[] =		/* IOCTL Handler Look-up Table */
{
	{ SDLAIO_MAPMEM,(void*)sdlaio_mapmem	},
	{ SDLAIO_READ,	(void*)sdlaio_read	},
	{ SDLAIO_WRITE,	(void*)sdlaio_write	},
	{ SDLAIO_PEEK,	(void*)sdlaio_peek	},
	{ SDLAIO_POKE,	(void*)sdlaio_poke	},
	{ SDLAIO_STATUS,(void*)sdlaio_status	},
	{ SDLAIO_SETUP,	(void*)sdlaio_setup	},
	{ SDLAIO_START,	(void*)sdlaio_start	},
	{ SDLAIO_STOP,	(void*)sdlaio_stop	},
	{ SDLAIO_DOWN,	(void*)sdlaio_down	},
	{ SDLAIO_BIND,	(void*)sdlaio_bind	},
	{ SDLAIO_UNBIND,(void*)sdlaio_unbind	},
	{ 0,		(void*)sdlaio_dummy	},	/* End of table */
};

/****** Module Wrappers *****************************************************/

/*============================================================================
 * Module Install Entry Point.
 *	Return 0 if O.k. to install.
 */
int init_module	(void)
{
	int stop_tick, i;

	printk(KERN_INFO "%s v%u.%02u %s\n",
		fullname, SDLA_VERSION, SDLA_RELEASE, copyright)
	;

	/*
	 * Initialize static data.
	 */
	memclr(sdla, sizeof(sdla));
	for (i = 0; i < SDLA_MAX_UNITS; ++i)
	{
		sdla[i].cardno = i;
	}

	/*
	 * Calibrate SDLA memory access delay.  Calculate the number of idle
	 * loops made within 1 second interval and then calculate number of
	 * loops that should be made to achive desired delay.
	 */
	for (sdla_delay = 0, stop_tick = jiffies + HZ;
	     jiffies < stop_tick;
	     ++sdla_delay)
	;
	sdla_delay = sdla_delay / 50000 + 1;
	if (SDLA_DEBUG > 2) printk(
		KERN_DEBUG "%s: sdla_delay = %d\n", drv_name, sdla_delay)
	;

	/*
	 * Register device driver.
	 */
	return register_chrdev(DRV_MAJOR_DEV, drv_name, &fops);
}

/*============================================================================
 * Module Remove Entry Point.
 */
void cleanup_module (void)
{
	int i;

	/*
	 * Shutdown adapters.
	 */
	for (i = 0; i < SDLA_MAX_UNITS; ++i)
	{
		if (sdla[i].state > SDLA_DISABLED) sdla_shutdown(&sdla[i]);
	}
	unregister_chrdev(DRV_MAJOR_DEV, drv_name);
}

/******* SDLA Driver Entry Points *******************************************/

/*============================================================================
 * Driver Open Entry Point.
 *	Return 0 if O.k.
 */
static int sdla_open (struct inode* inode_p, struct file* file_p)
{
	int card = MINOR(inode_p->i_rdev);

	if (card >= SDLA_MAX_UNITS)
		return -ENODEV		/* no such device */
	;
	sdla[card].flags |= FLAG_OPEN;
	return 0;
}

/*============================================================================
 * Driver Close Entry Point.
 */
static void sdla_close (struct inode* inode_p, struct file* file_p)
{
	int card = MINOR(inode_p->i_rdev);

	if (card < SDLA_MAX_UNITS)
	{
		sdla[card].flags &= ~FLAG_OPEN;
	}
}

/*============================================================================
 * Driver READ Entry Point.
 */
int sdla_read (struct inode* inode_p, struct file* file_p, char* buf, int cnt)
{
	sdlaData_t* ads = &sdla[MINOR(inode_p->i_rdev)];

	if ((ads->state < SDLA_BOUND) || (ads->read == NULL))
		return -ENXIO
	;
	return ads->read(ads, file_p, buf, cnt);
}

/*============================================================================
 * Driver WRITE Entry Point.
 */
int sdla_write (
	struct inode*	inode_p,
	struct file*	file_p,
	const char*	buf,
	int		cnt)
{
	sdlaData_t* ads = &sdla[MINOR(inode_p->i_rdev)];

	if ((ads->state < SDLA_BOUND) || (ads->write == NULL))
		return -ENXIO
	;
	return ads->write(ads, file_p, buf, cnt);
}

/*============================================================================
 * Driver IOCTL Entry Point.
 *	o Verify arguments, signature and version.
 *	o Check IOCTL 'magic' number. If it's SDLA_MAGIC then call appropriate
 *	  IOCTL handler.
 *	o Otherwise if adapter is bound to protocol-specific module, call
 *	  module's IOCTL entry point.
 *	Return 0 if O.k.
 *	Note:	if IOCTL handler returns EIO then 'ioerr' field of ADS will
 *		contain SDLA error code, which will be copied into 'result'
 *		field of sdlaCtl_t structure.
 */
static int sdla_ioctl (struct inode* inode_p, struct file* file_p,
			unsigned int cmd, unsigned long arg)
{
#define	ioctl_p	((sdlaCtl_t*)arg)

	int err = 0;			/* system return code (errno) */
	int card = MINOR(inode_p->i_rdev);
	sdlaData_t* ads;

	if (SDLA_DEBUG > 2) printk(
		KERN_DEBUG "%s: sdla_ioctl(): cmd = %d\n", drv_name, cmd)
	;

	/*
	 * Verify IOCTL request arguments.
	 */
	if (card >= SDLA_MAX_UNITS)
		return -ENODEV		/* no such device */
	;
	if ((ioctl_p == NULL) ||
	    (verify_area(VERIFY_WRITE, ioctl_p, sizeof(sdlaCtl_t))))
		return -EFAULT		/* invalid pointer */
	;
#if (_OS_VER_ < 200)
	if (get_user_long((void *)&ioctl_p->signature) != SDLA_SIGNATURE)
#else
	if (get_user(&ioctl_p->signature) != SDLA_SIGNATURE)
#endif
		return -EINVAL		/* invalid signature */
	;

	ads = &sdla[card];
	ads->cardno = card;
	ads->ioerr  = 0;		/* clear IOCTL error code */
#if (_OS_VER_ < 200)
	if (get_user_long(&ioctl_p->version) != SDLA_VERSION)
#else
	if (get_user(&ioctl_p->version) != SDLA_VERSION)
#endif
		ads->ioerr = SDLAER_BADVER	/* incorrect driver version */
	;
	else if ((cmd >> 8) == SDLA_MAGIC)
		/*
		 * This is 'our' IOCTL.  Call appropriate handler.
		 */
		err = ((Handler*)lookup(cmd, ioctl_table))(ads, &ioctl_p->u)
	;
	else if ((ads->state == SDLA_BOUND) && ads->ioctl)
		/*
		 * Call protocol-specific module's IOCTL entry point.
		 */
		err = ads->ioctl(ads, cmd, &ioctl_p->u)
	;
	else return -EINVAL;

	/*
	 * If I/O error has occured, pass extended error code to the user.
	 */
	if (err == -EIO)
	{
		if (SDLA_DEBUG > 1) printk(
			KERN_ERR "%s: sdla%u ioctl() %u I/O error %d.\n",
			drv_name, card, cmd, ads->ioerr)
		;
#if (_OS_VER_ < 200)
		put_user_long(ads->ioerr, &ioctl_p->result);
#else
		put_user(ads->ioerr, &ioctl_p->result);
#endif
	}
	return err;

#undef	ioctl_p
}

/*============================================================================
 * SDLA Interrupt Service Routine.
 *	o Find adapter hooked up to this IRQ line.
 *	o Acknowledge SDLA hardware interrupt.
 *	o Call protocol driver's interrupt service routine, if any.
 */
#if (_OS_VER_ < 200)
static void sdla_intr (int irq, struct pt_regs *regs)
#else
static void sdla_intr (int irq, void* dev_id, struct pt_regs *regs)
#endif
{
	int i;
	sdlaData_t* ads;

	for (i = 0, ads = sdla; i < SDLA_MAX_UNITS; ++i, ++ads)
	{
		/*
		 * Find adapter hooked up to this IRQ line.
		 */
		if ((ads->state >= SDLA_CONFIGURED) && (ads->irq == irq))
		{
			++ads->int_count;	/* update interrupt count */

			/*
			 * Acknowledge hardware interrupt.
			 */
			sdlactl.int_ack(ads);

			/*
			 * If adapter is bound then call protocol driver's
			 * interrupt service routine, if any.
			 */
			if ((ads->state == SDLA_BOUND) &&
			    (ads->intr != NULL))
				ads->intr(ads)
			;
			break;
		}
	}
}

/****** Kernel Interface ****************************************************/

/*============================================================================
 * Register Protocol Module.
 */
int sdla_register (unsigned long sig, unsigned ver, sdlaMod_t *mod)
{
	/*
	 * Verify signature, API version, etc.
	 */
	if ((mod == NULL) || (sig != SDLA_SIGNATURE))
		return SDLAER_BADARG
	;
	if (ver != SDLA_VERSION)
		return SDLAER_BADVER
	;

	/*
	 * Link module to the linked list.
	 */
	mod->next = modlist;
	modlist = mod;
	if (SDLA_DEBUG) printk(
		KERN_INFO "%s: registering module %s\n", drv_name, mod->name)
	;
	return 0;
}

/*============================================================================
 * Unregister Protocol Module.
 */
int sdla_unregister (char *name)
{
	sdlaMod_t* cur_p;
	sdlaMod_t* prev_p;

	/*
	 * Verify arguments.
	 */
	if (name == NULL)
		return SDLAER_BADARG
	;

	/*
	 * Unlink module.
	 */

	for (cur_p = modlist, prev_p = NULL;
	     cur_p;
	     prev_p = cur_p, cur_p = cur_p->next)
	{
		if (strcmp(cur_p->name, name) == 0)
		{
			if (SDLA_DEBUG) printk(
				KERN_INFO "%s: unregistering module %s\n",
				drv_name, name)
			;
			if (prev_p) prev_p->next = cur_p->next;
			else modlist = cur_p->next;
			return 0;
		}
	}
	return SDLAER_NOTFOUND;
}

/*============================================================================
 * Execute Adapter Command.
 *	o Set exec flag.
 *	o Busy-wait until flag is reset.
 *	o Return number of loops made, or 0 if command timed out.
 */
int sdla_exec (sdlaData_t* ads, int offs)
{
	volatile unsigned char* flag_p = (char*)ads->vbase + offs;
	int stop_tick;
	int nloop, delay;

	if (*flag_p == 0)
	{
		*flag_p = 1;
		for (stop_tick = SYS_TIMER + SDLA_TIMEOUT, nloop = 1;
		     SYS_TIMER < stop_tick;
        	     ++nloop)
		{
			for (delay = sdla_delay; delay; --delay); /* delay */
			if (*flag_p == 0) return nloop;
		}
	}
	return 0;
}

/****** IOCTL Handlers ******************************************************/

/*============================================================================
 * Dummy IOCTL Handler.
 */
static int sdlaio_dummy (sdlaData_t *ads, sdlaUsr_t *usr_p)
{
	ads->ioerr = SDLAER_BADCMD;
	return -EIO;
}

/*============================================================================
 * SDLAIO_SETUP Handler.
 *	Set up adapter:
 *	 o autodetect adapter type (if not specified)
 *	 o verify hardware options
 *	 o check for hardware conflicts
 *	 o set up adapter hardware.
 *	Initial adapter state:	SDLA_DISABLED
 *	Resulting adapter state:	SDLA_CONFIGURED
 *	On error return:
 *		SDLAER_BADSTATE	- adapter is not in DISABLED state
 *		SDLAER_BADHWOPT	- illegal hardware option
 *		SDLAER_CONFLICT	- hardware conflict
 *		SDLAER_BADTYPE	- unknown adapter type
 *		SDLAER_MEMCONFL	- memory area is unusable for DPM window
 *		SDLAER_FAILURE	- adapter hardware failure
 */
static int sdlaio_setup (sdlaData_t* ads, sdlaUsr_t* usr_p)
{
	int err = 0;

	/*
	 * Verify adapter state.
	 */
	if (ads->state > SDLA_DISABLED)
		ads->ioerr = SDLAER_BADSTATE	/* already configured */
	;
	else
	{
		/*
		 * Get setup data from user address space and perform setup.
		 */
		memcpy_fromfs((void*)ads, (void*)usr_p, sizeof(sdlaSetup_t));
		ads->ioerr = sdla_setup(ads);

		if (ads->ioerr)
			memclr(ads, sizeof(sdlaSetup_t))
		;
	}
	return (ads->ioerr) ? -EIO : err;
} 

/*============================================================================
 * SDLAIO_START Handler.
 *	Fire up adapter:
 *	 o position DPM window over adapter's cold boot entry point.
 *	 o put boot code at adapter's cold boot entry point.
 *	 o start CPU, set default DPM window vector and enable h/w interrupts.
 *	Initial state:	SDLA_CONFIGURED
 *	Resulting state:	SDLA_RUNNING
 *	On error, return:
 *		SDLAER_BADSTATE	- adapter is not in SDLA_CONFIGURED state
 *		SDLAER_FAILURE	- adapter hardware failure
 */
static int sdlaio_start (sdlaData_t *ads, sdlaUsr_t *usr_p)
{
  int	err	= 0;

  /*
   * Verify adapter state.
   */
  if (ads->state != SDLA_CONFIGURED)
    ads->ioerr = SDLAER_BADSTATE
  ;
  else
  {
    sdlaStart_t	start;		/* firmware start info */
    sdlaInfo_t	info;		/* adapter hardware options */
    int		offs, boot;

    /*
     * Get firmware start data from user address space.
     */
    memcpy_fromfs((void *)&start, (void *)usr_p, sizeof(sdlaStart_t));

    /*
     * Position DPM window over adapter's cold boot entry point.
     */
    sdlactl.get_info(ads, &info);
    offs = info.boot_offs % ads->dpm_size;
    if (sdlactl.mapmem(ads, info.boot_offs - offs))
      ads->ioerr = SDLAER_FAILURE
    ;
    else
    { 
      /*
       * Put boot code at adapter's cold boot entry point.
       */
      boot = ads->vbase;
      *((char *)boot) = 0xC3;			/* Z80: jp */
      *((short *)(boot + 1)) = start.entry_point;

      /*
       * Start CPU, set default DPM window vector and enable h/w interrupts.
       */
      cli();
      if (sdlactl.cpu_start(ads) ||
          sdlactl.mapmem(ads, start.dpm_vector) ||
          (ads->irq && sdlactl.int_enable(ads)))
        ads->ioerr = SDLAER_FAILURE
      ;
      else
      {
        ads->firmware_id = start.firmware_id;
        ads->timestamp   = jiffies;
        ads->state       = SDLA_RUNNING;
      }
      sti();
    }
  }
  return (ads->ioerr) ? -EIO : err;
} /* sdlaio_start() */

/*============================================================================
 * SDLAIO_STOP Handler.
 *	Place adapter into configured, inactive state (i.e. for diagnostic).
 *	Initial state:	SDLA_RUNNING
 *	Resulting state:	SDLA_CONFIGURED
 *	On error, return:
 *		SDLAER_BADSTATE	- adapter is not in SDLA_RUNNING state
 *		SDLAER_FAILURE	- adapter hardware failure
 */
static int sdlaio_stop (sdlaData_t *ads, sdlaUsr_t *usr_p)
{
	int err = 0;

	/*
	 * Verify adapter state.
	 */
	if (ads->state != SDLA_RUNNING)
		ads->ioerr = SDLAER_BADSTATE
	;
	else
	{
		cli();
		if (sdlactl.cpu_stop(ads))
			ads->ioerr = SDLAER_FAILURE
		;
		else
		{
			ads->timestamp = jiffies;
			ads->state = SDLA_CONFIGURED;
		}
		sti();
	}
	return (ads->ioerr) ? -EIO : err;
}

/*============================================================================
 * SDLAIO_DOWN Handler.
 *	Shut down adapter.
 *	Initial state:	any, except for SDLA_BOUND
 *	Resulting state:	SDLA_DISABLED
 *	On error, return:
 *		SDLAER_BADSTATE	- adapter is in SDLA_BOUND state
 *		SDLAER_FAILURE	- adapter hardware failure
 */
static int sdlaio_down (sdlaData_t* ads, sdlaUsr_t* usr_p)
{
	/*
	 * Verify adapter state.
	 */
	if (ads->state >= SDLA_BOUND)
		ads->ioerr = SDLAER_BADSTATE
	;
	else if (ads->state > SDLA_DISABLED)
		sdla_shutdown(ads)
	;
	return (ads->ioerr) ? -EIO : 0;
}

/*============================================================================
 * SDLAIO_BIND Handler.
 *	Bind adapter to protocol-specific module.
 *	Initial state:	SDLA_RUNNING
 *	Resulting state:	SDLA_BOUND
 *	On error, return:
 *		SDLAER_DOWN	- adapter is not running
 *		SDLAER_BADSTATE	- adapter is already bound
 */
static int sdlaio_bind (sdlaData_t* ads, sdlaUsr_t* usr_p)
{
  int err = 0;

  /*
   * Verify adapter state.
   */
  if (ads->state < SDLA_RUNNING)
    ads->ioerr = SDLAER_DOWN
  ;
  else if (ads->state >= SDLA_BOUND)
    ads->ioerr = SDLAER_BADSTATE
  ;
  else
  {
    sdlaMod_t* mod;

    /*
     * Get module name from user address space.
     */
    memcpy_fromfs((void*)ads->modname, (void*)usr_p, SDLA_MODNAME_SZ);
    ads->modname[SDLA_MODNAME_SZ - 1] = '\0';	/* better safe than sorry */

    /*
     * Search registered modules for the module with matching name.
     */
    for (mod = modlist; mod; mod = mod->next)
    {
      if (strcmp(ads->modname, mod->name) == 0) break;
    }

    if (mod == NULL)
      ads->ioerr = SDLAER_NOTFOUND
    ;
    else
    {
      ads->read = NULL;
      ads->write = NULL;
      ads->ioctl = NULL;
      ads->intr = NULL;
      ads->halt = NULL;
      ads->unbind = NULL;
      if (mod->bind) ads->ioerr = mod->bind(ads);
    }

    if (!ads->ioerr)
    {
      ads->state     = SDLA_BOUND;
      ads->timestamp = jiffies;
    }
  }
  return (ads->ioerr) ? -EIO : err;
} /* sdlaio_bind() */

/*============================================================================
 * SDLAIO_UNBIND Handler.
 *	Unbind adapter from protocol-specific module.
 *	Initial state:	SDLA_BOUND
 *	Resulting state:	SDLA_RUNNING
 *	On error, return:
 *		SDLAER_BADSTATE	- adapter is not bound
 */
static int sdlaio_unbind (sdlaData_t* ads, sdlaUsr_t* usr_p)
{
	/*
	 * Verify adapter state.
	 */
	if (ads->state < SDLA_RUNNING)
		ads->ioerr = SDLAER_BADSTATE
	;
	else
	{
		if (ads->unbind) ads->ioerr = ads->unbind(ads);
		ads->state = SDLA_RUNNING;
		ads->timestamp = jiffies;
	}
	return (ads->ioerr) ? -EIO : 0;
}

/*============================================================================
 * SDLAIO_STATUS Handler.
 *	Read adapter status.
 */
static int sdlaio_status (sdlaData_t* ads, sdlaUsr_t* usr_p)
{
	int i;

	if (ads->state >= SDLA_CONFIGURED)
		for (i = 0; i < ads->io_range; ++i)
			ads->regs_rd[i] = inb(ads->io_base + i)
	;
	memcpy_tofs((void*)usr_p, ads, sizeof(sdlaStatus_t));
	return 0;
}

/*============================================================================
 * SDLAIO_MAPMEM Handler.
 */
static int sdlaio_mapmem (sdlaData_t* ads, sdlaUsr_t* usr_p)
{
  int	err	= 0;

  /*
   * Verify adapter state.
   */
  if ((ads->state < SDLA_CONFIGURED) || (ads->state >= SDLA_BOUND))
    ads->ioerr = SDLAER_BADSTATE
  ;
  {
    unsigned long vector;

    /*
     * Get new DPM window vector from user address space.
     */
#if (_OS_VER_ < 200)
    vector = get_user_long((void*)usr_p);
#else
    vector = get_user(&usr_p->vector);
#endif

    if (vector > ads->mem_size)
      ads->ioerr = SDLAER_BADARG
    ;
    else
    {
      cli();
      if (sdlactl.mapmem(ads, vector))
        ads->ioerr = SDLAER_FAILURE
      ;
      sti();
    }
  }
  return (ads->ioerr) ? -EIO : err;
} /* sdlaio_mapmem() */

/*============================================================================
 * SDLAIO_READ Handler.
 *	Read block of data from adapter's shared memory window.
 */
static int sdlaio_read (sdlaData_t* ads, sdlaUsr_t* usr_p)
{
  int	err	= 0;

  /*
   * Verify adapter state.
   */
  if (ads->state < SDLA_CONFIGURED)
    ads->ioerr = SDLAER_BADSTATE
  ;
  else
  {
    sdlaXfer_t	xfer;

    /*
     * Get transfer control block from user address space.
     */
    memcpy_fromfs((void*)&xfer, (void*)usr_p, sizeof(xfer));

    /*
     * Verify arguments.
     */
    if (xfer.offset + xfer.length > ads->dpm_size)
      ads->ioerr = SDLAER_BADARG
    ;
    else memcpy_tofs(
		(void*)(xfer.ptr),
		(void*)(ads->vbase + xfer.offset),
		xfer.length)
    ;
  }
  return (ads->ioerr) ? -EIO : err;
} /* sdlaio_read() */

/*============================================================================
 * SDLAIO_WRITE Handler.
 */
static int sdlaio_write (sdlaData_t* ads, sdlaUsr_t* usr_p)
{
  int	err	= 0;

  /*
   * Verify adapter state.
   */
  if (ads->state < SDLA_CONFIGURED)
    ads->ioerr = SDLAER_BADSTATE
  ;
  else
  {
    sdlaXfer_t	xfer;

    /*
     * Get transfer control block from user address space.
     */
    memcpy_fromfs((void*)&xfer, (void*)usr_p, sizeof(xfer));

    /*
     * Verify arguments.
     */
    if (xfer.offset + xfer.length > ads->dpm_size)
      ads->ioerr = SDLAER_BADARG
    ;
    else memcpy_fromfs(
		(void*)(ads->vbase + xfer.offset),
		(void*)(xfer.ptr),
		xfer.length)
    ;
  }
  return (ads->ioerr) ? -EIO : err;
} /* sdlaio_write() */

/*============================================================================
 * SDLAIO_PEEK Handler.
 */
static int sdlaio_peek (sdlaData_t* ads, sdlaUsr_t* usr_p)
{
  int		err	= 0;
  sdlaXfer_t    xfer;

  if (ads->state < SDLA_CONFIGURED)	/* verify adapter state */
    ads->ioerr = SDLAER_BADSTATE
  ;
  else
  {
    /*
     * Get transfer control block from user address space.
     */
    memcpy_fromfs((void*)&xfer, (void*)usr_p, sizeof(xfer));

    /*
     * Verify arguments.
     */
    if (xfer.offset + xfer.length > ads->mem_size)
      ads->ioerr = SDLAER_BADARG
    ;
    else
    {
      unsigned      winsize = ads->dpm_size;
      unsigned      pos, len;		/* current offset and block size */
      unsigned long vec, oldvec;	/* DPM window vector */

      /*
       * Transfer data from adapter's memory to user data buffer.
       * Note:	it's possible that block length is bigger than DPM window
       *	size, so we must take care of moving DPM window around when
       *	nesessary.
       */
      oldvec = ads->mem_vector;
      while (xfer.length)
      {
        pos = xfer.offset % winsize;		/* current window offset */
        vec = xfer.offset - pos;		/* current window vector */
        len = (xfer.length > (winsize - pos)) ? (winsize - pos) : xfer.length;
        cli();
        sdlactl.mapmem(ads, vec);		/* relocate window */
        sti();
        memcpy_tofs(
		(void*)(xfer.ptr),
		(void*)(ads->vbase + pos),
		len)
        ;
        xfer.length     -= len;
        xfer.offset     += len;
        (char*)xfer.ptr += len;
      }

      /*
       * Restore DPM window position.
       */
      cli();
      sdlactl.mapmem(ads, oldvec);
      sti();
    }
  }
  return (ads->ioerr) ? -EIO : err;
} /* sdlaio_peek() */

/*============================================================================
 * SDLAIO_POKE Handler.
 */
static int sdlaio_poke (sdlaData_t* ads, sdlaUsr_t* usr_p)
{
  int		err	= 0;
  sdlaXfer_t    xfer;

  if (ads->state < SDLA_CONFIGURED)	/* verify adapter state */
    ads->ioerr = SDLAER_BADSTATE
  ;
  else
  {
    /*
     * Get transfer control block from user address space.
     */
    memcpy_fromfs((void*)&xfer, (void*)usr_p, sizeof(xfer));

    /*
     * Verify arguments.
     */
    if (xfer.offset + xfer.length > ads->mem_size)
      ads->ioerr = SDLAER_BADARG
    ;
    else
    {
      unsigned      winsize = ads->dpm_size;
      unsigned      pos, len;		/* current offset and block size */
      unsigned long vec, oldvec;	/* DPM window vector */

      /*
       * Transfer data from user data buffer to adapter's memory.
       * Note:	it's possible that block length is bigger than DPM window
       *	size, so we must take care of moving DPM window around when
       *	nesessary.
       */
      oldvec = ads->mem_vector;
      while (xfer.length)
      {
        pos = xfer.offset % winsize;		/* current window offset */
        vec = xfer.offset - pos;		/* current window vector */
        len = (xfer.length > (winsize - pos)) ? (winsize - pos) : xfer.length;
        cli();
        sdlactl.mapmem(ads, vec);		/* relocate window */
        sti();
        memcpy_fromfs(
		(void*)(ads->vbase + pos),
		(void*)(xfer.ptr),
		len)
        ;
        xfer.length     -= len;
        xfer.offset     += len;
        (char*)xfer.ptr += len;
      }

      /*
       * Restore DPM window position.
       */
      cli();
      sdlactl.mapmem(ads, oldvec);
      sti();
    }
  }
  return (ads->ioerr) ? -EIO : err;
} /* sdlaio_poke() */

/****** Support Functions ***************************************************/

/*============================================================================
 * Set up SDLA hardware.
 */
static int sdla_setup (sdlaData_t* ads)
{
	sdlaInfo_t info;		/* adapter hardware options */
	int i;

	/*
	 * Verify I/O port address and detect adapter.
	 */
	if (ads->io_base == 0)
		return SDLAER_BADHWOPT	/* I/O port must be specified */
	;
	if (check_region(ads->io_base, SDLA_MAX_RANGE))
	{
		printk(KERN_ERR "%s: I/O port conflict @ 0x%X!\n",
			drv_name, ads->io_base)
		;
		return SDLAER_CONFLICT;	/* I/O ports conflict */
	}
	if (sdlactl.detect(ads))
	{
		printk(KERN_ERR "%s: Adapter not detected @ 0x%X!\n",
			drv_name, ads->io_base)
		;
		return SDLAER_BADTYPE;	/* unknown SDLA type */
	}

	/*
	 * Verify adapter hardware options and set defaults.
	 */
	sdlactl.get_info(ads, &info);
        ads->io_range = info.io_range;

	if (ads->clock_rate == 0) ads->clock_rate = info.clock[1];
	else if (!verify_opt(info.clock, ads->clock_rate))
	{
		printk(KERN_ERR "%s: Invalid CPU clock option (%d)!\n",
			drv_name, ads->clock_rate)
		;
		return SDLAER_BADHWOPT;
	}

	if (ads->dpm_size == 0) ads->dpm_size = info.dpm_size[1];
	else if (!verify_opt(info.dpm_size, ads->dpm_size))
	{
		printk(KERN_ERR "%s: Invalid memory window size option (0x%X)!\n",
			drv_name, ads->dpm_size)
		;
		return SDLAER_BADHWOPT;
	} 

	if (ads->dpm_base == 0)
	{
		for (i = info.dpm_base[0]; i; --i)
		{
			ads->dpm_base = info.dpm_base[i];
			if (sdla_setdpm(ads) == 0) break;
		}
		if (i == 0)
		{
			printk(KERN_ERR "%s: No memory region available!\n", drv_name);
			return SDLAER_BADHWOPT;
		}
	}
	else if (!verify_opt(info.dpm_base, ads->dpm_base))
	{
		printk(KERN_ERR "%s: Invalid memory window base option (0x%lX)!\n",
			drv_name, ads->dpm_base)
		;
		return SDLAER_MEMCONFL;
	} 
	else if (sdla_setdpm(ads))
	{
		printk(KERN_ERR "%s: Memory conflict @ 0x%lX!\n",
			drv_name, ads->dpm_base)
		;
		return SDLAER_MEMCONFL;
	} 

	if (ads->irq)
	{
		int err = 0;
		int irq = (ads->irq == 2) ? 9 : ads->irq;

		if (!verify_opt(info.irq, ads->irq))
		{
			printk(KERN_ERR "%s: Invalid IRQ option (%d)!\n",
				drv_name, ads->irq)
			;
			err = SDLAER_BADHWOPT;
		} 
		else if (
#if (_OS_VER_ < 200)
			request_irq(irq, sdla_intr, 0, drv_name)
#else
			request_irq(irq, sdla_intr, 0, drv_name, NULL)
#endif
			)
		{
			printk(KERN_ERR "%s: IRQ %d conflict!\n",
				drv_name, ads->irq)
			;
			err = SDLAER_CONFLICT;
		}
		else ads->irq = irq;

		if (err)
		{
  			sdlactl.down(ads);	/* shutdown adapter */
			return err;
		}
	}

	/*
	 * Finish initialization.
	 */
	request_region(ads->io_base, ads->io_range, drv_name);
	sdlactl.mapmem(ads, 0);
	ads->timestamp = jiffies;
	ads->state = SDLA_CONFIGURED;
	return 0;
}

/*============================================================================
 * Set up DPM window. 
 */
static int sdla_setdpm (sdlaData_t *ads)
{
  int		ioerr	= 0;
  unsigned char	*ch_p;		/* DPM window pointer */
  int		i, tmp;

  /*
   * Since Linux does not allow to register memory region let's at least
   * check for memory conflicts with other configured SDLA cards.
   */
  for (i = 0; i < SDLA_MAX_UNITS; ++i)
  {
    if ((&sdla[i] != ads) &&
        (sdla[i].state > SDLA_DISABLED) &&
        (sdla[i].dpm_base == ads->dpm_base))
    {
      ioerr = SDLAER_CONFLICT;
      break;
    }
  } 

  if (!ioerr)
  {
    /*
     * Get virtual memory address.
     * Linux maps kernel address space to physical memory.
     */
    ads->vbase = ads->dpm_base;

    /*
     * Shut adapter down to force memory window to close, then check
     * memory window address space to make sure it is not populated.
     */
    sdlactl.down(ads);
    ch_p = (void *)ads->vbase;
    for (i = 0; i < ads->dpm_size; ++i, ++ch_p)
    {
      if (*ch_p != 0xFF)
      {
        ioerr = SDLAER_MEMCONFL;	/* memory area is in use */
        break;
      }
    }
  }

  if (!ioerr)
  {
    /*
     * Initialize adapter and enable DPM window access.
     */
    if (sdlactl.init(ads))
      ioerr = SDLAER_FAILURE
    ;
    else
    { 
      /*
       * Test adapter memory by mapping DPM window into SDLA local
       * address space and testing it segment by segment.
       */
      ads->mem_size = 0;
      while (!sdlactl.mapmem(ads, ads->mem_size))
      { 
        tmp = memtest(
		(void *)ads->vbase,
		ads->dpm_size)
        ;
        ads->mem_size += tmp;
        if (tmp != ads->dpm_size) break;
      }
      if (ads->mem_size == 0)
      {
        ioerr = SDLAER_MEMCONFL;	/* memory test failed */
        sdlactl.down(ads);
      }
    } 
  }
  return ioerr;
}

/*============================================================================
 * Shut down SDLA Hardware. 
 */
static int sdla_shutdown (sdlaData_t *ads)
{
  /*
   * Shut down adapter hardware.
   */
  cli();
  ads->state = SDLA_DISABLED;
  sdlactl.down(ads);
  sti();

  /*
   * Release system resources.
   */
  if (ads->irq)
#if (_OS_VER_ < 200)
    free_irq(ads->irq)
#else
    free_irq(ads->irq, NULL)
#endif
  ;
  release_region(ads->io_base, ads->io_range);

  /*
   * Clear adapter status structure.
   */
  memclr(ads, sizeof(sdlaStatus_t));
  ads->timestamp = jiffies;
  return 0;
}

/*============================================================================
 * Verify hardware option.
 *	Return option's ordinal number (1..N) if verifies o.k. or zero if
 *	invalid option.
 */
static int verify_opt (unsigned* optlist, unsigned optval)
{
	int i;

	for (i = 1; i <= optlist[0]; ++i)
		if (optlist[i] == optval) return i 
	;
	return 0;
}

/*============================================================================
 * Test memory region.
 *	Return size of region that passed the test.
 *	Note:	Region size must be multiple of 2 !
 */
static unsigned memtest (void* ptr, unsigned size)
{
  unsigned short* w_ptr;
  unsigned size_w = size >> 1;	/* region size in words */
  unsigned i;

	for (i = 0, w_ptr = ptr; i < size_w; ++i, ++w_ptr)
		*w_ptr = 0xAA55
	;
	for (i = 0, w_ptr = ptr; i < size_w; ++i, ++w_ptr)
		if (*w_ptr != 0xAA55)
		{
			size_w = i;
			break;
		}
	;
	for (i = 0, w_ptr = ptr; i < size_w; ++i, ++w_ptr)
		*w_ptr = 0x55AA
	;
	for (i = 0, w_ptr = ptr; i < size_w; ++i, ++w_ptr)
		if (*w_ptr != 0x55AA)
		{
			size_w = i;
			break;
		}
	;
	for (i = 0, w_ptr = ptr; i < size_w; ++i, ++w_ptr) *w_ptr = 0;
	return size_w << 1;
}

/*============================================================================
 * Get a pointer from a look-up table.
 */
static void* lookup (int code, LookUp_t* table)
{
	while (table->code && (table->code != code)) table++;
	return table->ptr;
}

/****** End *****************************************************************/

