/*****************************************************************************
* PROJECT:	Sangoma Synchronous Data Link Adapter Driver
* FILE:		sdlactl.c	SDLA Hardware-independent Control Routines.
* TARGET OS:	SCO Unix, Linux
* AUTHOR:	Gene Kozin, Sangoma Technologies Inc.
* ----------------------------------------------------------------------------
* (c) Copyright 1995-1996 Sangoma Technologies Inc.  All Rights Reserved.
* ============================================================================
* Rev.#	Date		Who	Description
* -----	-----------	---	----------------------------------------------
*  3.10	Oct.17,1996	GK	
*  3.00	Jun.12,1996	GK	'All-in-one' version.  Handles all card types.
*  1.00	Dec.06,1994	GK	Initial version.
*****************************************************************************/

/*============================================================================
 * This module implements low-level hardware independent interface for Sangoma
 * SDLA cards.
 *
 * S502A Notes:
 *  o There is no separate DPM window enable/disable control on S502A.  Window
 *    is open immediately after it's configured through the HMCR register.
 *    To close DPM window, HMCR has to be written a value ????1111b (e.g. 0x0F
 *    or 0xFF).
 *
 *  o On S502A DPM window cannot be located at offset E000 (e.g. 0xAE000).
 *
 *  o There should be a delay of .... before reading back S502A status
 *    register.
 *
 * S502E Notes:
 *  o S502E has a h/w bug: although default IRQ line state is HIGH, enabling
 *    interrupts by setting bit 1 of control register (BASE) to '1'causes it
 *    to go LOW!  Therefore, disabling interrupts by setting that bit to '0'
 *    causes low-to-high transition on IRQ line (ghosty interrupt).  The same
 *    occurs when disabling CPU by resetting bit 0 of CPU control register
 *    (BASE+3) - see the next note.
 *
 *  o S502E CPU and DPM control is limited:
 *	- CPU cannot be stopped independently. Resetting bit 0 of CPU control
 *	  register (BASE+3) shuts the board down entirely;
 *	- DPM access cannot be controlled dynamically. Ones CPU is started,
 *	  bit 1 of control register (BASE) is used to enable/disable IRQ, so
 *	  access to shared memory cannot be disabled while CPU is running.
 ****************************************************************************/

#include "sdladrv.h"		/* common SDLA driver defines */

/****** Defines *************************************************************/

#define	SDLA_IODELAY	100	/* 10 works for 486DX2-66 */

/*
 * I/O port address range.
 */
#define S502A_IORANGE	3
#define S502E_IORANGE	4
#define S503_IORANGE	3
#define S507_IORANGE	4
#define S508_IORANGE	4

/*
 * Maximum amount of memory.
 */
#define S502_MAXMEM	0x10000L
#define S503_MAXMEM	0x10000L
#define S507_MAXMEM	0x40000L
#define S508_MAXMEM	0x40000L

/*
 * Minimum amount of memory.
 */
#define S502_MINMEM	0x8000L
#define S503_MINMEM	0x8000L
#define S507_MINMEM	0x20000L
#define S508_MINMEM	0x20000L

/*
 * Adaptrer cold boot entry point offset.
 */
#define	S502A_BOOTOFS	0x0066
#define	S502E_BOOTOFS	0x0000
#define	S503_BOOTOFS	0x0000
#define	S507_BOOTOFS	0x0000
#define	S508_BOOTOFS	0x0000

/****** Function Prototypes *************************************************/

static int sdla_init	(sdlaData_t* ads_p);
static int sdla_down	(sdlaData_t* ads_p);
static int sdla_mapmem	(sdlaData_t* ads_p, unsigned long addr);
static int sdla_start	(sdlaData_t* ads_p);
static int sdla_stop	(sdlaData_t* ads_p);
static int sdla_inten	(sdlaData_t* ads_p);
static int sdla_intde	(sdlaData_t* ads_p);
static int sdla_intack	(sdlaData_t* ads_p);
static int sdla_isr	(sdlaData_t* ads_p);
static int sdla_getinfo	(sdlaData_t* ads_p, sdlaInfo_t* info_p);
static int sdla_detect	(sdlaData_t* ads_p);

static int init_S502A	(sdlaData_t* ads_p);
static int init_S502E	(sdlaData_t* ads_p);
static int init_S503	(sdlaData_t* ads_p);
static int init_S507	(sdlaData_t* ads_p);
static int init_S508	(sdlaData_t* ads_p);

static int detect_S502A	(int port);
static int detect_S502E	(int port);
static int detect_S503	(int port);
static int detect_S507	(int port);
static int detect_S508	(int port);

static int opt_index	(unsigned* optlist, unsigned optval);

/****** Public Data *********************************************************/

const sdlaHwCtl_t sdlactl =	/* SDLA control entry points */
{
	sdla_getinfo,		/* .get_info() */
	sdla_detect,		/* .detect() */
	sdla_init,		/* .init() */
	sdla_down,		/* .down() */
	sdla_mapmem,		/* .mapmem() */
	sdla_start,		/* .cpu_start() */
	sdla_stop,		/* .cpu_stop() */
	sdla_inten,		/* .int_enable() */
	sdla_intde,		/* .int_disable() */
	sdla_intack,		/* .int_ack() */
	sdla_isr,		/* .isr() */
};

/****** Local Data **********************************************************/

/*
 * Hardware Configuration Options (for verification).
 * First element of an options array is the number of options.
 */
static unsigned	S502_portopt[] =
	{ 4, 0x250, 0x300, 0x350, 0x360 }
;
static unsigned	S503_portopt[] =
	{ 8, 0x250, 0x254, 0x300, 0x304, 0x350, 0x354, 0x360, 0x364 }
;
static unsigned	S508_portopt[] =
	{ 8, 0x250, 0x270, 0x280, 0x300, 0x350, 0x360, 0x380, 0x390 }
;

static unsigned S502A_irqopt[] = { 0 };
static unsigned S502E_irqopt[] = { 4, 2, 3, 5, 7 };
static unsigned S503_irqopt[]  = { 5, 2, 3, 4, 5, 7 };
static unsigned S508_irqopt[]  = { 8, 3, 4, 5, 7, 10, 11, 12, 15 };

static unsigned S502A_memopt[]  =
{
	28,
	0xA0000, 0xA2000, 0xA4000, 0xA6000, 0xA8000, 0xAA000, 0xAC000,
	0xC0000, 0xC2000, 0xC4000, 0xC6000, 0xC8000, 0xCA000, 0xCC000,
	0xD0000, 0xD2000, 0xD4000, 0xD6000, 0xD8000, 0xDA000, 0xDC000,
	0xE0000, 0xE2000, 0xE4000, 0xE6000, 0xE8000, 0xEA000, 0xEC000,
};
static unsigned S502E_memopt[]  =
{
	32,
	0xA0000, 0xA2000, 0xA4000, 0xA6000, 0xA8000, 0xAA000, 0xAC000, 0xAE000,
	0xC0000, 0xC2000, 0xC4000, 0xC6000, 0xC8000, 0xCA000, 0xCC000, 0xCE000,
	0xD0000, 0xD2000, 0xD4000, 0xD6000, 0xD8000, 0xDA000, 0xDC000, 0xDE000,
	0xE0000, 0xE2000, 0xE4000, 0xE6000, 0xE8000, 0xEA000, 0xEC000, 0xEE000,
};
static unsigned S507_memopt[] =
{
	32,
	0xA0000, 0xA2000, 0xA4000, 0xA6000, 0xA8000, 0xAA000, 0xAC000, 0xAE000,
	0xB0000, 0xB2000, 0xB4000, 0xB6000, 0xB8000, 0xBA000, 0xBC000, 0xBE000,
	0xC0000, 0xC2000, 0xC4000, 0xC6000, 0xC8000, 0xCA000, 0xCC000, 0xCE000,
	0xE0000, 0xE2000, 0xE4000, 0xE6000, 0xE8000, 0xEA000, 0xEC000, 0xEE000,
};

static unsigned	S502_winopt[] = { 2, 0x2000, 0x10000 };	/* incl. S503 */
static unsigned	S507_winopt[] = { 2, 0x2000, 0x4000 };
static unsigned	S508_winopt[] = { 1, 0x2000 };

static unsigned	S502A_clkopt[] = { 2, 3600, 7200 };
static unsigned	S502E_clkopt[] = { 5, 3600, 5000, 7200, 8000, 10000 };
static unsigned	S503_clkopt[]  = { 3, 7200, 8000, 10000 };
static unsigned	S507_clkopt[]  = { 1, 12288 };
static unsigned	S508_clkopt[]  = { 1, 16000 };

/*
 * Host memory control register masks.
 */
static unsigned char S502A_hmcr[] =
{
	0x10, 0x12, 0x14, 0x16, 0x18, 0x1A, 0x1C,	/* A0000 - AC000 */
	0x20, 0x22, 0x24, 0x26, 0x28, 0x2A, 0x2C,	/* C0000 - CC000 */
	0x00, 0x02, 0x04, 0x06, 0x08, 0x0A, 0x0C,	/* D0000 - DC000 */
	0x30, 0x32, 0x34, 0x36, 0x38, 0x3A, 0x3C,	/* E0000 - EC000 */
};
static unsigned char S502E_hmcr[] =
{
	0x10, 0x12, 0x14, 0x16, 0x18, 0x1A, 0x1C, 0x1E,	/* A0000 - AE000 */
	0x20, 0x22, 0x24, 0x26, 0x28, 0x2A, 0x2C, 0x2E,	/* C0000 - CE000 */
	0x00, 0x02, 0x04, 0x06, 0x08, 0x0A, 0x0C, 0x0E,	/* D0000 - DE000 */
	0x30, 0x32, 0x34, 0x36, 0x38, 0x3A, 0x3C, 0x3E,	/* E0000 - EE000 */
};
static unsigned char S507_hmcr[] =
{
	0x00, 0x02, 0x04, 0x06, 0x08, 0x0A, 0x0C, 0x0E,	/* A0000 - AE000 */
	0x40, 0x42, 0x44, 0x46, 0x48, 0x4A, 0x4C, 0x4E,	/* B0000 - BE000 */
	0x80, 0x82, 0x84, 0x86, 0x88, 0x8A, 0x8C, 0x8E,	/* C0000 - CE000 */
	0xC0, 0xC2, 0xC4, 0xC6, 0xC8, 0xCA, 0xCC, 0xCE,	/* E0000 - EE000 */
};
static unsigned char S508_hmcr[] =
{
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* A0000 - AE000 */
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,	/* C0000 - CE000 */
	0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,	/* D0000 - DE000 */
	0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,	/* E0000 - EE000 */
};

static unsigned char S507_irqmask[] =
{
	0x00, 0x20, 0x40, 0x60, 0x80, 0xA0, 0xC0, 0xE0
};

/****** Interface Functions *************************************************/

/*============================================================================
 * SDLA Control Routine.
 *	This function performs low-level SDLA hardware control and imposes no
 *	restrictions on requested operations. All intelligence is assumed to
 *	be in higher levels.
 *
 * On error return:
 * 	CTLERR_BADCMD	- illegal command
 * 	CTLERR_BADARG	- argument out of range
 * 	CTLERR_BADSTATE	- incorrect adapter state
 * 	CTLERR_FATAL	- hardware failure
 */
int sdla_ctl (sdlaData_t *ads_p, int cmd, unsigned long lpar)
{
	int result;

	switch (cmd)		/* process command */
	{
	case SDLACTL_MAPMEM:	/* map memory window into SDLA adress space */
		/*
		 * Note: lpar argument contains address of the SDLA memory
		 * location to be visible through the shared memory window.
		 */
		result = sdla_mapmem(ads_p, lpar);
		break;

	case SDLACTL_INTACK:	/* acknowledge SDLA interrupt */
		result = sdla_intack(ads_p);
		break;

	case SDLACTL_INTEN:	/* enable interrupt generation */
		result = sdla_inten(ads_p);
		break;

	case SDLACTL_INTDE:	/* disable interrupt generation */
		result = sdla_intde(ads_p);
		break;

	case SDLACTL_INIT:	/* configure SDLA */
		result = sdla_init(ads_p);
		break;

	case SDLACTL_DOWN:	/* shut down SDLA */
		result = sdla_down(ads_p);
		break;

	case SDLACTL_START:	/* start CPU */
		result = sdla_start(ads_p);
		break;

	case SDLACTL_STOP:	/* stop on-board CPU */
		result = sdla_stop(ads_p);
		break;

	case SDLACTL_GETINFO:	/* get adapter hardware options */
		/*
		 * Note: lpar -> sdlaInfo_t
		 */
		result = sdla_getinfo(ads_p, (void*)lpar);
		break;

	case SDLACTL_DETECT:	/* detect adapter hardware */
		result = sdla_detect(ads_p);
		break;

	case SDLACTL_INTRPT:	/* issue an interrupt to SDLA */
	default:			/* unsupported command */
		result = CTLERR_BADCMD;
	}
	return result;
} /* sdla_ctl() */

/******	LOCAL FUNCTIONS *****************************************************/

/*============================================================================
 * Get SDLA Hardware Options and Model-specific Characteristics.
 */
static int sdla_getinfo (sdlaData_t* ads_p, sdlaInfo_t* info_p)
{
	int err	= 0;

	switch (ads_p->type)
	{
	case SDLA_S502A:
		info_p->hw_opt    = 0;
		info_p->io_range  = S502A_IORANGE;
		info_p->boot_offs = S502A_BOOTOFS;
		info_p->port      = S502_portopt;
		info_p->irq       = S502A_irqopt;
		info_p->dpm_base  = S502A_memopt;
		info_p->dpm_size  = S502_winopt;
		info_p->clock	  = S502A_clkopt;
		break;

	case SDLA_S502E:
		info_p->hw_opt    = 0;
		info_p->io_range  = S502E_IORANGE;
		info_p->boot_offs = S502E_BOOTOFS;
		info_p->port      = S502_portopt;
		info_p->irq       = S502E_irqopt;
		info_p->dpm_base  = S502E_memopt;
		info_p->dpm_size  = S502_winopt;
		info_p->clock     = S502E_clkopt;
		break;

	case SDLA_S503:
		info_p->hw_opt    = 0;
		info_p->io_range  = S503_IORANGE;
		info_p->boot_offs = S503_BOOTOFS;
		info_p->port      = S503_portopt;
		info_p->irq       = S503_irqopt;
		info_p->dpm_base  = S502E_memopt;
		info_p->dpm_size  = S502_winopt;
		info_p->clock     = S503_clkopt;
		break;

	case SDLA_S508:
		info_p->hw_opt    = 0;
		info_p->io_range  = S508_IORANGE;
		info_p->boot_offs = S508_BOOTOFS;
		info_p->port      = S508_portopt;
		info_p->irq       = S508_irqopt;
		info_p->dpm_base  = S502E_memopt;
		info_p->dpm_size  = S508_winopt;
		info_p->clock     = S508_clkopt;
		break;

	case SDLA_S507:
		info_p->hw_opt    = 0;
		info_p->io_range  = S507_IORANGE;
		info_p->boot_offs = S507_BOOTOFS;
		info_p->port      = S508_portopt;
		info_p->irq       = S508_irqopt;
		info_p->dpm_base  = S507_memopt;
		info_p->dpm_size  = S507_winopt;
		info_p->clock     = S507_clkopt;
		break;

	default:
		err = CTLERR_FATAL;
	}
	return err;
}

/*============================================================================
 * Detect adapter type.
 *	Note: This test is destructive! Adapter will be left in shutdown
 *	state after the test.
 */
static int sdla_detect (sdlaData_t* ads_p)
{
	int err = 0;
	unsigned port = ads_p->io_base;

	if (detect_S502A(port))
		ads_p->type = SDLA_S502A
	;
	else if (detect_S502E(port))
		ads_p->type = SDLA_S502E
	;
	else if (detect_S503(port))
		ads_p->type = SDLA_S503
	;
	else if (detect_S507(port))
		ads_p->type = SDLA_S507
	;
	else if (detect_S508(port))
		ads_p->type = SDLA_S508
	;
	else err = CTLERR_FATAL;
	return err;
}

/*============================================================================
 * Configure SDLA: setup memory window, IRQ, etc.
 *	o Reset adapter and initialize registers mirror.
 *	o Check adapter presence.
 *	o Verify configuration options.
 *	o Configure adapter hardware.
 */
static int sdla_init (sdlaData_t* ads_p)
{
	int err = 0;
	int i;

	for (i = 0; i < SDLA_MAX_RANGE; ++i)
		ads_p->regs_wr[i] = ads_p->regs_rd[i] = 0
	;
	switch (ads_p->type)
	{
	case SDLA_S502A:
		err = init_S502A(ads_p);
		break;

	case SDLA_S502E:
		err = init_S502E(ads_p);
		break;

	case SDLA_S503:
		err = init_S503(ads_p);
		break;

	case SDLA_S507:
		err = init_S507(ads_p);
		break;

	case SDLA_S508:
		err = init_S508(ads_p);
		break;

	default:
		err = CTLERR_FATAL;
	}
	return err;
}

/*============================================================================
 * Shut down SDLA: disable shared memory, CPU, IRQ, etc.
 */
static int sdla_down (sdlaData_t* ads_p)
{
	int err = 0;
	unsigned port = ads_p->io_base;
	int i;

	switch (ads_p->type)
	{
	case SDLA_S502A:
		_OUTB(port, 0x08);		/* halt CPU */
		_OUTB(port, 0x08);
		_OUTB(port, 0x08);
		ads_p->regs_wr[0] = 0x08;
		_OUTB(port + 1, 0xFF);		/* close memory window */
		ads_p->regs_wr[1] = 0xFF;
		break;

	case SDLA_S502E:
		_OUTB(port + 3, 0);		/* stop CPU */
		_OUTB(port, 0);			/* reset board */
		for (i = 0; i < S502E_IORANGE; ++i)
			ads_p->regs_wr[i] = ads_p->regs_rd[i] = 0
		;
		break;

	case SDLA_S503:
	case SDLA_S507:
	case SDLA_S508:
		_OUTB(port, 0);			/* reset board logic */
		ads_p->regs_wr[0] = 0;
		break;

	default:
		err = CTLERR_FATAL;
	}
	return err;
}

/*============================================================================
 * Map shared memory window into SDLA adress space.
 */
static int sdla_mapmem (sdlaData_t* ads_p, unsigned long addr)
{
	int err	= 0;
	unsigned port = ads_p->io_base;
	register int tmp;

	switch (ads_p->type)
	{
	case SDLA_S502A:
	case SDLA_S502E:
		if (addr >= S502_MAXMEM)	/* verify parameter */
			err = CTLERR_BADARG
		;
		else
		{
			tmp = addr >> 13;	/* convert to register mask */
			_OUTB(port + 2, tmp);
			ads_p->regs_wr[2] = tmp;
		}
		break;

	case SDLA_S503:
		if (addr >= S503_MAXMEM)	/* verify parameter */
			err = CTLERR_BADARG
		;
		else
		{
			tmp = (ads_p->regs_wr[0] & 0x8F) | ((addr >> 9) & 0x70);
			_OUTB(port, tmp);
			ads_p->regs_wr[0] = tmp;
		}
		break;

	case SDLA_S507:
		if (addr >= S507_MAXMEM)
			err = CTLERR_BADARG
		;
		else if (!(_INB(port) & 0x02))
			err = CTLERR_BADSTATE
		;
		else
		{
			tmp = addr >> 13;	/* convert to register mask */
			_OUTB(port + 2, tmp);
			ads_p->regs_wr[2] = tmp;
		}
		break;

	case SDLA_S508:
		if (addr >= S508_MAXMEM)
			err = CTLERR_BADARG
		;
		else
		{
			tmp = addr >> 13;	/* convert to register mask */
			_OUTB(port + 2, tmp);
			ads_p->regs_wr[2] = tmp;
		}
		break;

	default:
		err = CTLERR_FATAL;
	}
	if (!err) ads_p->mem_vector = addr & 0xFFFFE000L;
	return err;
}

/*============================================================================
 * Start adapter's CPU.
 */
static int sdla_start (sdlaData_t* ads_p)
{
	int err	= 0;
	unsigned port = ads_p->io_base;
	int tmp, i;

	switch (ads_p->type)
	{
	case SDLA_S502A:
/*
		if (!(_INB(port) & 0x40))
			err = CTLERR_BADSTATE
		;
		else
*/
		{
			_OUTB(port, 0x10);	/* issue NMI to CPU */
			ads_p->regs_wr[0] = 0x10;
		}
		break;

	case SDLA_S502E:
		if (_INB(port) & 0x01)
			err = CTLERR_BADSTATE
		;
		else
		{
			_OUTB(port + 3, 0x01);		/* start CPU */
			ads_p->regs_wr[3] = 0x01;
			for (i = 0; i < SDLA_IODELAY; ++i);
			if (!(_INB(port) & 0x01))	/* verify */
				err = CTLERR_FATAL
			;
			else
			{
				/*
				 * Enabling CPU changes functionality of the
				 * control register, so we have to reset its
				 * mirror.
				 */
				_OUTB(port, 0);	/* disable interrupts */
				ads_p->regs_wr[0] = 0;
			}
		}
		break;

	case SDLA_S503:
		if (_INB(port) & 0x01)
			err = CTLERR_BADSTATE
		;
		else
		{
			tmp = ads_p->regs_wr[0] | 0x09;	/* set bits 0 and 3 */
			_OUTB(port, tmp);
			ads_p->regs_wr[0] = tmp;	/* update mirror */
			for (i = 0; i < SDLA_IODELAY; ++i);
			if (!(_INB(port) & 0x01))	/* verify */
				err = CTLERR_FATAL
			;
		}
		break;

	case SDLA_S507:
		if ((_INB(port) & 0x06) != 0x02)
			err = CTLERR_BADSTATE
		;
		else
		{
			tmp = ads_p->regs_wr[0] | 0x02;
			_OUTB(port, tmp);
			ads_p->regs_wr[0] = tmp;	/* update mirror */
			for (i = 0; i < SDLA_IODELAY; ++i);
			if (!(_INB(port) & 0x04))	/* verify */
				err = CTLERR_FATAL
			;
		}
		break;

	case SDLA_S508:
		if (_INB(port + 1) & 0x02)
			err = CTLERR_BADSTATE
		;
		else
		{
			tmp = ads_p->regs_wr[0] | 0x02;
			_OUTB(port, tmp);
			ads_p->regs_wr[0] = tmp;	/* update mirror */
			for (i = 0; i < SDLA_IODELAY; ++i);
			if (!(_INB(port + 1) & 0x02))	/* verify */
				err = CTLERR_FATAL
			;
		}
		break;

	default:
		err = CTLERR_FATAL;
	}
	return err;
}

/*============================================================================
 * Stop adapter's CPU.
 */
static int sdla_stop (sdlaData_t* ads_p)
{
	int err = 0;
	unsigned port = ads_p->io_base;
	int tmp, i;

	switch (ads_p->type)
	{
	case SDLA_S502A:
		_OUTB(port, 0x08);			/* halt CPU */
		_OUTB(port, 0x08);
		_OUTB(port, 0x08);
		ads_p->regs_wr[0] = 0x08;
		break;

	case SDLA_S502E:
		_OUTB(port + 3, 0);			/* stop CPU */
		ads_p->regs_wr[3] = 0;
		for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
		if (_INB(port) & 0x01)			/* verify */
			err = CTLERR_FATAL
		;
		else
		{
			/*
			 * Resetting CPU closes DPM window and forces IRQ high,
			 * so for compatibility with other SDLA types we
			 * re-enable memory access.
			 */
			_OUTB(port, 0x02);		/* re-enable memory */
			ads_p->regs_wr[0] = 0x02;
		}
		break;

	case SDLA_S503:
		tmp = ads_p->regs_wr[0] & ~0x09;	/* reset bits 0 and 3 */
		_OUTB(port, tmp);
		ads_p->regs_wr[0] = tmp;		/* update mirror */
		for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
		if (_INB(port) & 0x01)			/* verify */
			err = CTLERR_FATAL
		;
		break;

	case SDLA_S507:
		if (!(_INB(port) & 0x02))
			err = CTLERR_BADSTATE
		;
		else
		{
			tmp = ads_p->regs_wr[0] & ~0x02;
			_OUTB(port, tmp);
			ads_p->regs_wr[0] = tmp;	/* update mirror */
			for (i = 0; i < SDLA_IODELAY; ++i);
			if (_INB(port) & 0x04)		/* verify */
				err = CTLERR_FATAL
			;
		}
		break;

	case SDLA_S508:
		if (!(_INB(port + 1) & 0x02))
			err = CTLERR_BADSTATE
		;
		else
		{
			tmp = ads_p->regs_wr[0] & ~0x02;
			_OUTB(port, tmp);
			ads_p->regs_wr[0] = tmp;    	/* update mirror */
			for (i = 0; i < SDLA_IODELAY; ++i);
			if (_INB(port + 1) & 0x02)  	/* verify */
				err = CTLERR_FATAL
			;
		}
		break;

	default:
		err = CTLERR_FATAL;
	}
	return err;
}

/*============================================================================
 * Enable interrupt generation.
 */
static int sdla_inten (sdlaData_t* ads_p)
{
	int err = 0;
	unsigned port = ads_p->io_base;
	int tmp, i;

	switch (ads_p->type)
	{
	case SDLA_S502E:
		/*
		 * Note thar interrupt control operations on S502E are allowed
		 * only if CPU is enabled (bit 0 of status register is set).
		 */
		if (_INB(port) & 0x01)
		{
			_OUTB(port, 0x02);	/* bit1 = 1, bit2 = 0 */
			_OUTB(port, 0x06);	/* bit1 = 1, bit2 = 1 */
			ads_p->regs_wr[0] = 0x06;
		}
		else err = CTLERR_BADSTATE;
		break;

	case SDLA_S503:
		tmp = ads_p->regs_wr[0] | 0x04;
		_OUTB(port, tmp);
		ads_p->regs_wr[0] = tmp;		/* update mirror */
		for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
		if (!(_INB(port) & 0x02))		/* verify */
			err = CTLERR_FATAL
		;
		break;

	case SDLA_S508:
		tmp = ads_p->regs_wr[0] | 0x10;
		_OUTB(port, tmp);
		ads_p->regs_wr[0] = tmp;		/* update mirror */
		for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
		if (!(_INB(port + 1) & 0x10))		/* verify */
			err = CTLERR_FATAL
		;
		break;

	case SDLA_S502A:
	case SDLA_S507:
		break;

	default:
		err = CTLERR_FATAL;
	}
	return err;
}

/*============================================================================
 * Disable interrupt generation.
 */
static int sdla_intde (sdlaData_t* ads_p)
{
	int err = 0;
	unsigned port = ads_p->io_base;
	int tmp, i;

	switch (ads_p->type)
	{
	case SDLA_S502E:
		/*
		 * Note that on S502E:
		 *  1) interrupt control operations are allowed only if CPU is
		 *     enabled (bit 0 of status register is set).
		 *  2) disabling interrupts using bit 1 of control register
		 *     causes IRQ line go high, therefore we are going to use
		 *     0x04 instead: lower it to inhibit interrupts to PC.
		 */
		if (_INB(port) & 0x01)
		{
			_OUTB(port, ads_p->regs_wr[0] & ~0x04);
			ads_p->regs_wr[0] &= ~0x04;
		}
		else err = CTLERR_BADSTATE;
		break;

	case SDLA_S503:
		tmp = ads_p->regs_wr[0] & ~0x04;
		_OUTB(port, tmp);
		ads_p->regs_wr[0] = tmp;		/* update mirror */
		for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
		if (_INB(port) & 0x02)			/* verify */
			err = CTLERR_FATAL
		;
		break;

	case SDLA_S508:
		tmp = ads_p->regs_wr[0] & ~0x10;
		_OUTB(port, tmp);
		ads_p->regs_wr[0] = tmp;		/* update mirror */
		for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
		if (_INB(port) & 0x10)			/* verify */
			err = CTLERR_FATAL
		;
		break;

	case SDLA_S502A:
	case SDLA_S507:
		break;

	default:
		err = CTLERR_FATAL;
	}
	return err;
}

/*============================================================================
 * Acknowledge SDLA hardware interrupt.
 */
static int sdla_intack (sdlaData_t* ads_p)
{
	int err = 0;
	unsigned port = ads_p->io_base;
	int tmp;

	switch (ads_p->type)
	{
	case SDLA_S502E:
		/*
		 * To acknoledge hardware interrupt we have to toggle bit 3 of
		 * control register: \_/
		 * Note that interrupt control operations on S502E are allowed
		 * only if CPU is enabled (bit 1 of status register is set).
		 */
		if (_INB(port) & 0x01)
		{
			tmp = ads_p->regs_wr[0] & ~0x04;
			_OUTB(port, tmp);
			tmp |= 0x04;
			_OUTB(port, tmp);
			ads_p->regs_wr[0] = tmp;
		}
		else err = CTLERR_BADSTATE;
		break;

	case SDLA_S503:
		if (_INB(port) & 0x04)
		{
			tmp = ads_p->regs_wr[0] & ~0x08;
			_OUTB(port, tmp);
			tmp |= 0x08;
			_OUTB(port, tmp);
			ads_p->regs_wr[0] = tmp;
		}
		break;

	case SDLA_S502A:
	case SDLA_S507:
	case SDLA_S508:
	break;

	default:
		err = CTLERR_FATAL;
	}
	return err;
}

/*============================================================================
 * Generate an interrupt to adapter's CPU.
 */
static int sdla_isr (sdlaData_t* ads_p)
{
	int err = 0;
	unsigned port = ads_p->io_base;

	switch (ads_p->type)
	{
	case SDLA_S502A:
		if (_INB(port) & 0x40)
			err = CTLERR_BADSTATE
		;
		else
		{
			_OUTB(port, 0x10);		/* issue NMI to CPU */
			ads_p->regs_wr[0] = 0x10;
		}
		break;

	case SDLA_S507:
		if ((_INB(port) & 0x06) != 0x06)
			err = CTLERR_BADSTATE
		;
		else _OUTB(port + 3, 0);
		break;

	case SDLA_S508:
		if (!(_INB(port + 1) & 0x02))
			err = CTLERR_BADSTATE
		;
		else _OUTB(port, 0x08);
		break;

	case SDLA_S502E:
	case SDLA_S503:
	default:
		err = CTLERR_FATAL;
	}
	return err;
}

/*============================================================================
 * Initialize S502A adapter.
 */
static int init_S502A (sdlaData_t* ads_p)
{
	unsigned port = ads_p->io_base;
	int tmp, i;

	if (!detect_S502A(port))
		return CTLERR_FATAL
	;
	ads_p->regs_wr[0] = 0x08;
	ads_p->regs_wr[1] = 0xFF;

	/*
	 * Verify configuration options.
	 */
	i = opt_index(S502A_memopt, ads_p->dpm_base);
	if (i == 0)
		return CTLERR_BADARG
	;

	tmp = S502A_hmcr[i - 1];
	switch (ads_p->dpm_size)
	{
	case 0x2000:
		tmp |= 0x01;
		break;

	case 0x10000L:
		break;

	default:
		return CTLERR_BADARG;
	}

	/*
	 * Setup dual-port memory window (this also enables memory
	 * access).
	 */
	_OUTB(port + 1, tmp);
	ads_p->regs_wr[1] = tmp;
	ads_p->regs_wr[0] = 0x08;
	return 0;
}

/*============================================================================
 * Initialize S502E adapter.
 */
static int init_S502E (sdlaData_t* ads_p)
{
	unsigned port = ads_p->io_base;
	int tmp, i;

	if (!detect_S502E(port))
		return CTLERR_FATAL
	;

	/*
	 * Verify configuration options.
	 */
	i = opt_index(S502E_memopt, ads_p->dpm_base);
	if (i == 0)
		return CTLERR_BADARG
	;

	tmp = S502E_hmcr[i - 1];
	switch (ads_p->dpm_size)
	{
	case 0x2000:
		tmp |= 0x01;
		break;

	case 0x10000L:
		break;

	default:
		return CTLERR_BADARG;
	}

	/*
	 * Setup dual-port memory window.
	 */
	_OUTB(port + 1, tmp);
	ads_p->regs_wr[1] = tmp;

	/*
	 * Enable memory access.
	 */
	_OUTB(port, 0x02);
	ads_p->regs_wr[0] = 0x02;
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	return (_INB(port) & 0x02) ? 0 : CTLERR_FATAL;
}

/*============================================================================
 * Initialize S503 adapter.
 * ---------------------------------------------------------------------------
 */
static int init_S503 (sdlaData_t* ads_p)
{
	unsigned port = ads_p->io_base;
	int tmp, i;

	if (!detect_S503(port))
		return CTLERR_FATAL
	;

	/*
	 * Verify configuration options.
	 */
	i = opt_index(S502E_memopt, ads_p->dpm_base);
	if (i == 0)
		return CTLERR_BADARG
	;

	tmp = S502E_hmcr[i - 1];
	switch (ads_p->dpm_size)
	{
	case 0x2000:
		tmp |= 0x01;
		break;

	case 0x10000L:
		break;

	default:
		return CTLERR_BADARG;
	}

	/*
	 * Setup dual-port memory window.
	 */
	_OUTB(port + 1, tmp);
	ads_p->regs_wr[1] = tmp;

	/*
	 * Enable memory access.
	 */
	_OUTB(port, 0x02);
	ads_p->regs_wr[0] = 0x02;	/* update mirror */
	return 0;
}

/*============================================================================
 * Initialize S507 adapter.
 */
static int init_S507 (sdlaData_t* ads_p)
{
	unsigned port = ads_p->io_base;
	int tmp, i;

	if (!detect_S507(port))
		return CTLERR_FATAL
	;

	/*
	 * Verify configuration options.
	 */
	i = opt_index(S507_memopt, ads_p->dpm_base);
	if (i == 0)
		return CTLERR_BADARG
	;

	tmp = S507_hmcr[i - 1];
	switch (ads_p->dpm_size)
	{
	case 0x2000:
		tmp |= 0x01;
		break;

	case 0x10000L:
		break;

	default:
		return CTLERR_BADARG;
	}

	/*
	 * Enable adapter's logic.
	 */
	_OUTB(port, 0x01);
	ads_p->regs_wr[0] = 0x01;
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if (!(_INB(port) & 0x20))
		return CTLERR_FATAL
	;

	/*
	 * Setup dual-port memory window.
	 */
	_OUTB(port + 1, tmp);
	ads_p->regs_wr[1] = tmp;

	/*
	 * Enable memory access.
	 */
	tmp = ads_p->regs_wr[0] | 0x04;
	if (ads_p->irq)
	{
		i = opt_index(S508_irqopt, ads_p->irq);
		if (i) tmp |= S507_irqmask[i - 1];
	}
	_OUTB(port, tmp);
	ads_p->regs_wr[0] = tmp;		/* update mirror */
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	return (_INB(port) & 0x08) ? 0 : CTLERR_FATAL;
}

/*============================================================================
 * Initialize S508 adapter.
 */
static int init_S508 (sdlaData_t* ads_p)
{
	unsigned port = ads_p->io_base;
	int tmp, i;

	if (!detect_S508(port))
		return CTLERR_FATAL
	;

	/*
	 * Verify configuration options.
	 */
	i = opt_index(S502E_memopt, ads_p->dpm_base);
	if (i == 0)
		return CTLERR_BADARG
	;

	/*
	 * Setup memory configuration.
	 */
	tmp = S508_hmcr[i - 1];
	_OUTB(port + 1, tmp);
	ads_p->regs_wr[1] = tmp;

	/*
	 * Enable memory access.
	 */
	_OUTB(port, 0x04);
	ads_p->regs_wr[0] = 0x04;		/* update mirror */
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	return (_INB(port + 1) & 0x04) ? 0 : CTLERR_FATAL;
}

/*============================================================================
 * Detect S502A adapter.
 *	Following tests are used to detect S502A adapter:
 *	1. All registers other than status (BASE) should read 0xFF
 *	2. After writing 00001000b to control register, status register should
 *	   read 01000000b.
 *	3. After writing 0 to control register, status register should still
 *	   read  01000000b.
 *	4. After writing 00000100b to control register, status register should
 *	   read 01000100b.
 *	Return 1 if detected o.k. or 0 if failed.
 *	Note:	This test is destructive! Adapter will be left in shutdown
 *		state after the test.
 */
static int detect_S502A (int port)
{
	int i, j;

	if (!opt_index(S502_portopt, port))
		return 0
	;
	for (j = 1; j < SDLA_MAX_RANGE; ++j)
	{
		if (_INB(port + j) != 0xFF)
			return 0
		;
		for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	}

	_OUTB(port, 0x08);			/* halt CPU */
	_OUTB(port, 0x08);
	_OUTB(port, 0x08);
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if (_INB(port) != 0x40)
		return 0
	;
	_OUTB(port, 0x00);
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if (_INB(port) != 0x40)
		return 0
	;
	_OUTB(port, 0x04);
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if (_INB(port) != 0x44)
		return 0
	;

	/*
	 * Reset adapter.
	 */
	_OUTB(port, 0x08);
	_OUTB(port, 0x08);
	_OUTB(port, 0x08);
	_OUTB(port + 1, 0xFF);
	return 1;
}

/*============================================================================
 * Detect S502E adapter.
 *	Following tests are used to verify adapter presence:
 *	1. All registers other than status (BASE) should read 0xFF.
 *	2. After writing 0 to CPU control register (BASE+3), status register
 *	   (BASE) should read 11111000b.
 *	3. After writing 00000100b to port BASE (set bit 2), status register
 *	   (BASE) should read 11111100b.
 *	Return 1 if detected o.k. or 0 if failed.
 *	Note:	This test is destructive! Adapter will be left in shutdown
 *		state after the test.
 */
static int detect_S502E (int port)
{
	int i, j;

	if (!opt_index(S502_portopt, port))
		return 0
	;
	for (j = 1; j < SDLA_MAX_RANGE; ++j)
	{
		if (_INB(port + j) != 0xFF)
			return 0
		;
		for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	}

	_OUTB(port + 3, 0);			/* CPU control reg. */
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if (_INB(port) != 0xF8)			/* read status */
		return 0
	;
	_OUTB(port, 0x04);			/* set bit 2 */
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if (_INB(port) != 0xFC)			/* verify */
		return 0
	;

	/*
	 * Reset adapter.
	 */
	_OUTB(port, 0);
	return 1;
}

/*============================================================================
 * Detect S503 adapter.
 *	Following tests are used to verify adapter presence:
 *	1. All registers other than status (BASE) should read 0xFF.
 *	2. After writing 0 to control register (BASE), status register (BASE)
 *	   should read 11110000b.
 *	3. After writing 00000100b (set bit 2) to control register (BASE),
 *	   status register should read 11110010b.
 *	Return 1 if detected o.k. or 0 if failed.
 *	Note:	This test is destructive! Adapter will be left in shutdown
 *		state after the test.
 */
static int detect_S503 (int port)
{
	int i, j;

	if (!opt_index(S503_portopt, port))
		return 0
	;
	for (j = 1; j < SDLA_MAX_RANGE; ++j)
	{
		if (_INB(port + j) != 0xFF)
			return 0
		;
		for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	}

	_OUTB(port, 0);				/* reset control reg.*/
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if (_INB(port) != 0xF0)			/* read status */
		return 0
	;
	_OUTB(port, 0x04);			/* set bit 2 */
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if (_INB(port) != 0xF2)			/* verify */
		return 0
	;

	/*
	 * Reset adapter.
	 */
	_OUTB(port, 0);
	return 1;
}

/*============================================================================
 * Detect S507 adapter.
 *	Following tests are used to detect S507 adapter:
 *	1. All ports should read the same value.
 *	2. After writing 0x00 to control register, status register should read
 *	   ?011000?b.
 *	3. After writing 0x01 to control register, status register should read
 *	   ?011001?b.
 *	Return 1 if detected o.k. or 0 if failed.
 *	Note:	This test is destructive! Adapter will be left in shutdown
 *		state after the test.
 */
static int detect_S507 (int port)
{
	int tmp, i, j;

	if (!opt_index(S508_portopt, port))
		return 0
	;
	tmp = _INB(port);
	for (j = 1; j < S507_IORANGE; ++j)
	{
		if (_INB(port + j) != tmp)
			return 0
		;
		for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	}

	_OUTB(port, 0x00);
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if ((_INB(port) & 0x7E) != 0x30)
		return 0
	;
	_OUTB(port, 0x01);
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if ((_INB(port) & 0x7E) != 0x32)
		return 0
	;

	/*
	 * Reset adapter.
	 */
	_OUTB(port, 0x00);
	return 1;
}

/*============================================================================
 * Detect S508 adapter.
 *	Following tests are used to detect S508 adapter:
 *	1. After writing 0x00 to control register, status register should read
 *	   ??000000b.
 *	2. After writing 0x10 to control register, status register should read
 *	   ??010000b
 *	Return 1 if detected o.k. or 0 if failed.
 *	Note:	This test is destructive! Adapter will be left in shutdown
 *		state after the test.
 */
static int detect_S508 (int port)
{
	int i;

	if (!opt_index(S508_portopt, port))
		return 0
	;
	_OUTB(port, 0x00);
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if ((_INB(port + 1) & 0x3F) != 0x00)
		return 0
	;
	_OUTB(port, 0x10);
	for (i = 0; i < SDLA_IODELAY; ++i);	/* delay */
	if ((_INB(port + 1) & 0x3F) != 0x10)
		return 0
	;

	/*
	 * Reset adapter.
	 */
	_OUTB(port, 0x00);
	return 1;
}

/*============================================================================
 * Get option's index into the options list.
 *	Return option's index (1 .. N) or zero if option is invalid.
 */
static int opt_index (unsigned* optlist, unsigned optval)
{
	int i;

	for (i = 1; i <= optlist[0]; ++i)
		if (optlist[i] == optval) return i
	;
	return 0;
}

/******	END *****************************************************************/
