/*****************************************************************************
* PROJECT:	Sangoma SDLA Driver Protocol-Specific Module Template
* FILE:		sppp.c		Main Module
* TARGET OS:	Linux 1.2.x, 2.0.x
* AUTHOR:	Gene Kozin, Sangoma Technologies Inc.
* ----------------------------------------------------------------------------
* Copyright (c) 1995-1996 Sangoma Technologies Inc.  All Rights Reserved.
* ============================================================================
* Rev.#	Date		Who	Details
* -----	-----------	---	----------------------------------------------
*  1.22 Feb.21,1997     FT      o Fixed '+1 bug' in pppReceive()
*  1.21	Nov.08,1996	GK	o Updated for SDLA driver v3.10
*  1.20	Oct.09,1996	GK	o Added transmit and connection status
*				  interrupt handlers.
*  1.11	Sep.17,1996	GK	o Fixed bug causing page fault in ioctl_stat().
*  1.10	Jul.15,1996	GK	o Changes for Linux 2.0 compatibility.
*  1.00	Jun.04,1996	GK	o Initial version.
*****************************************************************************/

#include <linux/config.h>	/* OS configuration options */
#include <linux/kernel.h>
#include <linux/version.h>	/* for UTS_RELEASE */
#include <linux/module.h>	/* support for loadable modules */
#include <linux/fs.h>		/* support for device drivers */
#include <linux/mm.h>		/* verify_area(), etc. */
#include <linux/malloc.h>	/* kmalloc(), kfree() */
#include <linux/errno.h>	/* return codes */
#include <asm/string.h>		/* inline memset(), etc. */
#include <asm/segment.h>	/* kernel <-> user copy */
				/*----- Network-specific stuff -------------*/
#include <linux/inet.h>		/* in_aton(), in_ntoa() prototypes */
#include <linux/netdevice.h>	/* support for network drivers */
#include <linux/if.h>		/* network interface stuff */
#include <linux/if_ether.h>	/* for ethertypes */
#include <linux/if_arp.h>	/* for ARPHRD_* */
#include <linux/skbuff.h>	/* for struct sk_buff */
#include <asm/byteorder.h>	/* htons(), etc. */

#include <sdladrv.h>		/* SDLA driver kernel API definitions */
#include <ppp_api.h>		/* PPP firmware API definitions */
#include <spppio.h>		/* this driver API definitions */

/****** Defines *************************************************************/

#define	MOD_NAME	"sppp"
#define	MOD_VERSION	SPPP_VERSION	/* version number */
#define	MOD_RELEASE	21		/* release (minor version) number */

#if (_OS_VER_ == 102)
#define	MOD_FULLNAME	"WANPIPE(tm) PPP Router Module for Linux 1.2.x";
#elif (_OS_VER_ == 103)
#define	MOD_FULLNAME	"WANPIPE(tm) PPP Router Module for Linux 1.3.x";
#elif (_OS_VER_ == 200)
#define	MOD_FULLNAME	"WANPIPE(tm) PPP Router Module for Linux 2.0.x";
#else
#error _OS_VER_ is not defined or invalid!
#endif

/*
 * Mins, maxes and defaults.
 */
#define	MAX_IFS		SDLA_MAX_UNITS	/* max number of interfaces */
#define	IF_NAME_LEN	16		/* max length of interface name */
#define	DFLT_MTU	1500		/* default interface MTU */

/*
 * Miscellaneous.
 */
#define	CMD_OK		0		/* normal firmware return code */
#define	CMD_TIMEOUT	0xFF		/* firmware command timed out */

/****** Macros **************************************************************/

/****** Data Types **********************************************************/

typedef struct	enet_statistics	IfStats_t;	/* for sake of shortness */

/*----------------------------------------------------------------------------
 * Interface Data Space.
 */
typedef struct	spppData
{
  struct device	dev;			/* Linux network device. Must be 1st */
  IfStats_t	ifstats;		/* statistics */
  char		name[IF_NAME_LEN];	/* interface name */
  int		ifstate;		/* interface state */
  int		ifno;			/* interface number */
  sdlaData_t*	ads;			/* -> adapter data space */
  spppConf_t	cfg;			/* interface configuration */
  unsigned long	cmd_vect;		/* vector of the control window */
  unsigned int	cmd_offs;		/* offset of PPP command structure */
  unsigned int	flg_offs;		/* offset of PPP status flags */
  unsigned int	txc_offs;		/* offset of transmit control blocks*/
  unsigned int	rxc_offs;		/* offset of receive control blocks */
  unsigned int	tx_buf_num;		/* number of transmit buffers */
  unsigned int	tx_buf_idx;		/* next transmit buffer index */
  unsigned int	rx_buf_num;		/* number of receive buffers */
  unsigned int	rx_buf_idx;		/* next receive buffer index */
  unsigned int	rx_buf_base;		/* base address of receive buffer */
  unsigned int	rx_buf_top;		/* top address of receive buffer */
  unsigned int	rx_intr_cnt;		/* receive interrupt count */
  unsigned int	tx_intr_cnt;		/* transmit interrupt count */
  unsigned int	ctl_intr_cnt;		/* connection control intr count */
  unsigned int	spurious_cnt;		/* spurious interrupt count */
} spppData_t;

/*
 * 'ifstate' defines.
 */
enum IfState
{
  IFS_DOWN	= 1,		/* initial state */
  IFS_ENABLED,			/* communications were enabled */
  IFS_READY			/* link is up, ready to send/receive data */
};

/****** Function Prototypes *************************************************/

/*
 * Driver Entry Points. (Called by the operating system).
 */
int init_module (void);
void cleanup_module (void);

/*
 * SDLA Module Entry Points. (Called by the SDLA Driver).
 */
static int mod_bind (sdlaData_t* ads);
static int mod_unbind (sdlaData_t* ads);
static int mod_ioctl (sdlaData_t* ads, int cmd, void* ptr);
static int mod_halt (sdlaData_t* ads);
static int mod_intr (sdlaData_t* ads);

/*
 * Linux Network Device Driver Entry Points.
 */
static int if_init (struct device* dev);
static int if_open (struct device* dev);
static int if_close (struct device* dev);
static int if_ioctl (struct device* dev, struct ifreq* ifr, int cmd);
static int if_send (struct sk_buff* skb, struct device* dev);
static IfStats_t *if_stats (struct device* dev);

#if (_OS_VER_ < 103)
static unsigned short if_type_trans (struct sk_buff *skb, struct device *dev);
#endif

/*
 * IOCTL Handlers.
 */
static int ioctl_conf (spppData_t* ifd_p, spppConf_t* conf_p);
static int ioctl_stat (spppData_t* ifd_p, spppStat_t* stat_p);
static int ioctl_info (spppData_t* ifd_p, spppInfo_t* info_p);

/*
 * Interrupt Handlers.
 */
static int intr_receive (spppData_t* ifd_p);
static int intr_transmit (spppData_t* ifd_p);
static int intr_connect (spppData_t* ifd_p);

/*
 * Firmware-Specific Functions.
 */
static int pppCodeVersion (spppData_t* ifd_p, char* str);
static int pppConfigure (spppData_t* ifd_p);
static int pppCommDisable (spppData_t* ifd_p);
static int pppCommEnable (spppData_t* ifd_p);
static int pppSetIntrMode (spppData_t* ifd_p, int mode);
static int pppCommStats (spppData_t* ifd_p, spppStat_t* stat_p);
static int pppSend (spppData_t* ifd_p, void* buf, int len, int proto);
static int pppReceive (spppData_t* ifd_p, void* buf, int* len_p, int* proto_p);
static int pppEvent (spppData_t* ifd_p, int event, int cmd);

/*
 * Miscellaneous Functions.
 */
static unsigned char bps_to_line_speed (unsigned long bps);

/****** Global Data *********************************************************/

/*
 * Module public data.
 */
#if (defined(_LINUX_) && (_OS_VER_ < 200))
char kernel_version[] = UTS_RELEASE;		/* must be public */
#endif

/*
 * Module private data (define as static).
 */
static char mod_name[] = MOD_NAME;
static char mod_fullname[] = MOD_FULLNAME;
static char mod_copyright[] = "(c) 1995-1996 Sangoma Technologies Inc.";

static sdlaMod_t modreg =		/* Module registration info */
{
  MOD_NAME,		/* .name */
  NULL,			/* .next */
  &mod_bind,		/* .bind */
};

static spppData_t ifdata[MAX_IFS];	/* interface data space */

/******* System Module Entry Points *****************************************/

/*============================================================================
 * Module Initialization Routine.
 * ---------------------------------------------------------------------------
 * Method:
 *  o Announce ourselves.
 *  o Initialize static data.
 *  o Register ourselves with SDLA driver.
 */
int init_module (void)
{
	printk("%s v%d.%d %s\n",
		mod_fullname, MOD_VERSION, MOD_RELEASE, mod_copyright)
	;

	/*
	 * Initialize static data.
	 */
	memclr(ifdata, sizeof(ifdata));

	/*
	 * Register with SDLA driver.
	 */
	if (sdla_register(SDLA_SIGNATURE, SDLA_VERSION, &modreg))
	{
		if (SDLA_DEBUG) printk(
			"%s: failed to register with SDLA driver!\n",
			mod_name)
		;
		return -1;
	}
	return 0;
}

/*============================================================================
 * Module Remove Entry Point.
 * ---------------------------------------------------------------------------
 */
void cleanup_module (void)
{
	if (SDLA_DEBUG > 2) printk("%s: shutting down...\n", mod_name);
	sdla_unregister(mod_name);
}

/******* SDLA Module Entry Points *******************************************/

/*============================================================================
 * mod_bind()		Bind module.
 * ---------------------------------------------------------------------------
 * o Check adapter state and firmware compatibility.
 * o Allocate interface data space.
 * o Register Linux network interface.
 * Return 0 if done or SDLAER_* if error.
 */
static int mod_bind (sdlaData_t* ads)
{
  int err = 0;
  spppData_t* ifd_p = &ifdata[ads->cardno];

  /*
   * Check adapter state and firmware compatibility.
   */
  if (ads->state != SDLA_RUNNING)
    err = SDLAER_BADSTATE
  ;
  else if (ifd_p->ifstate != 0)
    err = SDLAER_BADSTATE
  ;
  else
  {
    /*
     * Initialize interface data space.
     */
    memclr(ifd_p, sizeof(spppData_t));
    sprintf(ifd_p->name, "%s%u", mod_name, ads->cardno);
    ifd_p->ads = ads;

    /*
     * Initialize firmware version-dependent parameters.
     */
    switch (ads->firmware_id)
    {
      case SFID_PPP502:
        ifd_p->cmd_vect = 0xA000;
        ifd_p->cmd_offs = 0x1C00;
        ifd_p->flg_offs = 0;
        break;

      case SFID_PPP508:
        ifd_p->cmd_vect = 0xE000;
        ifd_p->cmd_offs = 0;
        ifd_p->flg_offs = 0x1000;
        break;

      default:
        err = SDLAER_BADTYPE;
    }

    if (!err)
    {
      /*
       * Register Linux network driver.
       */
      ifd_p->dev.name = ifd_p->name;
      ifd_p->dev.init = if_init;
      if(register_netdev(&ifd_p->dev))
        err = SDLAER_FAILURE
      ;
      else
      {
        /*
         * Setup module entry points and private data.
         */
        ads->read	= NULL;
        ads->write	= NULL;
        ads->ioctl	= &mod_ioctl;
        ads->intr	= &mod_intr;
        ads->halt	= &mod_halt;
        ads->unbind	= &mod_unbind;
        ads->priv	= ifd_p;
        MOD_INC_USE_COUNT;		/* prevent module from unloading */
      }
    }
  }
  return err;
} /* mod_bind() */

/*============================================================================
 * mod_unbind()		Unbind module.
 * ---------------------------------------------------------------------------
 * o Unregister Linux network interface.
 * o Release resources.
 */
static int mod_unbind (sdlaData_t* ads)
{
  int		err	= 0;
  spppData_t*	ifd_p	= ads->priv;

  if ((ifd_p == NULL) || ifd_p->dev.start)
    err = SDLAER_BADSTATE		/* interface is in use */
  ;
  else
  {
    unregister_netdev(&ifd_p->dev);	/* de-register interface */
    ifd_p->ifstate = 0;
    ads->priv  = NULL;
    MOD_DEC_USE_COUNT;			/* allow unloading the module */
  }
  return err;
} /* mod_unbind() */

/*============================================================================
 * mod_halt()		Module shutdown.
 * ---------------------------------------------------------------------------
 */
static int mod_halt (sdlaData_t* ads)
{
	if (SDLA_DEBUG > 2) printk("%s: mod_halt() called\n", mod_name);
	return 0;
}

/*============================================================================
 * mod_ioctl()		Module IOCTL Entry Point
 * ---------------------------------------------------------------------------
 * This routine is called by the SDLA driver when system ioctl() call is
 * invoked.  At this point all verifications are done and argument 'ptr'
 * points to 'user' area of sdlaCtl_t structure in user data space.
 */
static int mod_ioctl (sdlaData_t* ads, int cmd, void* ptr)
{
	return -EINVAL;
}

/*============================================================================
 * mod_intr()		Module Interrupt Routine.
 * ---------------------------------------------------------------------------
 * Note: when interrupt occurs the DPM window vector is undefined.  Therefore
 *       we have to switch it to adapter's mailbox before accessing shared
 *       memory and restore it on exit.
 */
static int mod_intr (sdlaData_t* ads)
{
	spppData_t* ifd_p  = ads->priv;
	pppFlags_t* flg_p  = (void*)(ads->vbase + ifd_p->flg_offs);
	unsigned long cur_vect = ads->mem_vector;	/* current vector */

	if (ifd_p->dev.interrupt)
	{
		if (SDLA_DEBUG > 1) printk (
			"%s: Interrupt re-entrancy on interface %s!\n",
			mod_name, ifd_p->name)
		;
	}
	else ifd_p->dev.interrupt = 1;	/* set 'in ISR' flag */

	/*
	 * Set DPM window vector for firmware mailbox access.
	 */
	if (cur_vect != ifd_p->cmd_vect)
		sdlactl.mapmem(ads, ifd_p->cmd_vect);
	;
	switch (flg_p->intr_flag)
	{
	case 0x01:		/* receive interrupt */
		++ifd_p->rx_intr_cnt;
		intr_receive(ifd_p);
		break;

	case 0x02:		/* transmit interrupt */
		++ifd_p->tx_intr_cnt;
		intr_transmit(ifd_p);
		break;

	case 0x10:		/* connection control interrupt */
		++ifd_p->ctl_intr_cnt;
		intr_connect(ifd_p);
		break;

	default:		/* unwanter interrupt */
		++ifd_p->spurious_cnt;
	}
	flg_p->intr_flag = 0;		/* clear interrupt condition */

	/*
	 * Restore original DPM window vector.
	 */
	if (cur_vect != ifd_p->cmd_vect)
		sdlactl.mapmem(ads, cur_vect);
	;
	ifd_p->dev.interrupt = 0;	/* clear 'in ISR' flag */
	return 0;
} /* mod_intr() */

/****** Linux Network Device Driver Entry Points ****************************/

/*============================================================================
 * Initialize Network Interface.
 * ---------------------------------------------------------------------------
 * This routine is called only once, during Linux network device registration.
 * Returning anything but zero will fail the registration.
 */
static int if_init (struct device *dev)
{
#define	ifd_p	((spppData_t*)dev)

  sdlaData_t*	ads	= ifd_p->ads;	/* -> adapter data space */
  char		str[80];
  int		i;

  /*
   * Read firmware code version.
   */
  if (pppCodeVersion(ifd_p, str) != CMD_OK)
    return -1
  ;
  if (SDLA_DEBUG > 1)
    printk("%s: interface %s running PPP firmware v%s\n",
           mod_name, ifd_p->name, str)
  ;

  /*
   * Disable communications and reset interface state.
   */
  pppCommDisable(ifd_p);
  ifd_p->ifstate = IFS_DOWN;

  /*
   * Initialize device structure.
   */
  dev->open		= if_open;
  dev->stop		= if_close;
  dev->do_ioctl		= if_ioctl;
  dev->hard_start_xmit	= if_send;
  dev->get_stats	= if_stats;
  dev->hard_header	= NULL;			/* built by adapter */
  dev->rebuild_header	= NULL;
#if (_OS_VER_ < 103)
  dev->type_trans	= if_type_trans;
#endif
  dev->irq		= ads->irq;		/* for reference only */
  dev->base_addr	= ads->io_base;	/* for reference only */
  dev->mem_start	= ads->dpm_base;	/* for reference only */
  dev->mem_end		= ads->dpm_base + ads->dpm_size;
  dev->mtu		= DFLT_MTU;
  dev->family		= AF_INET;		/* address family */
  dev->type		= ARPHRD_PPP;		/* ARP interface h/w type */
  dev->pa_alen		= sizeof(unsigned long); /* protocol address length */
  dev->hard_header_len	= 0;			/* media header length */
  dev->addr_len		= 0;			/* hardware address length */
/*  dlci_to_Q922(ifd_p->dlci, dev->dev_addr);*/	/* hardware address */

  /*
   * Initialize socket buffers.
   */
  for (i = 0; i < DEV_NUMBUFFS; ++i)
    skb_queue_head_init(&dev->buffs[i])
  ;

  /*
   * Set default PPP link-level configuration.
   */
  ifd_p->cfg.interface     = 0;		/* RS-232 */
  ifd_p->cfg.clocking      = 0;		/* External */
  ifd_p->cfg.speed         = 0;		/*  */
  ifd_p->cfg.restart_tmr   = 30;	/*  */
  ifd_p->cfg.auth_rsrt_tmr = 30;	/*  */
  ifd_p->cfg.auth_wait_tmr = 300;	/*  */
  ifd_p->cfg.mdm_fail_tmr  = 5;		/*  */
  ifd_p->cfg.dtr_drop_tmr  = 10;	/*  */
  ifd_p->cfg.connect_tmout = 0;		/*  */
  ifd_p->cfg.conf_retry    = 10;	/*  */
  ifd_p->cfg.term_retry    = 2;		/*  */
  ifd_p->cfg.fail_retry    = 5;		/*  */
  ifd_p->cfg.auth_retry    = 10;	/*  */
  ifd_p->cfg.auth_options  = 0;		/*  */
  ifd_p->cfg.ip_options    = 0;		/*  */

  return 0;

#undef	ifd_p
} /* if_init() */

/*============================================================================
 * Open Network Interface.
 * ---------------------------------------------------------------------------
 * o Configure PPP firmware.
 * o Enable communications.
 * Return 0 if O.k. or errno.
 */
static int if_open (struct device* dev)
{
#define	ifd_p	((spppData_t*)dev)

  sdlaData_t*	ads	= ifd_p->ads;	/* -> adapter data space */
  int		err	= 0;

  if (dev->start)
    err = -EBUSY			/* only one open is allowed */
  ;
  else
  {
    /*
     * Adjust some interface data.
     */
    dev->irq	= ads->irq;		/* not used, but may be confusing */
    dev->mtu	= max(dev->mtu, 1500);
    dev->mtu	= min(dev->mtu, 4500);

    /*
     * Configure adapter firmware.
     */
    if ((pppCommDisable(ifd_p)       != CMD_OK) ||
        (pppConfigure(ifd_p)         != CMD_OK) ||
        (pppSetIntrMode(ifd_p, 0x13) != CMD_OK))
      err = -EIO
    ;
    else
    {
      /*
       * Initialize Rx/Tx buffer control fields.
       */
      ifd_p->tx_buf_idx = 0;
      ifd_p->rx_buf_idx = 0;
      switch (ads->firmware_id)
      {
        case SFID_PPP502:
          ifd_p->txc_offs    = *(unsigned short*)(ads->vbase + 0x12);
          ifd_p->rxc_offs    = *(unsigned short*)(ads->vbase + 0x1A);
          ifd_p->tx_buf_num  = *(unsigned short*)(ads->vbase + 0x10);
          ifd_p->rx_buf_num  = *(unsigned short*)(ads->vbase + 0x18);
          ifd_p->rx_buf_base = 0;	/* not used */
          ifd_p->rx_buf_top  = 0;	/* not used */
          break;

        case SFID_PPP508:
        default:
          ifd_p->txc_offs =
            *(unsigned long*)(ads->vbase + 0x1102) - ads->mem_vector
          ;
          ifd_p->rxc_offs =
            *(unsigned long*)(ads->vbase + 0x1122) - ads->mem_vector
          ;
          ifd_p->tx_buf_num  = *(unsigned short*)(ads->vbase + 0x1100);
          ifd_p->rx_buf_num  = *(unsigned short*)(ads->vbase + 0x1120);
          ifd_p->rx_buf_base = *(unsigned long*)(ads->vbase + 0x112A);
          ifd_p->rx_buf_top  = *(unsigned long*)(ads->vbase + 0x1130);
      }

      /*
       * Enable communications.
       */
      cli();
      if (pppCommEnable(ifd_p) != CMD_OK)
        err = -EIO
      ;
      else
      {
        ifd_p->ifstate = IFS_ENABLED;
        dev->tbusy     = 0;
        dev->interrupt = 0;
        dev->start     = 1;
      }
      sti();
    }
  }
  return err;

#undef	ifd_p
} /* if_open() */

/*============================================================================
 * Close Network Interface.
 * ---------------------------------------------------------------------------
 * o Disable PPP communications.
 * o Reset flags.
 */
static int if_close (struct device* dev)
{
#define	ifd_p	((spppData_t*)dev)

	pppCommDisable(ifd_p);
	pppSetIntrMode(ifd_p, 0);
	ifd_p->ifstate = IFS_DOWN;
	dev->start = 0;
	return 0;

#undef	ifd_p
}

/*============================================================================
 * Get Interface Statistics.
 * ---------------------------------------------------------------------------
 */
static IfStats_t* if_stats (struct device* dev)
{
#define	ifd_p	((spppData_t*)dev)

	return &ifd_p->ifstats;

#undef	ifd_p
}

/*============================================================================
 * Send a Packet on a Network Interface.
 * ---------------------------------------------------------------------------
 */
static int if_send (struct sk_buff* skb, struct device* dev)
{
#define	ifd_p	((spppData_t*)dev)

	pppFlags_t* flg_p;

	if ((skb == NULL) || (dev == NULL)) return 0;
	if (dev->interrupt || set_bit(0, (void*)&dev->tbusy))
	{
		if (SDLA_DEBUG > 1) printk (
			"%s: if_send() re-entrancy on interface %s!\n",
			mod_name, ifd_p->name)
		;
		++ifd_p->ifstats.tx_errors;
		return 1;	/* sending or serving interrupt, try later */
	}

	flg_p = (void*)(ifd_p->ads->vbase + ifd_p->flg_offs);
	if (flg_p->lcp_state != 9)		/* check link status */
	{
		if (SDLA_DEBUG > 1) printk (
			"%s: Link is down on interface %s! LCP state: 0x%02X\n",
			mod_name, ifd_p->name, flg_p->lcp_state)
		;
		++ifd_p->ifstats.tx_dropped;
	}
	else if (pppSend(ifd_p, skb->data, skb->len, 0) == CMD_OK)
	{
		++ifd_p->ifstats.tx_packets;
	}
	else
	{
		flg_p->intr_mask |= 0x02;	/* unmask Tx interrupts */
		return 1;			/* busy, try later */
	}

	dev_kfree_skb(skb, FREE_WRITE);		/* release socket buffer */
	dev->tbusy = 0;				/* clear 'busy' flag */
	return 0;

#undef	ifd_p
}

#if (_OS_VER_ < 103)
/*============================================================================
 * Determine packet's protocol ID.
 * ---------------------------------------------------------------------------
 */
static unsigned short if_type_trans (struct sk_buff* skb, struct device* dev)
{
	return (unsigned short)skb->pkt_class;
}
#endif

/*============================================================================
 * Network Interface IOCTL.
 * ---------------------------------------------------------------------------
 * Return 0 if O.k.
 */
static int if_ioctl (struct device* dev, struct ifreq* ifr, int cmd)
{
#define	ifd_p	((spppData_t*)dev)

  int		err	= 0;		/* system return code (errno) */
  int		ioerr	= 0;		/* SPPP I/O result */ 
  spppCtl_t*	ioctl_p;

  /*
   * Get a pointer to SPPP control block.
   */
  ioctl_p = (void *)ifr->ifr_ifru.ifru_data;

  /*
   * Verify IOCTL request arguments.
   */
  if (!ioctl_p || verify_area(VERIFY_WRITE, ioctl_p, sizeof(spppCtl_t)))
    err = -EFAULT
  ;
#if (_OS_VER_ < 200)
  else if (get_user_long((void *)&ioctl_p->signature) != SPPP_SIGNATURE)
#else
  else if (get_user(&ioctl_p->signature) != SPPP_SIGNATURE)
#endif
    err = -EINVAL		/* invalid signature */
  ;
#if (_OS_VER_ < 200)
  else if (get_user_long(&ioctl_p->version) != SPPP_VERSION)
#else
  else if (get_user(&ioctl_p->version) != SPPP_VERSION)
#endif
    ioerr = SPPPER_BADVER	/* incorrect driver version */
  ;
  else switch (cmd)		/* process IOCTL request */
  {
    case SPPPIO_CONF:
      ioerr = ioctl_conf(ifd_p, &ioctl_p->v.conf);
      break;
 
    case SPPPIO_STAT:
      ioerr = ioctl_stat(ifd_p, &ioctl_p->v.stat);
      break;
 
    case SPPPIO_INFO:
      ioerr = ioctl_info(ifd_p, &ioctl_p->v.info);
      break;
 
    default:
      ioerr = SPPPER_BADCMD;
  }

  /*
   * If IOCTL error has occured, pass error code to the user and
   * return EIO to the system.
   */
  if (ioerr)
  {
#if (_OS_VER_ < 200)
    put_user_long(ioerr, &ioctl_p->result);
#else
    put_user(ioerr, &ioctl_p->result);
#endif
    err = -EIO;
    if (SDLA_DEBUG > 1)
      printk("%s: error %d on interface %s during IOCTL %d.\n",
             mod_name, ioerr, ifd_p->name, cmd - SIOCDEVPRIVATE)
    ;
  }
  return err;

#undef	ifd_p
} /* if_ioctl() */

/****** IOCTL Handlers ******************************************************/

/*============================================================================
 * SPPPIO_CONF Handler.
 * ---------------------------------------------------------------------------
 */
static int ioctl_conf (spppData_t* ifd_p, spppConf_t* conf_p)
{
	int err = 0;

	if (ifd_p->ifstate != IFS_DOWN)
		err = SDLAER_BADSTATE
	;
	else
	{
		/*
		 * Get configuration data from user address space.
		 */
		memcpy_fromfs(
			(void*)&ifd_p->cfg, (void*)conf_p, sizeof(spppConf_t))
		;
		switch (pppConfigure(ifd_p))
		{
		case CMD_OK:
			break;

		case CMD_TIMEOUT:
			err = SDLAER_TIMEOUT;
			break;

		default:
			err = -20;
		}
	}
	return err;
}

/*============================================================================
 * SPPPIO_STAT Handler.
 * ---------------------------------------------------------------------------
 */
static int ioctl_stat (spppData_t* ifd_p, spppStat_t* stat_p)
{
	int err = 0;
	spppStat_t stat;

	memclr(&stat, sizeof(spppStat_t));
	if (pppCommStats(ifd_p, &stat) != CMD_OK)
		err = SDLAER_TIMEOUT
	;
	else
	{
		stat.rx_intr_cnt  = ifd_p->rx_intr_cnt;
		stat.tx_intr_cnt  = ifd_p->tx_intr_cnt;
		stat.ctl_intr_cnt = ifd_p->ctl_intr_cnt;
		stat.spurious_cnt = ifd_p->spurious_cnt;
		memcpy_tofs((void*)stat_p, (void*)&stat, sizeof(spppStat_t));
	}
	return err;
}

/*============================================================================
 * SPPPIO_INFO Handler.
 * ---------------------------------------------------------------------------
 */
static int ioctl_info (spppData_t* ifd_p, spppInfo_t* info_p)
{
	int err = 0;

	return err;
}

/****** Interrupt Handlers **************************************************/

/*============================================================================
 * Receive Interrupt Handler.
 * ---------------------------------------------------------------------------
 */
static int intr_receive (spppData_t* ifd_p)
{
	struct sk_buff* skb_p = alloc_skb(ifd_p->dev.mtu, GFP_ATOMIC);
	int proto;

	if (skb_p == NULL)
	{
		++ifd_p->ifstats.rx_dropped;
		return 0;
	}

	if (pppReceive(ifd_p, skb_p->data, (void*)&skb_p->len, &proto))
	{
		++ifd_p->ifstats.rx_errors;
		dev_kfree_skb(skb_p, FREE_WRITE);	/* release buffer */
		return 0;
	}

	/*
	 * Fill the rest of socket buffer fields and pass it up the protocol
	 * stack.
	 */
	++ifd_p->ifstats.rx_packets;
	proto = (proto == 1) ? ETH_P_IPX : ETH_P_IP;
#if (_OS_VER_ < 103)
	skb_p->pkt_class = htons(proto);
#else
	skb_p->protocol = htons(proto);
#endif
	skb_p->dev = &ifd_p->dev;
	netif_rx(skb_p);
	return 0;
}

/*============================================================================
 * Transmit Interrupt Handler.
 * ---------------------------------------------------------------------------
 */
static int intr_transmit (spppData_t* ifd_p)
{
	pppFlags_t* flg_p = (void*)(ifd_p->ads->vbase + ifd_p->flg_offs);

	ifd_p->dev.tbusy = 0;		/* clear 'Tx busy' flag */
	flg_p->intr_mask &= ~0x02;	/* inhibit further interrupts */
	return 0;
}

/*============================================================================
 * Connection Status Interrupt Handler.
 * ---------------------------------------------------------------------------
 */
static int intr_connect (spppData_t* ifd_p)
{
	pppFlags_t* flg_p = (void*)(ifd_p->ads->vbase + ifd_p->flg_offs);

	if (SDLA_DEBUG > 1) printk(
		"%s: Link goes down on interface %s! Cause: 0x%02X\n",
		mod_name, ifd_p->name, flg_p->disc_cause)
	;
	return 0;
}

/****** SDLA Firmware-Specific Functiond ************************************/

/*============================================================================
 * Read Firmware Code Version.
 * ---------------------------------------------------------------------------
 * Put code version as ASCII string in str. 
 */
static int pppCodeVersion (spppData_t* ifd_p, char* str)
{
  sdlaData_t*	ads = ifd_p->ads;
  pppCmd_t*	cmd_p = (void*)(ads->vbase + ifd_p->cmd_offs);
  int		cmdres, len;

  memclr(cmd_p, sizeof(pppCmd_t));
  cmd_p->command = PPP_READ_CODE_VERSION;
  cmdres = sdla_exec(ads, ifd_p->cmd_offs) ? cmd_p->result : CMD_TIMEOUT;
  if (cmdres != CMD_OK)
    pppEvent(ifd_p, cmdres, PPP_READ_CODE_VERSION)
  ;
  else if (str)
  {
    len = cmd_p->length;
    memcpy(str, cmd_p->data, len);
    str[len] = '\0';
  }
  return cmdres;
}

/*============================================================================
 * Configure PPP.
 * ---------------------------------------------------------------------------
 */
static int pppConfigure (spppData_t* ifd_p)
{
  sdlaData_t*	ads = ifd_p->ads;
  pppCmd_t*	cmd_p = (void*)(ads->vbase + ifd_p->cmd_offs);
  int		cmdres, buflen;
  unsigned char	buf[0x6F];

  /*
   * Prepare PPP configuration structure.
   */
  memclr(&buf, 0x6F);
  switch (ads->firmware_id)
  {
    case SFID_PPP502:
      buflen = 0x68;
      if (ifd_p->cfg.clocking)
        buf[0] = bps_to_line_speed(ifd_p->cfg.speed)
      ;
      *(unsigned short*)(&buf[0x01]) = 4;		/* num. of Tx bufs */
      *(unsigned short*)(&buf[0x05]) = ifd_p->dev.mtu;	/* local MTU/MRU */
      *(unsigned short*)(&buf[0x07]) = ifd_p->dev.mtu;	/* min. remote MRU */
      *(unsigned short*)(&buf[0x09]) = ifd_p->cfg.restart_tmr;
      *(unsigned short*)(&buf[0x0B]) = ifd_p->cfg.auth_rsrt_tmr;
      *(unsigned short*)(&buf[0x0D]) = ifd_p->cfg.auth_wait_tmr;
      *(unsigned short*)(&buf[0x0F]) = ifd_p->cfg.mdm_fail_tmr;
      *(unsigned short*)(&buf[0x11]) = ifd_p->cfg.dtr_drop_tmr;
      *(unsigned short*)(&buf[0x13]) = ifd_p->cfg.connect_tmout;
      *(unsigned short*)(&buf[0x15]) = ifd_p->cfg.conf_retry;
      *(unsigned short*)(&buf[0x17]) = ifd_p->cfg.term_retry;
      *(unsigned short*)(&buf[0x19]) = ifd_p->cfg.fail_retry;
      *(unsigned short*)(&buf[0x1B]) = ifd_p->cfg.auth_retry;
      buf[0x1E] = 0x81;		/* IP options: enable, local IP assignment */
      *(unsigned long*)(&buf[0x1F])  = ifd_p->dev.pa_addr;
      *(unsigned long*)(&buf[0x23])  = ifd_p->dev.pa_dstaddr;
      break;

    case SFID_PPP508:
    default:
      buflen = 0x6F;
      if (ifd_p->cfg.clocking)
        *(unsigned long*)(&buf[0])   = ifd_p->cfg.speed	/* line speed */
      ;
      *(unsigned short*)(&buf[0x04]) = 60;		/* % of Tx bufs */
      if (ifd_p->cfg.interface == 0)
        *(unsigned short*)(&buf[0x06]) |= 0x0020;	/* RS-232 mode */
      ;
      *(unsigned short*)(&buf[0x08]) = ifd_p->dev.mtu;	/* local MTU/MRU */
      *(unsigned short*)(&buf[0x0A]) = ifd_p->dev.mtu;	/* min. remote MRU */
      *(unsigned short*)(&buf[0x0C]) = ifd_p->cfg.restart_tmr;
      *(unsigned short*)(&buf[0x0E]) = ifd_p->cfg.auth_rsrt_tmr;
      *(unsigned short*)(&buf[0x10]) = ifd_p->cfg.auth_wait_tmr;
      *(unsigned short*)(&buf[0x12]) = ifd_p->cfg.mdm_fail_tmr;
      *(unsigned short*)(&buf[0x14]) = ifd_p->cfg.dtr_drop_tmr;
      *(unsigned short*)(&buf[0x16]) = ifd_p->cfg.connect_tmout;
      *(unsigned short*)(&buf[0x18]) = ifd_p->cfg.conf_retry;
      *(unsigned short*)(&buf[0x1A]) = ifd_p->cfg.term_retry;
      *(unsigned short*)(&buf[0x1C]) = ifd_p->cfg.fail_retry;
      *(unsigned short*)(&buf[0x1E]) = ifd_p->cfg.auth_retry;
      buf[0x21] = 0x81;		/* IP options: enable, local IP assignment */
      *(unsigned long*)(&buf[0x22])  = ifd_p->dev.pa_addr;
      *(unsigned long*)(&buf[0x26])  = ifd_p->dev.pa_dstaddr;
      break;
  }

  /*
   * Issue command to the board.
   */
  memclr(cmd_p, sizeof(pppCmd_t));
  memcpy(cmd_p->data, buf, buflen);
  cmd_p->command = PPP_SET_CONFIG;
  cmd_p->length  = buflen;
  cmdres = sdla_exec(ads, ifd_p->cmd_offs) ? cmd_p->result : CMD_TIMEOUT;
  if (cmdres != CMD_OK)
    pppEvent(ifd_p, cmdres, PPP_SET_CONFIG)
  ;
  return cmdres;
}

/*============================================================================
 * Enable Communications.
 * ---------------------------------------------------------------------------
 */
static int pppCommEnable (spppData_t* ifd_p)
{
  sdlaData_t*	ads = ifd_p->ads;
  pppCmd_t*	cmd_p = (void*)(ads->vbase + ifd_p->cmd_offs);
  int		cmdres;

  memclr(cmd_p, sizeof(pppCmd_t));
  cmd_p->command = PPP_COMM_ENABLE;
  cmdres = sdla_exec(ads, ifd_p->cmd_offs) ? cmd_p->result : CMD_TIMEOUT;
  if (cmdres != CMD_OK)
    pppEvent(ifd_p, cmdres, PPP_COMM_ENABLE)
  ;
  return cmdres;
}

/*============================================================================
 * Disable Communications.
 * ---------------------------------------------------------------------------
 */
static int pppCommDisable (spppData_t* ifd_p)
{
  sdlaData_t*	ads = ifd_p->ads;
  pppCmd_t*	cmd_p = (void *)(ads->vbase + ifd_p->cmd_offs);
  int		cmdres;

  memclr(cmd_p, sizeof(pppCmd_t));
  cmd_p->command = PPP_COMM_DISABLE;
  cmdres = sdla_exec(ads, ifd_p->cmd_offs) ? cmd_p->result : CMD_TIMEOUT;
  if (cmdres != CMD_OK)
    pppEvent(ifd_p, cmdres, PPP_COMM_DISABLE)
  ;
  return cmdres;
}

/*============================================================================
 * Set Interrupt Mode.
 * ---------------------------------------------------------------------------
 */
static int pppSetIntrMode (spppData_t* ifd_p, int mode)
{
  sdlaData_t*	ads = ifd_p->ads;
  pppCmd_t*	cmd_p = (void *)(ads->vbase + ifd_p->cmd_offs);
  int		cmdres;

  memclr(cmd_p, sizeof(pppCmd_t));
  switch (ads->firmware_id)
  {
    case SFID_PPP502:
      cmd_p->data[0] = mode;
      cmd_p->length  = 1;
      break;

    case SFID_PPP508:
    default:
      cmd_p->data[0] = mode;
      cmd_p->data[1] = ads->irq;
      cmd_p->length = 2;
  }
  cmd_p->command = PPP_SET_INTR_FLAGS;
  cmdres = sdla_exec(ads, ifd_p->cmd_offs) ? cmd_p->result : CMD_TIMEOUT;
  if (cmdres != CMD_OK)
    pppEvent(ifd_p, cmdres, PPP_SET_INTR_FLAGS)
  ;
  return cmdres;
}

/*============================================================================
 * Send Data.
 * ---------------------------------------------------------------------------
 */
static int pppSend (spppData_t* ifd_p, void* buf, int len, int proto)
{
	sdlaData_t* ads = ifd_p->ads;
	pppBufCtl_t* bcb_p = (void*)(ads->vbase + ifd_p->txc_offs +
				     ifd_p->tx_buf_idx * 16);
	unsigned long bufaddr, flags;

	if (bcb_p->flag) return 1;	/* no transmit buffers available */

	switch (ads->firmware_id)
	{
	case SFID_PPP502:
		bufaddr = (bcb_p->buf.o_p[1] << 8) + bcb_p->buf.o_p[0]; 
		break;

	case SFID_PPP508:
	default:
		bufaddr = bcb_p->buf.ptr;
	}
	save_flags(flags);
	cli();
	sdla_poke(ads, bufaddr, buf, len);
	restore_flags(flags);

	bcb_p->length = len;		/* frame length */
	bcb_p->proto  = proto;		/* protocol */
	bcb_p->flag   = 0x01;		/* request transmission */

	/*
	 * Update transmit buffer control fields.
	 */
	if (++ifd_p->tx_buf_idx >= ifd_p->tx_buf_num)
		ifd_p->tx_buf_idx = 0
	;
	return 0;
}

/*============================================================================
 * Receive Data.
 * ---------------------------------------------------------------------------
 */
static int pppReceive (spppData_t* ifd_p, void* buf, int* len_p, int* proto_p)
{
	sdlaData_t* ads = ifd_p->ads;
	pppBufCtl_t* bcb_p = (void*)(ads->vbase + ifd_p->rxc_offs +
				     ifd_p->rx_buf_idx * 16);
	int cmdres = CMD_OK;
	int len;
	unsigned long bufaddr;

	if (bcb_p->flag != 0x01)
	{
		if (SDLA_DEBUG > 1) printk(
			"%s: interface %s receive buffers corrupted!.\n",
			mod_name, ifd_p->name)
		;
		return 1;	/* data not available */
	}

	len = bcb_p->length;
	if (len > ifd_p->dev.mtu)
	{
		if (SDLA_DEBUG > 1) printk(
			"%s: interface %s received frame too long!.\n",
			mod_name, ifd_p->name)
		;
		cmdres = 2;	/* frame too long */
	}
	else
	{
		*len_p   = len;
		*proto_p = bcb_p->proto;
		switch (ads->firmware_id)
		{
		case SFID_PPP502:
			bufaddr = (bcb_p->buf.o_p[1] << 8) + bcb_p->buf.o_p[0]; 
			sdla_peek(ads, bufaddr, buf, len);
			bcb_p->flag = 0xFF;	/* mark buffer as processed */
			break;

	        case SFID_PPP508:
	        default:
			bufaddr = bcb_p->buf.ptr;
			if ((bufaddr + len) > ifd_p->rx_buf_top + 1)
			{
				/*
				 * Handle split frame.
				 */
				int tmp = ifd_p->rx_buf_top - bufaddr + 1;

				sdla_peek(ads, bufaddr, buf, tmp);
				sdla_peek(
					ads,
					ifd_p->rx_buf_base,
					(char*)buf + tmp,
					len - tmp)
				;
			}
			else sdla_peek(ads, bufaddr, buf, len);
			bcb_p->flag = 0x00;	/* mark buffer as processed */
		}
	}

	/*
	 * Update receive buffer control fields.
	 */
	if (++ifd_p->rx_buf_idx >= ifd_p->rx_buf_num)
		ifd_p->rx_buf_idx = 0
	;
	return cmdres;
}

/*============================================================================
 * Read Communications Statistics.
 * ---------------------------------------------------------------------------
 */
static int pppCommStats (spppData_t* ifd_p, spppStat_t* stat_p)
{
  sdlaData_t*	ads	= ifd_p->ads;
  pppCmd_t*	cmd_p	= (void *)(ads->vbase + ifd_p->cmd_offs);
  pppFlags_t*	flg_p	= (void *)(ads->vbase + ifd_p->flg_offs);
  int		cmdres;

  /*
   * Read link/modem status.
   */
  stat_p->link_status = 0;
  if (flg_p->mdm_status & PPP_MDM_DCD)
    stat_p->link_status |= SPPP_MODEM_DCD
  ;
  if (flg_p->mdm_status & PPP_MDM_CTS)
    stat_p->link_status |= SPPP_MODEM_CTS
  ;
  if (flg_p->lcp_state == 9)
    stat_p->link_status |= SPPP_LINK_UP
  ;

  /*
   * Read operational statistics.
   */
  memclr(cmd_p, sizeof(pppCmd_t));
  cmd_p->command = PPP_READ_STATISTICS;
  cmdres = sdla_exec(ads, ifd_p->cmd_offs) ? cmd_p->result : CMD_TIMEOUT;
  if (cmdres == CMD_OK)
  {
    stat_p->rx_byte_cnt = *(unsigned long*)(&cmd_p->data[0x12]);
    stat_p->rx_frm_cnt  = *(unsigned long*)(&cmd_p->data[0x0E]);
    stat_p->tx_byte_cnt = *(unsigned long*)(&cmd_p->data[0x0A]);
    stat_p->tx_frm_cnt  = *(unsigned long*)(&cmd_p->data[0x06]);
    stat_p->tx_retries  = *(unsigned short*)(&cmd_p->data[0x04]);

    switch (ads->firmware_id)
    {
      case SFID_PPP502:
        stat_p->rx_dropped = *(unsigned short*)(&cmd_p->data[0x00]) +
                             *(unsigned short*)(&cmd_p->data[0x02])
        ;
        break;

      case SFID_PPP508:
        stat_p->rx_bad_len = *(unsigned short*)(&cmd_p->data[0x02]);
      default:
    }
  }
  else pppEvent(ifd_p, cmdres, PPP_READ_STATISTICS);

  /*
   * Read communications errors statistics.
   */
  memclr(cmd_p, sizeof(pppCmd_t));
  cmd_p->command = PPP_READ_ERROR_STATS;
  cmdres = sdla_exec(ads, ifd_p->cmd_offs) ? cmd_p->result : CMD_TIMEOUT;
  if (cmdres == CMD_OK)
  {
    stat_p->rx_overruns  = cmd_p->data[0x00];
    stat_p->rx_crc_err   = cmd_p->data[0x01];
    stat_p->rx_aborts    = cmd_p->data[0x02];
    stat_p->tx_aborts    = cmd_p->data[0x04];
    stat_p->tx_underruns = cmd_p->data[0x05];
    stat_p->dcd_change   = cmd_p->data[0x08];
    stat_p->cts_change   = cmd_p->data[0x09];

    switch (ads->firmware_id)
    {
      case SFID_PPP502:
        stat_p->rx_bad_len = cmd_p->data[0x03];
        break;

      case SFID_PPP508:
      default:
        stat_p->rx_dropped = cmd_p->data[0x03];
    }
  }
  else pppEvent(ifd_p, cmdres, PPP_READ_ERROR_STATS);
  return cmdres;
}

/*============================================================================
 * Firmware asynchronous event/error handler.
 * ---------------------------------------------------------------------------
 * Return zero if previous command has to be cancelled.
 */
static int pppEvent (spppData_t* ifd_p, int event, int cmd)
{
	int retry = 0;

	if (SDLA_DEBUG > 1)
		printk("%s: interface %s: command 0x%02X returned 0x%02X.\n",
			mod_name, ifd_p->name, cmd, event)
	;
	return retry;
} /* pppEvent() */

/****** Miscellaneous Functions *********************************************/

/*============================================================================
 * Convert line speed in bps to a number used by S502 code.
 * ---------------------------------------------------------------------------
 */
static unsigned char bps_to_line_speed (unsigned long bps)
{
	unsigned char	number;

	if (bps <= 1200)        number = 0x01 ;
	else if (bps <= 2400)   number = 0x02;
	else if (bps <= 4800)   number = 0x03;
	else if (bps <= 9600)   number = 0x04;
	else if (bps <= 19200)  number = 0x05;
	else if (bps <= 38400)  number = 0x06;
	else if (bps <= 45000)  number = 0x07;
	else if (bps <= 56000)  number = 0x08;
	else if (bps <= 64000)  number = 0x09;
	else if (bps <= 74000)  number = 0x0A;
	else if (bps <= 112000) number = 0x0B;
	else if (bps <= 128000) number = 0x0C;
	else number = 0x0D;

	return number;
}

/****** End *****************************************************************/
