/*****************************************************************************
* PROJECT:	flipcfg		Sangoma Frame Relay Driver Configurator
* FILE:		flipcfg.c	Main
* TARGET OS:	UNIX
* AUTHOR:	Gene Kozin
* ----------------------------------------------------------------------------
* (c) Copyright 1995-1996 Sangoma Technologies Inc.  All Rights Reserved.
* ============================================================================
*  Rev	Date		Who	Details
* -----	-----------	---	----------------------------------------------
*  2.00	Apr.30,1996	GK	Added some frame relay-level statistics and
*				status info.
*  1.10	Mar.11,1996	GK	Requires FLIP Version 1.
*  1.00	Dec.15,1995	GK	Beta version.
*  0.10	Nov.22,1995	GK	Uses socket-style IOCTL to access FLIP master
*				control device.
*  0.01	Sep.21,1995	GK	Alpha release.
*****************************************************************************/

/*****************************************************************************
* Usage:
*   flipcfg [{switch} ...] {command} [{argument} ...]
*
* Where:
*   {switch}	Optional switches:
*		 -v		verbose mode
*		 -h or -?	display help screen
*   {command}	Tells what to do:
*		 c {port} {config_file}	- configure Frame Relay port
*		 r {port}		- reset Frame Relay port
*		 s {port}		- get Frame Realy port status
*		 a {ifno} {port} {dlci}	- add network interface
*		 d {ifno}		- delete network interface
*		 i {ifno}		- get interface status
*   {argument}	Command-specific arguments:
*		 {port}	- SDLA port number 0 .. N (decimal)
*		 {dlci} - DLCI number (decimal)
*		 {ifno} - interface number (decimal)
*****************************************************************************/

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <flipio.h>		/* FLIP driver API */

/****** Defines *************************************************************/

#ifndef	min
#define	min(a,b)	(((a)<(b))?(a):(b))
#endif

#define	PROG_NAME	"flipcfg"
#define	DEV_NAME	"flip"

/*
 *----- Error Codes ----------------------------------------------------------
 */
#define	ERR_SYSTEM	1	/* System error */
#define	ERR_SWITCH	2	/* Invalid command line switch */
#define ERR_SYNTAX	3	/* Invalid command line syntax */

#define	ERR_DRIVER	10	/* General driver failure */
#define	ERR_CFG_SYNTAX	11	/* Configuration file syntax error */

/*
 *----- Command Line Parsing Stuff -------------------------------------------
 */
#define ARG_SWITCH	'-'	/* switch follows */
#define	ARG_CMDFILE	'@'	/* command file follows */

#define	SW_GETHELP	'h'	/* 'get help' switch */
#define	SW_ALTHELP	'?'	/* same */
#define	SW_VERBOSE	'v'	/* enable verbose output */

/*
 *----- Commands -------------------------------------------------------------
 */
#define	CMD_CONFIG	'c'	/* configure port */
#define	CMD_RESET	'r'	/* reset port */
#define	CMD_STATUS	's'	/* get port status */
#define	CMD_ADD		'a'	/* add network interface */
#define	CMD_DELETE	'd'	/* delete network interface */
#define	CMD_IFSTATUS	'i'	/* get interface status */

#define	MAX_ARG		3	/* maximum number of arguments */

/*
 *----- Configuration Stuff --------------------------------------------------
 */
#define	MAX_CFGLINE	256
#define	MAX_CMDLINE	256
#define	MAX_TOKENS	32

/*
 *----- Miscellaneous --------------------------------------------------------
 */
#define	FLAG_VERBOSE	0x0001	/* enable verbose output */
#define	FLAG_HELP	0x0002	/* show help screen */

/****** Data Types **********************************************************/

typedef struct	LookUp			/* Look-up Table entry */
{
  int	code;				/* look-up code */
  void	*ptr;				/* pointer */
} LookUp_t;

typedef struct SymName			/* Symbolic Name structure */
{
  char	*str;
  int	val;
} SymName_t;

typedef struct CfgRec			/* Configuration Record */
{
  char	*keyword;			/* keyword */
  int	argtype;			/* type of argument */
  int	offs;				/* offset into configuration struct.*/
  void	*argvals;			/* transcription data */
} CfgRec_t;

typedef int	CmdProc	(int argc, char *argv[]);

/*----- Argument types -----------------------------------------------------*/
#define	ARGT_SYMNAME	1		/* symbolic name */
#define	ARGT_NUMERIC	2		/* number */
#define	ARGT_NUMLIST	3		/* list of numbers */

/****** Function Prototypes *************************************************/

/*
 * Command handlers.
 */
int	cmd_dummy	(int argc, char *argv[]);
int	cmd_config	(int argc, char *argv[]);
int	cmd_reset	(int argc, char *argv[]);
int	cmd_add		(int argc, char *argv[]);
int	cmd_delete	(int argc, char *argv[]);
int	cmd_status	(int argc, char *argv[]);
int	cmd_ifstatus	(int argc, char *argv[]);

/*
 * Support functions.
 */
int	switch_proc	(char *str);
int	cmdfile_proc	(char *name);
void	show_help	(void);
void	show_error	(int err);
int	parse_cfg	(char *cfgline, CfgRec_t *cfgtable, void *cfgdata);
void	*lookup		(int code, LookUp_t *table);
int	tokenize	(char *str, char *space, char *eol, char **tokens);


/****** Global Data *********************************************************/

/*
 *----- Messages -------------------------------------------------------------
 */
char banner[]	=
	"FLIP Configurator. v2.00 (c) 1995 Sangoma Technologies Inc."
;
char helptext[] =
	"Usage:\n"
	"  flipcfg [sw] c {port} {cfg_file}  Configure Frame Relay port\n"
	"  flipcfg [sw] r {port}             Reset Frame Relay port\n"
	"  flipcfg [sw] s {port}             Get Frame Relay port status\n"
	"  flipcfg [sw] a {if} {port} {dlci} Add network interface\n"
	"  flipcfg [sw] d {if}               Delete network interface\n"
	"  flipcfg [sw] i {if}               Get network interface status\n"
	"  flipcfg [sw] @{cmd_file}          Process command file\n"
	"Where:\n"
	"  [sw]        optional switches:\n"
	"               -h or -?   display help screen\n"
	"               -v         enable verbose output\n"
	"  {port}      SDLA port number 0..N (decimal)\n"
	"  {dlci}      DLCI number (decimal)\n"
	"  {if}        interface number (decimal)\n"
	"  {cfg_file}  configuration file name\n"
	"  {cmd_file}  command file\n"
;

LookUp_t	err_table[] =		/* Error messages */
{
  { ERR_SWITCH,		"Invalid command line switch"		},
  { ERR_SYNTAX,		"Invalid command line syntax"		},
  { ERR_CFG_SYNTAX,	"Configuration file syntax error"	},
  { 0,			"Unknown error code"			}, /* EOT */
};

LookUp_t	drverr_table[] =	/* Driver error messages */
{
  { FLIPER_BADCMD,	"Invalid FLIP IOCTL command"		},
  { FLIPER_BADARG,	"Invalid FLIP IOCTL argument"		},
  { FLIPER_BADVER,	"Incompatible driver version"		},
  { FLIPER_BADSTATE,	"Invalid device state"			},
  { 0,			"FLIP IOCTL failure"			}, /* EOT */
};

/*
 *----- Miscellaneous --------------------------------------------------------
 */
char		progname[] = PROG_NAME;
char		devname[] = DEV_NAME;
int		flags;
int		syserr;			/* system error number */
int		drverr;			/* device error code */
int		sock;			/* FLIP master control socket */
struct ifreq	ifr;

flipCtl_t	flip_cb =	/* FLIP control block */
{
  FLIP_SIGNATURE,
  FLIP_VERSION,
  0,
  0,
  {
    { 0, }
  }
};

LookUp_t	cmd_table[] =		/* Command handlers */
{
  { CMD_CONFIG,		cmd_config	},
  { CMD_RESET,		cmd_reset	},
  { CMD_ADD,		cmd_add		},
  { CMD_DELETE,		cmd_delete	},
  { CMD_STATUS,		cmd_status	},
  { CMD_IFSTATUS,	cmd_ifstatus	},
  { 0,			cmd_dummy	}, /* EOT */
};

/*
 *----- Configuration Data ---------------------------------------------------
 */
SymName_t	yesno_opt[] =
{
  { "Yes",	1	}, 
  { "No",	0	}, 
  { NULL,	0	}, 
};

SymName_t	onoff_opt[] =
{
  { "On",	1	}, 
  { "Off",	0	}, 
  { NULL,	0	}, 
};

SymName_t	port_opt[] =
{
  { "RS232",	1	},
  { "V35",	0	},
  { NULL,	0	}, 
};

SymName_t	clock_opt[] =
{
  { "Internal",	1	}, 
  { "External",	0	}, 
  { NULL,	0	}, 
};

SymName_t	station_opt[] =
{
  { "Node",	1	}, 
  { "CPE",	0	}, 
  { NULL,	0	}, 
};

SymName_t	lmi_opt[] =
{
  { "None",	3	}, 
  { "Q933",	2	}, 
  { "LMI",	1	}, 
  { "AnnexD",	0	}, 
  { NULL,	0	}, 
};

CfgRec_t	cfgtable[] =
{
  { "Station",	ARGT_SYMNAME, offsetof(flipPortCfg_t, station), station_opt },
  { "Clocking",	ARGT_SYMNAME, offsetof(flipPortCfg_t, clocking), clock_opt },
  { "PortType",	ARGT_SYMNAME, offsetof(flipPortCfg_t, type), port_opt },
  { "Signalling", ARGT_SYMNAME, offsetof(flipPortCfg_t, signalling), lmi_opt },
  { "BaudRate",	ARGT_NUMERIC, offsetof(flipPortCfg_t, speed),	0	 },
  { "CIR",	ARGT_NUMERIC, offsetof(flipPortCfg_t, CIR),	0	 },
  { "MTU",	ARGT_NUMERIC, offsetof(flipPortCfg_t, mtu),	0	 },
  { "T391",	ARGT_NUMERIC, offsetof(flipPortCfg_t, T391),	0	 },
  { "T392",	ARGT_NUMERIC, offsetof(flipPortCfg_t, T392),	0	 },
  { "N391",	ARGT_NUMERIC, offsetof(flipPortCfg_t, N391),	0	 },
  { "N392",	ARGT_NUMERIC, offsetof(flipPortCfg_t, N392),	0	 },
  { "N393",	ARGT_NUMERIC, offsetof(flipPortCfg_t, N393),	0	 },
  { "DLCI",	ARGT_NUMERIC, offsetof(flipPortCfg_t, dlci),	0	 },
  { NULL,	0, 0, 0 }
};

char	cfgspaces[]	= "\t =:";
char	cfgeol[]	= "\n\r#";

/****** Entry Point *********************************************************/

int main (int argc, char *argv[])
{
  int	err = 0;	/* return code */

  /*
   * Process command line switches.
   */
  while (!err && --argc && (**++argv == ARG_SWITCH))
    err = switch_proc(*argv)
  ;

  /*
   * Display banner.
   */
  if (flags & FLAG_VERBOSE) puts(banner);

  /*
   * Perform requested action.
   */
  if (!err)
  {
    if (flags & FLAG_HELP) show_help();
    else if (argc == 0) err = ERR_SYNTAX;
    else
    {
      /*
       * Create a socket and initialize its IOCTL structure.
       */
      sock = socket(AF_INET, SOCK_STREAM, 0);
      if (sock <= 0)
      {
        err = ERR_SYSTEM;
        syserr = errno;
      }
      else
      {
        strcpy(ifr.ifr_ifrn.ifrn_name, devname);
        ifr.ifr_ifru.ifru_data = (void *)&flip_cb;
        if (**argv == ARG_CMDFILE)
        {
          /*
           * Process command file.
           */
          err = cmdfile_proc(&(*argv)[1]);
        }
        else err = ((CmdProc*)lookup(**argv, cmd_table))(--argc, ++argv);
        close(sock);
      }
    }
  }
  if (err) show_error(err);	/* display error message */
  return err;
} /* main() */

/****** Command Handlers ****************************************************/

/*============================================================================
 * Dummy Command Handler
 * ---------------------------------------------------------------------------
 */
int  cmd_dummy (int argc, char *argv[])
{
  return ERR_SYNTAX;
} /* cmd_dummy() */

/*============================================================================
 * CMD_CONFIG Handler
 * ---------------------------------------------------------------------------
 */
int  cmd_config (int argc, char *argv[])
{
  int	err = 0;		/* error code */
  int	port;
  FILE	*cfgfile;
  char	cfgline[MAX_CFGLINE];	/* line buffer */

  if ((argc < 2) || !isdigit(**argv))
    err = ERR_SYNTAX
  ;
  else
  {
    port = atoi(argv[0]);
    if (flags & FLAG_VERBOSE)
    {
      printf(
	"Configuring Frame Relay port %d using configuration file %s ...\n",
	port, argv[1])
      ;
    }
    cfgfile = fopen(argv[1], "r");
    if (cfgfile == NULL)
    {
      err = ERR_SYSTEM;
      syserr = errno;
    }
    else
    {
      memset(&flip_cb.u.portcfg, 0, sizeof(flipPortCfg_t));
      flip_cb.u.portcfg.port	= port;
      while (fgets(cfgline, MAX_CFGLINE, cfgfile))
      {
        if (parse_cfg(cfgline, cfgtable, &flip_cb.u.portcfg))
        {
          err = ERR_CFG_SYNTAX;
          break;
        }
      }
      fclose(cfgfile);

      if (!err)
      {
        if (ioctl(sock, FLIPIO_CONFIG, &ifr) < 0)
        {
          if (errno == EIO)
          {
            err = ERR_DRIVER;
            drverr = flip_cb.result;
          }
          else
          {
            err = ERR_SYSTEM;
            syserr = errno;
          }
        }
      }
    }
  }
  return err;
} /* cmd_config() */

/*============================================================================
 * CMD_RESET Handler
 * ---------------------------------------------------------------------------
 */
int cmd_reset (int argc, char *argv[])
{
  int	err = 0;
  int	port;

  if ((argc < 1) || !isdigit(**argv))
    err = ERR_SYNTAX
  ;
  else
  {
    port = atoi(argv[0]);
    if (flags & FLAG_VERBOSE)
    {
      printf("Closing Frame Relay port %d ...\n", port);
    }
    flip_cb.u.port = port;
    if (ioctl(sock, FLIPIO_RESET, &ifr) < 0)
    {
      if (errno == EIO)
      {
        err = ERR_DRIVER;
        drverr = flip_cb.result;
      }
      else
      {
        err = ERR_SYSTEM;
        syserr = errno;
      }
    }
  }
  return err;
} /* cmd_reset() */

/*============================================================================
 * CMD_STATUS Handler
 * ---------------------------------------------------------------------------
 */
int cmd_status (int argc, char *argv[])
{
  int	err = 0;
  int	port, i;

  if ((argc < 1) || !isdigit(**argv))
    err = ERR_SYNTAX
  ;
  else
  {
    port = atoi(argv[0]);
    if (flags & FLAG_VERBOSE)
    {
      printf("Getting status for the Frame Relay port %d ...\n", port);
    }
    flip_cb.u.port = port;
    if (ioctl(sock, FLIPIO_STATUS, &ifr) < 0)
    {
      if (errno == EIO)
      {
        err = ERR_DRIVER;
        drverr = flip_cb.result;
      }
      else
      {
        err = ERR_SYSTEM;
        syserr = errno;
      }
    }
    else
    {
      printf("Link status ................... ");
      printf("DCD %s : ", (flip_cb.u.portsts.status & FLIP_MODEM_DCD) ?
             "UP" : "DOWN")
      ;
      printf("CTS %s : ", (flip_cb.u.portsts.status & FLIP_MODEM_CTS) ?
             "UP" : "DOWN")
      ;
      printf("LINK %s\n", (flip_cb.u.portsts.status & FLIP_LINK_UP) ?
             "UP" : "DOWN")
      ;
      printf("Number of active DLCIs ........ %d\n", flip_cb.u.portsts.dlci_num);
      if (flip_cb.u.portsts.dlci_num)
      {
        printf("List of active DLCI(s) ........");
        for (i = 0; i < flip_cb.u.portsts.dlci_num; ++i)
          printf(" %d", flip_cb.u.portsts.dlci[i])
        ;
        putchar('\n');
      }
      printf("Configured interfaces ......... %d\n", flip_cb.u.portsts.ifcnt);
      printf("Receive interrupts ............ %d\n", flip_cb.u.portsts.rxint_cnt);
      printf("Transmit interrupts ........... %d\n", flip_cb.u.portsts.txint_cnt);
      printf("Spurious interrupts ........... %d\n", flip_cb.u.portsts.spuri_cnt);
      printf("Receiver overrun errors ....... %d\n", flip_cb.u.portsts.rx_overruns);
      printf("Aborted frames received ....... %d\n", flip_cb.u.portsts.rx_aborts);
      printf("Too long frames received ...... %d\n", flip_cb.u.portsts.rx_too_long);
      printf("CRC errors .................... %d\n", flip_cb.u.portsts.rx_bad_crc);
      printf("Discarded frames .............. %d\n", flip_cb.u.portsts.rx_dropped);
      printf("Frames with invalid DLCI ...... %d\n", flip_cb.u.portsts.rx_bad_dlci);
      printf("Frames with invalid format .... %d\n", flip_cb.u.portsts.rx_bad_format);
      printf("Transmitter underrun errors ... %d\n", flip_cb.u.portsts.tx_underruns);
      printf("Aborted frames transmitted .... %d\n", flip_cb.u.portsts.tx_aborts);
      printf("Re-transmitted frames ......... %d\n", flip_cb.u.portsts.tx_retries);
      printf("Number of times DCD dropped ... %d\n", flip_cb.u.portsts.dcd_low_cnt);
      printf("Number of times CTS dropped ... %d\n", flip_cb.u.portsts.cts_low_cnt);
      printf("\n");
    }
  }
  return err;
} /* cmd_status() */

/*============================================================================
 * CMD_ADD Handler
 * ---------------------------------------------------------------------------
 */
int  cmd_add (int argc, char *argv[])
{
  int	err = 0;	/* error code */
  int	ifno, port, dlci;

  if ((argc < 3) ||
      !isdigit(*argv[0]) ||
      !isdigit(*argv[1]) ||
      !isdigit(*argv[2]))
    err = ERR_SYNTAX
  ;
  else
  {
    ifno = atoi(argv[0]);
    port = atoi(argv[1]);
    dlci = atoi(argv[2]);
    if (flags & FLAG_VERBOSE)
    {
      printf(
	"Creating interface %s%d for DLCI %d on Frame Relay port %d ...\n",
	devname, ifno, dlci, port)
      ;
    }
    flip_cb.u.ifcfg.ifno = ifno;
    flip_cb.u.ifcfg.port = port;
    flip_cb.u.ifcfg.dlci = dlci;
    if (ioctl(sock, FLIPIO_ADD, &ifr) < 0)
    {
      if (errno == EIO)
      {
        err = ERR_DRIVER;
        drverr = flip_cb.result;
      }
      else
      {
        err = ERR_SYSTEM;
        syserr = errno;
      }
    }
  }
  return err;
} /* cmd_add() */

/*============================================================================
 * CMD_DELETE Handler
 * ---------------------------------------------------------------------------
 */
int cmd_delete (int argc, char *argv[])
{
  int	err = 0;
  int	ifno;

  if ((argc < 1) || !isdigit(**argv))
    err = ERR_SYNTAX
  ;
  else
  {
    ifno = atoi(argv[0]);
    if (flags & FLAG_VERBOSE)
    {
      printf("Deleting interface %s%d ...\n", devname, ifno);
    }
    flip_cb.u.ifno = ifno;
    if (ioctl(sock, FLIPIO_DELETE, &ifr) < 0)
    {
      if (errno == EIO)
      {
        err = ERR_DRIVER;
        drverr = flip_cb.result;
      }
      else
      {
        err = ERR_SYSTEM;
        syserr = errno;
      }
    }
  }
  return err;
} /* cmd_delete() */

/*============================================================================
 * CMD_IFSTATUS Handler
 * ---------------------------------------------------------------------------
 */
int cmd_ifstatus (int argc, char *argv[])
{
  int	err = 0;
  int	ifno;

  if ((argc < 1) || !isdigit(**argv))
    err = ERR_SYNTAX
  ;
  else
  {
    ifno = atoi(argv[0]);
    if (flags & FLAG_VERBOSE)
    {
      printf("Getting status of the interface %s%d ...\n", devname, ifno);
    }
    flip_cb.u.ifno = ifno;
    if (ioctl(sock, FLIPIO_IFSTATUS, &ifr) < 0)
    {
      if (errno == EIO)
      {
        err = ERR_DRIVER;
        drverr = flip_cb.result;
      }
      else
      {
        err = ERR_SYSTEM;
        syserr = errno;
      }
    }
  }
  return err;
} /* cmd_ifstatus() */

/***** Helper Functions *****************************************************/

/*============================================================================
 * Process command line switches.
 * ---------------------------------------------------------------------------
 * Return 0 if processed o.k., otherwise return error code.
 */
int switch_proc (char *str)
{
  int	err = 0;

  switch (*++str)
  {
    case SW_GETHELP:
    case SW_ALTHELP:
      flags |= (FLAG_HELP | FLAG_VERBOSE);
      break;

    case SW_VERBOSE:
      flags |= FLAG_VERBOSE;
      break;

    default:
      err = ERR_SWITCH;
  }
  return err;
} /* switch_proc() */

/*============================================================================
 * Process command line switches.
 * ---------------------------------------------------------------------------
 * Return 0 if processed o.k., otherwise return error code.
 */
int cmdfile_proc (char *name)
{
  int	err = 0;
  FILE	*cmdfile;
  char	cmdline[MAX_CMDLINE];	/* command line buffer */
  char	*token[MAX_TOKENS];
  int	toknum;

  if (flags & FLAG_VERBOSE)
  {
    printf("Processing command file %s ...\n", name);
  }
  cmdfile = fopen(name, "r");
  if (cmdfile == NULL)
  {
    err = ERR_SYSTEM;
    syserr = errno;
  }
  else
  {
    while (!err && fgets(cmdline, MAX_CMDLINE, cmdfile))
    {
      toknum = tokenize(cmdline, " \t", "\r\n", token);
      if (toknum && (**token != '#'))
      {
        err = ((CmdProc*)lookup(**token, cmd_table))(toknum - 1, &token[1]);
      }
    }
    fclose(cmdfile);
  }
  return err;
} /* cmdfile_proc() */

/*============================================================================
 * Show help 
 * ---------------------------------------------------------------------------
 */
void show_help (void)
{
  puts(helptext);
} /* show_help() */

/*============================================================================
 * Show error message
 * ---------------------------------------------------------------------------
 */
void show_error (int err)
{
  char	*str;

  switch (err)
  {
    case ERR_SYSTEM:
      str = strerror(syserr);
      break;

    case ERR_DRIVER:
      str = lookup(drverr, drverr_table);
      break;

    default:
      str = lookup(err, err_table);
  }
  fprintf(stderr, "%s error %u: %s!\n", progname, err, str);
} /* show_error() */

/*============================================================================
 * Parse Configuration Line.
 * ---------------------------------------------------------------------------
 * Return zero if o.k., otherwise return:
 *	1 -  in case of syntax error
 */
int parse_cfg (char *cfgline, CfgRec_t *cfgtable, void *cfgdata)
{
  int		err	= 0;
  int		toknum;
  char		*token[MAX_TOKENS];
  SymName_t	*vals;

  toknum = tokenize(cfgline, cfgspaces, cfgeol, token);
  if (toknum > 1)
  {
    for (;
         cfgtable->keyword && strcmp(token[0], cfgtable->keyword);
         ++cfgtable)
    ;
    if (!cfgtable->keyword) err = 1;
    else switch (cfgtable->argtype)
    {
      case ARGT_SYMNAME:
        vals = cfgtable->argvals;
        for (;
             vals->str && strcmp(token[1], vals->str);
             ++vals)
        ;
        if (!vals->str) err = 1;
        else
        {
          *(int *)((char *)cfgdata + cfgtable->offs) = vals->val;
        }
        break;

      case ARGT_NUMERIC:
        *(int *)((char *)cfgdata + cfgtable->offs) =
          strtoul(token[1], NULL, 10)
        ;
        break;
    }
  }
  return err;
} /* parse_cfg() */

/*============================================================================
 * Get a pointer from a look-up table 
 * ---------------------------------------------------------------------------
 */
void *lookup (int code, LookUp_t *table)
{
  while (table->code && (table->code != code)) table++;
  return table->ptr;
} /* lookup() */

/*============================================================================
 * Tokenize string.
 * ---------------------------------------------------------------------------
 * Return number of tokens
 */
int tokenize (char *str, char *space, char *eol, char **tokens)
{
  int	toknum	= 0;
  int	toklen;

  str[strcspn(str, eol)] = '\0';		/* terminate string with 0 */
  while (*str && (toknum < MAX_TOKENS))
  {
    while (*str && strchr(space, *str)) ++str;	/* skip leading spaces */
    toklen = strcspn(str, space);
    if (toklen)
    {
      tokens[toknum] = str;
      ++toknum;
      str += toklen;
      if (*str) *str++ = '\0';
    }
  }
  return toknum;
} /* tokenize() */

/****** End *****************************************************************/

