/*****************************************************************************
* PROJECT:	sdlald		Sangoma Synchronous Data Link Adapter Loader
* FILE:		sdlald.c	Main
* TARGET OS:	Unix, Linux
* AUTHOR:	Gene Kozin
* ----------------------------------------------------------------------------
* (c) Copyright 1995-1996 Sangoma Technologies Inc.  All Rights Reserved.
* ============================================================================
*  Rev	Date		Who	Details
* -----	-----------	---	----------------------------------------------
*  3.10	Nov.18,1996	GK	o Uses SDLA driver v4.x
*				o Added default configuration for X.25
*  3.00	May 29,1996	GK	o Uses SDLA driver v3.x
*				o Added 'bind' and 'unbind' functions
*				o Command line option changed
*  2.03	May 07,1996	GK	o Added default boot-time firmware
*				  configuration routine
*  2.02	Apr.30,1996	GK	o Status output format changed
*				o Miscellaneous bux fixes
*  2.01	Apr.08,1996	GK	o Uses CCITT polynomial CRC calculation in
*				  checksum() function
*  2.00	Mar.11,1996	GK	o Uses SFM format version 2
*				o Does not use _IOxx macros
*  1.01	Dec.15,1995	GK	o Added a check for adapter type
*				o Changed ConfigProc definition 
*				o Fixed some bugs in action_load() routine
*  1.00	Dec.04,1995	GK	o Initial version
*****************************************************************************/

/*****************************************************************************
* Usage:
*   sdlald [-v] {device} {firmware} [{config}]	Load adapter firmware
*   sdlald -b {device} {module}			Bind to protocol module
*   sdlald -c {device} [{options}]		Configure adapter (h/w setup)
*   sdlald -d {device} [{offset} [{length}]]	Dump adapter memory
*   sdlald -f {firmware}			Show SFM header info
*   sdlald -h | -?				Display help screen
*   sdlald -s {device}				Show adapter status
*   sdlald -u {device}				Unbind from protocol module
*   sdlald -r {device}				Reset adapter
*   sdlald -z {device}				Shut down adapter (Zap)
*
* Where:
*   {device}	SDLA device node (e.g. /dev/sdla0)
*   {firmware}	SDLA firmware module name (SFM-file)
*   {config}	Configuration file name
*   {options}	Hardware configuration options:
*		 p=port		set SDLA base I/O port address (hex)
*		 m=addr		set SDLA base memory window address (hex)
*		 i=irq		set SDLA IRQ level (decimal)
*		 z=clock	set SDLA clock rate (KHz, decimal)
*
* Other switches are:
*   -v		Verbose mode
*
* Note, that switches -c, -d, -s, -u, -m and -h (-?) are mutually exclusive -
* only the last one will take an effect.
*****************************************************************************/

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sdlaio.h>		/* SDLA driver API */
#include <sfmdef.h>		/* SDLA Firmware Module definitions */

/****** Defines *************************************************************/

#define	PROG_NAME	"sdlald"

#ifndef	min
#define	min(a,b)	(((a)<(b))?(a):(b))
#endif	/* min */

enum ActionCodes		/*----- Action Codes ----------------------*/
{
	ACTION_HELP,		/* get help */
	ACTION_LOAD,		/* load adapter */
	ACTION_STATUS,		/* show adapter status */
	ACTION_SETUP,		/* setup (configure) adapter */
	ACTION_RESET,		/* reset adapter */
	ACTION_DOWN,		/* shutdown adapter */
	ACTION_SHOW_SFM,	/* show SFM header */
	ACTION_DUMP,		/* adapter memory dump */
	ACTION_BIND,		/* bind to protocol-specific module */
	ACTION_UNBIND,		/* unbind from protocol-specific module */
};

enum ErrorCodes			/*----- Error Codes ------------------------*/
{
	ERR_SYSTEM = 1,		/* System error */
	ERR_SWITCH,		/* Invalid command line switch */
	ERR_SYNTAX,		/* Invalid command line syntax */
	ERR_MALLOC,		/* Cannot allocate memory */
	ERR_DRIVER,		/* Driver request failure */
	ERR_BAD_DRIVER,		/* Not an SDLA driver */
	ERR_BAD_STATE,		/* Invalid adapter state */
	ERR_HW_COMPAT,		/* Incompatible hardware */
	ERR_NOT_SFM,		/* Not an SFM file */
	ERR_SFM_VERSION,	/* Infalid SFM file version */
	ERR_BAD_SFM,		/* Firmware module corrupted */
	ERR_CFG_UNKNOWN,	/* Unknown firmware type */
	ERR_BAD_OPTION,		/* Invalid option */
	ERR_LIMIT
};

/*
 * Command Line Parsing Stuff.
 */
#define ARG_SWITCH	'-'	/* command line switch indicator */
#define	SW_CONFIGURE	'c'	/* configure adapter */
#define	SW_BIND		'b'	/* bind to protocol-specific module */
#define	SW_UNBIND	'u'	/* unbind from protocol-specific module */
#define	SW_STATUS	's'	/* show adapter status */
#define	SW_SHUTDOWN	'z'	/* shut down adapter */
#define	SW_RESET	'r'	/* reset adapter */
#define	SW_DUMP		'd'	/* dump adapter memory */
#define	SW_SHOW_SFM	'f'	/* display SFM file header */
#define	SW_GETHELP	'h'	/* help */
#define	SW_ALTHELP	'?'	/* same */
#define	SW_VERBOSE	'v'	/* verbose mode */

#define	OPT_SEPARATOR	'='
#define	OPT_SDLA_TYPE	't'	/* adapter type */
#define OPT_IOPORT	'p'	/* I/O port option */
#define OPT_ADDRESS	'm'	/* DPM base option */
#define	OPT_IRQ		'i'	/* IRQ option */
#define	OPT_CLOCK	'z'	/* CPU clock rate option */

/*
 * Execution Option Flags.
 */
#define	FLAG_VERBOSE	0x0001	/* enable verbose output mode */

/****** Data Types **********************************************************/

typedef int ActionProc	(int argc, char* argv[]); /* action handler */

typedef struct	LookUp		/* Look-up Table Entry */
{
	int code;		/* look-up code */
	void* ptr;		/* pointer */
} LookUp_t;

/****** Function Prototypes *************************************************/

/*
 * Action handlers.
 */
int action_load		(int argc, char* argv[]);
int action_bind		(int argc, char* argv[]);
int action_unbind	(int argc, char* argv[]);
int action_status	(int argc, char* argv[]);
int action_setup	(int argc, char* argv[]);
int action_reset	(int argc, char* argv[]);
int action_down		(int argc, char* argv[]);
int action_dump		(int argc, char* argv[]);
int action_show_sfm	(int argc, char* argv[]);

/*
 * Helper functions.
 */
int switch_proc	(char* str);
int parse_opt	(char* str, sdlaSetup_t* setup_p);
void show_error	(int err);
void show_status(sdlaStatus_t* status);
void show_sfm	(SFM_t* sfm_p);
SFM_t* read_sfm	(char* sfm_name, int* err_p);
int open_sdla	(char* name, int* sdla_p, sdlaCtl_t* sdlacb_p);
int load_sdla	(int sdla, SFMInfo_t* info_p, void* code, void* data);
unsigned long filesize (FILE* file);
unsigned short checksum (unsigned char* buf, unsigned int len);
void* lookup	(int code, LookUp_t* table);
int name2val	(char* name, LookUp_t* table);
int hexdump	(char* str, unsigned char* data, int length, int limit);

/*
 * Firmware configuration routines.
 */
char* makeBootConf (sdlaStatus_t* stat, SFMInfo_t* info, char* cfgfile);
int makeX25Conf (char* conf);

/****** Global Data *********************************************************/

/*
 * Messages.
 */
char banner[]	=
	"SDLA Loader. v3.10 (c) 1995-1996 Sangoma Technologies Inc."
;
char helptext[] =
	"Usage:\n"
	"  sdlald [-v] {device} {firmware}       Load adapter firmware\n"
	"  sdlald -s {device}                    Show adapter status\n"
	"  sdlald -b {device} {module}           Bind to protocol module\n"
	"  sdlald -u {device}                    Unbind from protocol module\n"
	"  sdlald -c {device} [{options}]        Configure adapter\n"
	"  sdlald -r {device}                    Reset adapter\n"
	"  sdlald -z {device}                    Shut down adapter\n"
	"  sdlald -d {device} [{addr} [{size}]]  Dump adapter memory\n"
	"  sdlald -f {firmware}                  Show SFM header info\n"
	"  sdlald -h | -?                        Display help screen\n"
	"Where:\n"
	"  {device}    SDLA device node (e.g. /dev/sdla0)\n"
	"  {firmware}  SDLA firmware module name (SFM-file)\n"
	"  {module}    Name of the protocol-specific module\n"
	"  {options}   Hardware configuration options:\n"
	"                p=port    set SDLA base I/O port address (hex)\n"
	"                m=addr    set SDLA base memory window address (hex)\n"
	"                i=irq     set SDLA IRQ level (decimal)\n"
	"                z=clock   set SDLA clock rate (KHz, decimal)\n"
	"  {addr}      Local SDLA memory address to dump from (hex)\n"
	"  {size}      Size of memory area to be dumped in bytes (hex)\n"
	" Other switches are:\n"
	"   -v         Verbose mode"
;

char *errMessages[] =		/* Error messages */
{
	"System error",
	"Invalid command line switch",
	"Invalid command line syntax",
	"Cannot allocate memory",
	"SDLA driver request error",
	"Not an SDLA driver",
	"Invalid adapter state",
	"Incompatible adapter hardware",
	"Not an SFM file",
	"Invalid SFM file version",
	"Corrupted SFM file",
	"Unknown firmware type",
	"Invalid configuration option",
	"Unknown error code",
};

LookUp_t drverr_table[] =	/* Driver error messages */
{
	{ SDLAER_BADVER,	"Wrong SDLA driver version"	},
	{ SDLAER_CONFLICT,	"Conflicting hardware option"	},
	{ SDLAER_BADTYPE,	"Unknown adapter type"		},
	{ SDLAER_BADHWOPT,	"Invalid hardware option"	},
	{ SDLAER_MEMCONFL,	"Memory window area unusable"	},
	{ SDLAER_NOTFOUND,	"Object not found"		},
	{ SDLAER_BADMEM,	"Adapter memory test failed"	},
	{ SDLAER_BADSTATE,	"Invalid adapter state"		},
	{ SDLAER_FAILURE,	"Hardware failure"		},
	{ 0,			"SDLA driver failure"		}, /* EOT */
};

LookUp_t sdla_types[] =		/* Adapter types */
{
	{ SDLA_S502A,		"S502A"		},
	{ SDLA_S502E,		"S502E"		},
	{ SDLA_S503,		"S503"		},
	{ SDLA_S507,		"S507"		},
	{ SDLA_S508,		"S508"		},
	{ SDLA_S509,		"S509"		},
	{ 0,			"UNKNOWN"	},
};

LookUp_t sdla_states[] =	/* Adapter states */
{
	{ SDLA_CONFIGURED,	"CONFIGURED"	},
	{ SDLA_RUNNING,		"RUNNING"	},
	{ SDLA_BOUND,		"BOUND"		},
	{ 0,			"SHUT DOWN"	},
};

/*
 * Miscellaneous.
 */
char progname[] = PROG_NAME;
int action = ACTION_LOAD;	/* requested action */
int flags;			/* execution options */
int syserr;			/* system error number (last errno) */
int drverr;			/* SDLA driver error code */

sdlaCtl_t sdla_cb =		/* SDLA control block */
{
	SDLA_SIGNATURE,
	SDLA_VERSION,
	0,
	0,
	{
		{ 0, }
	}
};

/****** Entry Point *********************************************************/

int main (int argc, char* argv[])
{
	int err = 0;	/* return code */

	/* Process command line switches */
	if (argc > 1) while (!err && --argc && (**++argv == ARG_SWITCH))
		err = switch_proc(*argv)
	;
	else action = ACTION_HELP;

	/* Display banner */
	if (flags & FLAG_VERBOSE) puts(banner);

	/* Perform requested action */
	if (!err) switch (action)
	{
	case ACTION_LOAD:
		err = action_load(argc, argv);
		break;
	
	case ACTION_BIND:
		err = action_bind(argc, argv);
		break;
	
	case ACTION_UNBIND:
		err = action_unbind(argc, argv);
		break;
	
	case ACTION_STATUS:
		err = action_status(argc, argv);
		break;

	case ACTION_SETUP:
		err = action_setup(argc, argv);
		break;
	
	case ACTION_RESET:
		err = action_reset(argc, argv);
		break;
	
	case ACTION_DOWN:
		err = action_down(argc, argv);
		break;

	case ACTION_DUMP:
		err = action_dump(argc, argv);
		break;
	
	case ACTION_SHOW_SFM:
		err = action_show_sfm(argc, argv);
		break;
	
	default:
		puts(helptext);
	}

	/* Display error message */
	if (err) show_error(err);
	return err;
}

/****** Action Handlers *****************************************************/

/*============================================================================
 * Load adapter firmware.
 *	arg[0]	- device name
 *	arg[1]	- firmware module name
 *	arg[2]	- configuration file name (optional)
 */
int action_load (int argc, char *argv[])
{
	int err = 0;			/* error code */
	void* databuf = NULL;		/* configuration data buffer */
	SFM_t* sfm_p;			/* -> SFM memory image */
	int sdla;			/* device file handle */

	if (argc < 2) return ERR_SYNTAX;

	if (flags & FLAG_VERBOSE) printf (
		"Loading firmware module %s onto adapter %s ...\n",
		argv[1], argv[0])
	;
	err = open_sdla(argv[0], &sdla, &sdla_cb);
	if (err) return err;

	/* Set adapter to RUNNING state */
	switch (sdla_cb.u.status.state)
	{
	case SDLA_CONFIGURED:
		break;

	case SDLA_RUNNING:
		if ((ioctl(sdla, SDLAIO_STOP, &sdla_cb) < 0) ||
		    (ioctl(sdla, SDLAIO_STATUS, &sdla_cb) < 0))
		{
			if (errno == EIO)
			{
				err = ERR_DRIVER;
				drverr = sdla_cb.result;
			}
			else
			{
				err = ERR_SYSTEM;
				syserr = errno;
			}
		}
		break;

	default:
		err = ERR_BAD_STATE;
	}
	if (err)
	{
		close(sdla);
		return err;
	}

	/* Read firmware module */
	sfm_p = read_sfm(argv[1], &err);
	if (err)
	{
		close(sdla);
		return err;
	}

	/*
	 * Check hardware compatibility.
	 */
	if (sdla_cb.u.status.mem_size < sfm_p->info.memsize)
		err = ERR_HW_COMPAT
	;
	else
	{
		int i;

		for (i = 0; i < SFM_MAX_SDLA; ++i)
			if (sdla_cb.u.status.type == sfm_p->info.adapter[i])
				break
		;
		if (i >= SFM_MAX_SDLA) err = ERR_HW_COMPAT;
		else
		{
			/*
			 * Prepare boot-time configuration data, if necessary.
			 */
			if (sfm_p->info.datasize)
			{ 
				databuf = makeBootConf(
					&sdla_cb.u.status,
					&sfm_p->info,
					(argc > 2) ? argv[2] : NULL)
				;
				if (databuf == NULL) err = ERR_SYSTEM;
			}
			if (!err)
			{
				/*
				 * Download firmware onto adapter.
				 */
				err = load_sdla(sdla, &sfm_p->info, sfm_p->image, databuf);
			}
			if (databuf) free(databuf);
		}
	}
	free(sfm_p);
	close(sdla);
	return err;
}

/*============================================================================
 * Bind adapter to a protocol support module.
 *	arg[0]	- device name
 *	arg[1]	- protocol module name
 */
int action_bind (int argc, char* argv[])
{
	int err = 0;		/* error code */
	int sdla;		/* device file handle */

	if ((argc < 2) || (strlen(argv[1]) >= SDLA_MODNAME_SZ))
		return ERR_SYNTAX
	;

	if (flags & FLAG_VERBOSE) printf(
		"Binding protocol module %s to adapter %s ...\n",
		argv[1], argv[0])
	;
	err = open_sdla(argv[0], &sdla, &sdla_cb);
	if (err) return err;

	strcpy(sdla_cb.u.modname, argv[1]);
	if (ioctl(sdla, SDLAIO_BIND, &sdla_cb) < 0)
	{
		if (errno == EIO)
		{
			err = ERR_DRIVER;
			drverr = sdla_cb.result;
		}
		else
		{
			err = ERR_SYSTEM;
			syserr = errno;
		}
	}
	close(sdla);
	return err;
}

/*============================================================================
 * Unbind adapter from protocol support module.
 *	arg[0]	- device name
 */
int action_unbind (int argc, char *argv[])
{
	int err = 0;		/* error code */
	int sdla;			/* device file handle */

	if (argc < 1) return ERR_SYNTAX;

	if (flags & FLAG_VERBOSE) printf(
		"Unbinding adapter %s from protocol module...\n", argv[0])
	;
	err = open_sdla(argv[0], &sdla, &sdla_cb);
	if (err) return err;

	if (ioctl(sdla, SDLAIO_UNBIND, &sdla_cb) < 0)
	{
		if (errno == EIO)
		{
			err = ERR_DRIVER;
			drverr = sdla_cb.result;
		}
		else
		{
			err = ERR_SYSTEM;
			syserr = errno;
		}
	}
	close(sdla);
	return err;
}

/*============================================================================
 * Get adapter status.
 *	arg[0]	- device name
 */
int action_status (int argc, char *argv[])
{
	int err = 0;
	int sdla;		/* device file handle */

	if (!argc) return ERR_SYNTAX;

	err = open_sdla(argv[0], &sdla, &sdla_cb);
	if (err) return err;

	if (flags & FLAG_VERBOSE)
	{
		printf("Reading status of the %s...\n", argv[0]);
		show_status(&sdla_cb.u.status);
	}
	else printf("%d\n", sdla_cb.u.status.state);
	close(sdla);
	return 0;
}

/*============================================================================
 * Setup adapter hardware.
 *	arg[0]	- device name
 *	arg[1..N]	- hardware configuration options
 */
int action_setup (int argc, char* argv[])
{
	int err = 0;
	int sdla;		/* device file handle */
	int i;
	char* str;

	if (!argc) return ERR_SYNTAX;

	if (flags & FLAG_VERBOSE)
		printf ("Configuring adapter %s ...\n", argv[0])
	;
	err = open_sdla(argv[0], &sdla, &sdla_cb);
	if (err) return err;

	/*
	 * Parse hardware configuration options.
	 */
	for (i = 1; !err && (i < argc); ++i)
		err = parse_opt(argv[i], &sdla_cb.u.setup)
	;
	sdla_cb.u.setup.dpm_size = 0x2000;
	sdla_cb.u.setup.options  = 0;
	if (!err)
	{
		if (flags & FLAG_VERBOSE)
		{
			printf("I/O base (hex) ..... %Xh\n",
				sdla_cb.u.setup.io_base)
			;
			printf("IRQ level .......... %d\n",
				sdla_cb.u.setup.irq)
			;
			printf("Window base (hex) .. %Xh\n",
				sdla_cb.u.setup.dpm_base)
			;
			printf("Window size (hex) .. %Xh\n",
				sdla_cb.u.setup.dpm_size)
			;
			printf("CPU clock, kHz ..... %d\n",
				sdla_cb.u.setup.clock_rate)
			;
		}

		/*
		 * Set up adapter.
		 */
		if (ioctl(sdla, SDLAIO_SETUP, &sdla_cb) < 0)
		{
			if (errno == EIO)
			{
				err = ERR_DRIVER;
				drverr = sdla_cb.result;
			}
			else
			{
				err = ERR_SYSTEM;
				syserr = errno;
			}
		}
		else
		{
			/*
			 * Calibrate adapter speed, if nesessary.
			 */
/*
			if (sdla_cb.u.status.clock_rate == 0)
			{
				printf("Calibrating adapter speed...\n");
				err = calibrate_sdla(sdla, &sdla_cb.u.status.clock_rate);
			}
*/
		}
	}
	close(sdla);
	return err;
}

/*============================================================================
 * Reset adapter.
 *	arg[0]	- device name
 */
int action_reset (int argc, char* argv[])
{
	int err = 0;
	int sdla;		/* device file handle */

	if (!argc) return ERR_SYNTAX;

	if (flags & FLAG_VERBOSE)
		printf ("Stopping device %s ...\n", argv[0])
	;
	err = open_sdla(argv[0], &sdla, &sdla_cb);
	if (err) return err;

	if (ioctl(sdla, SDLAIO_STOP, &sdla_cb) < 0)
	{
		if (errno == EIO)
		{
			drverr = sdla_cb.result;
			err = ERR_DRIVER;
		}
		else
		{
			syserr = errno;
			err = ERR_SYSTEM;
		}
	}
	close(sdla);
	return err;
}

/*============================================================================
 * Shut down adapter.
 *	arg[0]	- device name
 */
int action_down (int argc, char* argv[])
{
	int err = 0;
	int sdla;		/* device file handle */

	if (!argc) return ERR_SYNTAX;

	if (flags & FLAG_VERBOSE)
		printf ("Shutting down device %s ...\n", argv[0])
	;
	err = open_sdla(argv[0], &sdla, &sdla_cb);
	if (err) return err;

	if (ioctl(sdla, SDLAIO_DOWN, &sdla_cb) < 0)
	{
		if (errno == EIO)
		{
			drverr = sdla_cb.result;
			err = ERR_DRIVER;
		}
		else
		{
			syserr = errno;
			err = ERR_SYSTEM;
		}
	}
	close(sdla);
	return err;
}

/*============================================================================
 * Dump adapter memory region.
 *	arg[0]	- device name
 *	arg[1]	- offset (optional)
 *	arg[2]	- length (optional)
 */
int action_dump (int argc, char* argv[])
{
  int		err	= 0;	/* error code */
  unsigned long	offset	= 0;
  unsigned long	length	= 0;
  unsigned char	*buffer;	/* dump buffer */
  char		dumpstr[256];
  int		sdla, i, cnt;

  if (!argc) err = ERR_SYNTAX;
  else for (i = 1; (i < argc) && !err; ++i)
  {
    if (!isxdigit(*argv[i])) err = ERR_SYNTAX;
    else switch (i)
    {
      case 1:	/* offset */
        offset = strtoul(argv[i], NULL, 16);
        break;

      case 2:	/* length */
        length = strtoul(argv[i], NULL, 16);
        break;

      default:
        err = ERR_SYNTAX;
    }
  }
  if (!err)
  {
    err = open_sdla(argv[0], &sdla, &sdla_cb);
    if (!err)
    {
      if (sdla_cb.u.status.state < SDLA_CONFIGURED)
        err = ERR_BAD_STATE
      ;
      else if (offset >= sdla_cb.u.status.mem_size)
        err = ERR_BAD_OPTION
      ;
      else
      {
        if (!length || ((offset + length) > sdla_cb.u.status.mem_size))
          length = sdla_cb.u.status.mem_size - offset
        ;
        if (flags & FLAG_VERBOSE)
          printf ("Dumping 0x%X bytes at offset 0x%X from device %s ...\n",
                  length, offset, argv[0])
        ;
        buffer = malloc(length);
        {
	  if (!buffer) err = ERR_MALLOC;
          else
          {
            /*
             * Read adapter memory.
             */
	    sdla_cb.u.xfer.offset = offset;	/* load offset */
	    sdla_cb.u.xfer.length = length;
            sdla_cb.u.xfer.ptr    = buffer;
            if (ioctl(sdla, SDLAIO_PEEK, &sdla_cb) < 0) 
            {
              if (errno == EIO)
              {
                err = ERR_DRIVER;
                drverr = sdla_cb.result;
              }
              else
              {
                err = ERR_SYSTEM;
                syserr = errno;
              }
            }
            else
            {
              if (length > 16)
              {
		cnt = 16 - (offset % 16);
                if (cnt)
                {
                  printf("%06X: ", offset);
                  offset += hexdump(dumpstr, buffer, cnt, 16);
                  puts(dumpstr);
		}
	      }
              else cnt = 0;
              while (cnt < length)
              {
                printf("%06X: ", offset);
                offset += hexdump(dumpstr, &buffer[cnt],
                                  min(16, length - cnt), 16)
                ;
                cnt += 16;
                puts(dumpstr);
              }
            }
            free(buffer);
          }
        }
      }
      close(sdla);
    }
  }
  return err;
} /* action_dump() */

/*============================================================================
 * Display SFM header.
 *	arg[0]	- firmware module name
 */
int action_show_sfm (int argc, char* argv[])
{
	int err = 0;
	SFM_t* sfm_p;

	if (!argc) return ERR_SYNTAX;

	if (flags & FLAG_VERBOSE)
		printf("Reading firmware module: %s\n", *argv)
	;
	sfm_p = read_sfm(*argv, &err);
	if (!err)
	{
		show_sfm(sfm_p);	/* display SFM header info */
		free(sfm_p);
	}
	return err;
}

/***** Helper Functions *****************************************************/

/*============================================================================
 * Process command line switches
 *	Return 0 if processed o.k., otherwise return error code.
 */
int switch_proc (char* str)
{
	int err = 0;
	char* sw;

	for (sw = ++str; *sw && !err; ++sw) switch (*sw)
	{
	case SW_CONFIGURE:
		action = ACTION_SETUP;
		break;

	case SW_STATUS:
		action = ACTION_STATUS;
		break;

	case SW_RESET:
		action = ACTION_RESET;
		break;

	case SW_SHUTDOWN:
		action = ACTION_DOWN;
		break;

	case SW_DUMP:
		action = ACTION_DUMP;
		break;

	case SW_BIND:
		action = ACTION_BIND;
		break;

	case SW_UNBIND:
		action = ACTION_UNBIND;
		break;

	case SW_SHOW_SFM:
		action = ACTION_SHOW_SFM;
		break;

	case SW_GETHELP:
	case SW_ALTHELP:
		action = ACTION_HELP;
		flags |= FLAG_VERBOSE;
		break;

	case SW_VERBOSE:
		flags |= FLAG_VERBOSE;
		break;

	default:
		err = ERR_SWITCH;
	}
	return err;
}

/*============================================================================
 * Parse SDLA hardware configuration options.
 */
int parse_opt (char* str, sdlaSetup_t* setup_p)
{
	int err = 0;

	switch (*str++)
	{
	case OPT_SDLA_TYPE:
/*
      if (*str++ != OPT_SEPARATOR) err = ERR_SYNTAX;
      else
      {
	setup_p->type = name2val(str, sdla_types);
	if (!setup_p->type) err = ERR_BAD_OPTION;
      }
*/
		break;

	case OPT_IOPORT:
		if (*str++ != OPT_SEPARATOR) err = ERR_SYNTAX;
		else setup_p->io_base = strtoul(str, NULL, 16);
		break;

	case OPT_ADDRESS:
		if (*str++ != OPT_SEPARATOR) err = ERR_SYNTAX;
		else setup_p->dpm_base = strtoul(str, NULL, 16);
		break;

	case OPT_IRQ:
		if (*str++ != OPT_SEPARATOR) err = ERR_SYNTAX;
		else setup_p->irq = strtoul(str, NULL, 10);
		break;

	case OPT_CLOCK:
		if (*str++ != OPT_SEPARATOR) err = ERR_SYNTAX;
		else setup_p->clock_rate = strtoul(str, NULL, 10);
		break;
	}
	return err;
}

/*============================================================================
 * Show error message.
 */
void show_error (int err)
{
	char* str;

	switch (err)
	{
	case ERR_SYSTEM:
		str = strerror(syserr);
		break;

	case ERR_DRIVER:
		str = lookup(drverr, drverr_table);
		break;

	default:
		str = errMessages[min(err, ERR_LIMIT) - 1];
	}
	fprintf(stderr, "%s error %u: %s!\n", progname, err, str);
}

/*============================================================================
 * Show adapter status.
 */
void show_status (sdlaStatus_t* status)
{
	char* str;

	str = lookup(status->state, sdla_states);
	printf("State .............. %s\n", str);
	if (status->state != SDLA_DISABLED)
	{
		int i;

		str = lookup(status->type, sdla_types);
		printf("Type ............... %s\n", str);
		printf("I/O port (hex) ..... %X\n", status->io_base);
		printf("IRQ level .......... %d\n", status->irq);
		printf("Window base (hex) .. %X\n", status->dpm_base);
		printf("Window size (hex) .. %X\n", status->dpm_size);
		printf("CPU clock, kHz ..... %d\n", status->clock_rate);
		printf("RAM size (hex) ..... %X\n", status->mem_size);
		printf("RAM vector (hex) ... %X\n", status->mem_vector);
		printf("Interrupt count .... %d\n", status->int_count);
		printf("Firmware ID ........ %d\n", status->firmware_id);
		if (status->state == SDLA_BOUND) printf(
			"Bound to module .... %s\n", status->modname)
		;
		printf("WR registers ....... ");
		for (i = 0; i < status->io_range; ++i)
			printf("%02X ", status->regs_wr[i])
		;
		putchar('\n');
		printf("RD registers ....... ");
		for (i = 0; i < status->io_range; ++i)
			printf("%02X ", status->regs_rd[i])
		;
		putchar('\n');
	}
}

/*============================================================================
 * Show SFM Header Info.
 */
void  show_sfm (SFM_t *sfm_p)
{
	int i;
	char* str;

	printf("Description: %s\n", sfm_p->descr);
	printf("Firmware ID ......................... %d\n", sfm_p->info.codeid);
	printf("Firmaware version number ............ %d\n", sfm_p->info.version);
	printf("Compatible adapter types ............ ");
	for (i = 0; sfm_p->info.adapter[i]; ++i)
	{
		str = lookup(sfm_p->info.adapter[i], sdla_types);
		printf("%s ", str);
	}
	putchar('\n');
	printf("Minimum adapter memory size (kB) .... %d\n",
		(unsigned short)(sfm_p->info.memsize / 1024));
	printf("Default memory window offset (hex) .. %X\n", sfm_p->info.winoffs);
	printf("Code entry point offset (hex) ....... %X\n", sfm_p->info.startoffs);
	printf("Code load offset (hex) .............. %X\n", sfm_p->info.codeoffs);
	printf("Code image size ..................... %d\n", sfm_p->info.codesize);
	printf("Configuration data offset (hex) ..... %X\n", sfm_p->info.dataoffs);
	printf("Configuration data size ............. %d\n", sfm_p->info.datasize);
}

/*============================================================================
 * Open SDLA device.
 *	o Open special device file
 *	o Verify file type
 *	o Read SDLA status
 *	o Verify device presence
 */
int open_sdla (char* name, int* sdla_p, sdlaCtl_t* sdlacb_p)
{
	int err = 0;
	int sdla;
	struct stat stat;

	sdla = open(name, O_RDWR);
	if (sdla <= 0)
	{
		syserr = errno;
		return ERR_SYSTEM;
	}

	/*
	 * Check to see if we've actually opened a device driver, then get
	 * adapter status. This will also verify driver version, device
	 * presence, etc.
	 */
	memset(&sdlacb_p->u.status, 0, sizeof(sdlacb_p->u.status));
	if (fstat(sdla, &stat) < 0)
	{
		err = ERR_SYSTEM;
		syserr = errno;
	}
	else if (!S_ISCHR(stat.st_mode))
		err = ERR_BAD_DRIVER
	;
	else if (ioctl(sdla, SDLAIO_STATUS, sdlacb_p) < 0)
	{
		if (errno == EIO)
		{
			err = ERR_DRIVER;
			drverr = sdlacb_p->result;
		}
		else
		{
			err = ERR_SYSTEM;
			syserr = errno;
		}
	}
	if (err)
	{
		close(sdla);
		*sdla_p = 0;
	}
	else *sdla_p = sdla;
	return err;
}

/*============================================================================
 * Read SDLA firmware module.
 */
SFM_t* read_sfm (char* sfm_name, int* err_p)
{
	int err = 0;
	SFM_t* sfm_p = NULL;
	FILE* sfm_file;
	unsigned long sfm_size;

	sfm_file = fopen(sfm_name, "rb");
	if (sfm_file == NULL)
	{
		syserr = errno;
		*err_p = ERR_SYSTEM;
		return NULL;
	}

	sfm_size = filesize(sfm_file);
	if (sfm_size <= sizeof(SFM_t))
	{
		fclose(sfm_file);
		*err_p = ERR_NOT_SFM;
		return NULL;
	}

	sfm_p = malloc(sfm_size);
	if (sfm_p == NULL)
	{
		fclose(sfm_file);
		syserr = errno;
		*err_p = ERR_SYSTEM;
		return NULL;
	}

	/*
	 * Read SFM file into memory.
	 */
	if (fread(sfm_p, 1, sfm_size, sfm_file) < sfm_size)
	{
		err = ERR_SYSTEM;
		syserr = errno;
	}
	else if (strcmp(sfm_p->signature, SFM_SIGNATURE))
		err = ERR_NOT_SFM
	;
	else if (sfm_p->version != SFM_VERSION)
		err = ERR_SFM_VERSION
	;
	else if ((sfm_size - offsetof(SFM_t, image) != sfm_p->info.codesize) ||
		 (checksum((void*)&sfm_p->info, sizeof(SFMInfo_t) + sfm_p->info.codesize)
		 != sfm_p->checksum))
		err = ERR_BAD_SFM
	;
	fclose(sfm_file);

	if (err)
	{
		free(sfm_p);
		*err_p = err;
		sfm_p = NULL;
	}
	return sfm_p;
}

/*============================================================================
 * Load SDLA.
 */
int load_sdla (int sdla, SFMInfo_t* info_p, void* code, void* data)
{
	/*
	 * Download firmware onto the board.
	 */
	sdla_cb.u.xfer.offset = info_p->codeoffs;	/* load offset */
	sdla_cb.u.xfer.length = info_p->codesize;
	sdla_cb.u.xfer.ptr    = code;
	if (ioctl(sdla, SDLAIO_POKE, &sdla_cb))
	{
		drverr = sdla_cb.result;
		return ERR_DRIVER;
	}

	/*
	 * Download configuration data onto the board.
	 */
	if (data)
	{
		sdla_cb.u.xfer.offset = info_p->dataoffs; /* load offset */
		sdla_cb.u.xfer.length = info_p->datasize;
		sdla_cb.u.xfer.ptr    = data;
		if (ioctl(sdla, SDLAIO_POKE, &sdla_cb))
		{
			drverr = sdla_cb.result;
			return ERR_DRIVER;
		}
	}

	/*
	 * Start the board up.
	 */
	sdla_cb.u.start.firmware_id = info_p->codeid;
	sdla_cb.u.start.entry_point = info_p->startoffs;
	sdla_cb.u.start.dpm_vector  = info_p->winoffs;
	if (ioctl(sdla, SDLAIO_START, &sdla_cb))
	{
		drverr = sdla_cb.result;
		return ERR_DRIVER;
	}
	return 0;
}

/*============================================================================
 * Dump data into a string in the following format:
 * XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX AAAAAAAAAAAAAAAA
 *	Return number of bytes dumped.
 *	NOTE: buffer string must be at least (limit * 4 + 2) bytes long.
 */
int hexdump (char* str, unsigned char* data, int length, int limit)
{
	int i, n;

	n = min(limit, length);
	if (n)
	{
		for (i = 0; i < n; ++i)
			str += sprintf(str, "%02X ", data[i])
		;
		for (i = 0; i < limit - n; ++i)
			str += sprintf(str, "   ")
		;
		str += sprintf(str, " ");
		for (i = 0; i < n; ++i)
			str += sprintf(str, "%c", (isprint(data[i])) ? data[i] : '.')
		;
		for (i = 0; i < limit - n; ++i)
			str += sprintf(str, " ")
		;
	}
	*str = '\0';
	return n;
}

/* ===========================================================================
 * Get file size
 *	Return file length or 0 if error.
 */
unsigned long filesize (FILE* file)
{
	unsigned long size = 0;
	unsigned long cur_pos;

	cur_pos = ftell(file);
	if ((cur_pos != -1L) && !fseek(file, 0, SEEK_END))
	{
		size = ftell(file);
		fseek(file, cur_pos, SEEK_SET);
	}
	return size;
}

/* ===========================================================================
 * Calculate 16-bit CRC using CCITT polynomial.
 */
unsigned short checksum (unsigned char* buf, unsigned int len)
{
	unsigned short crc = 0;
	unsigned mask, flag;

	for (; len; --len, ++buf)
	{
		for (mask = 0x80; mask; mask >>= 1)
		{
			flag = (crc & 0x8000);
			crc <<= 1;
			crc |= ((*buf & mask) ? 1 : 0);
			if (flag) crc ^= 0x1021;
		}
	}
	return crc;
}

/*============================================================================
 * Get a pointer from a look-up table.
 */
void *lookup (int code, LookUp_t* table)
{
	while (table->code && (table->code != code)) table++;
	return table->ptr;
}

/*============================================================================
 * Look up a symbolic name in name table.
 *	Return a numeric value associated with that name or zero, if name was
 *	not found.
 */
int name2val (char* name, LookUp_t* table)
{
	while (table->code && strcmp(name, table->ptr)) table++;
	return table->code;
}

/****** Firmware Configuration Routines *************************************/

/*============================================================================
 * Prepare boot-time firmware configuration block.
 */
char* makeBootConf (sdlaStatus_t* stat, SFMInfo_t* info, char* cfgfile)
{
	char* conf = NULL;

	conf = malloc(info->datasize);
	if (conf)
	{
		memset(conf, 0, info->datasize);

		switch (stat->clock_rate)
		{
			case 5000:  conf[0] = 0x01; break;
			case 7200:  conf[0] = 0x02; break;
			case 8000:  conf[0] = 0x03; break;
			case 10000: conf[0] = 0x04; break;
			case 16000: conf[0] = 0x05; break;
		}
		switch (stat->type)
		{
			case SDLA_S502E: conf[0] |= 0x80; break;
			case SDLA_S503:  conf[0] |= 0x40; break;
		}

		switch (info->codeid)
		{
		case SFID_X25_502:
		case SFID_X25_508:
			makeX25Conf(&conf[1]);
			break;
		}
	}
	return conf;
}

/*============================================================================
 * Prepare X.25 boot-time firmware configuration block.
 */
int makeX25Conf (char* conf)
{
	conf[0x01] = 3;			/* T1 timer */
	conf[0x03] = 10;		/* N2 */
	conf[0x06] = 7;			/* HDLC window size */
	conf[0x0B] = 1;			/* DTE */
	conf[0x0C] = 2;			/* X.25 packet window size */
	*(short*)&conf[0x0D] = 128;	/* default X.25 data field size */
	*(short*)&conf[0x0F] = 128;	/* maximum X.25 data field size */

	return 0;
}

/****** End *****************************************************************/

