/*****************************************************************************
* PROJECT:	spppcfg		Sangoma Synchronous PPP Driver Configurator
* FILE:		spppcfg.c	Main
* TARGET OS:	Linux
* AUTHOR:	Gene Kozin
* ----------------------------------------------------------------------------
* (c) Copyright 1995-1996 Sangoma Technologies Inc.  All Rights Reserved.
* ============================================================================
* Rev.#	Date		Who	Details
* -----	-----------	---	----------------------------------------------
*  1.00	Jun.20,1996	GK	Initial version.
*****************************************************************************/

/*****************************************************************************
* Usage:
*   spppcfg {interface} {conf_file}	Configure interface
*   spppcfg {interface}			Get interface status
*
* Where:
*   {interface}	Interface name, e.g. sppp0
*   {conf_file}	Configuration file name
*
* Optional command line swithces:
*   -v		verbose mode
*   -h or -?	display help screen
*****************************************************************************/

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include "spppio.h"		/* SPPP driver API */

/****** Defines *************************************************************/

#ifndef	min
#define	min(a,b)	(((a)<(b))?(a):(b))
#endif

#define	PROG_NAME	"spppcfg"

/*
 *----- Error Codes ----------------------------------------------------------
 */
#define	ERR_SYSTEM	1	/* System error */
#define	ERR_SWITCH	2	/* Invalid command line switch */
#define ERR_SYNTAX	3	/* Invalid command line syntax */

#define	ERR_DRIVER	10	/* General driver failure */
#define	ERR_CFG_SYNTAX	11	/* Configuration file syntax error */

/*
 * Command Line Parsing Stuff.
 */
#define ARG_SWITCH	'-'	/* switch follows */
#define	SW_VERBOSE	'v'	/* enable verbose output */
#define	SW_GETHELP	'h'	/* 'get help' switch */
#define	SW_ALTHELP	'?'	/* same */

/*
 * Configuration Stuff.
 */
#define	MAX_CFGLINE	256
#define	MAX_CMDLINE	256
#define	MAX_TOKENS	32

/*
 * Miscellaneous.
 */
#define	FLAG_VERBOSE	0x0001	/* enable verbose output */
#define	FLAG_HELP	0x0002	/* show help screen */

/****** Data Types **********************************************************/

typedef struct	LookUp			/* Look-up Table entry */
{
  int	code;				/* look-up code */
  void	*ptr;				/* pointer */
} LookUp_t;

typedef struct CfgRec			/* Configuration Record */
{
  char	*keyword;			/* keyword */
  int	argtype;			/* type of argument */
  int	offs;				/* offset into configuration struct.*/
  void	*argvals;			/* transcription data */
} CfgRec_t;

/*
 * Argument types.
 */
#define	ARGT_SYMNAME	1		/* symbolic name */
#define	ARGT_NUMERIC	2		/* number */
#define	ARGT_NUMLIST	3		/* list of numbers */

/****** Function Prototypes *************************************************/

/*
 * Command handlers.
 */
int	cmd_config	(char *ifname, char *filename);
int	cmd_status	(char *ifname);

/*
 * Support functions.
 */
int	switch_proc	(char *str);
void	show_help	(void);
void	show_error	(int err);
int	read_conf	(spppConf_t *cfg_p, char *filename);
int	parse_cfg	(char *cfgline, CfgRec_t *cfgtable, void *cfgdata);
int	tokenize	(char *str, char *space, char *eol, char **tokens);
void	*lookup		(int code, LookUp_t *table);
int	name2val	(char *name, LookUp_t *table);

/****** Global Data *********************************************************/

/*
 *----- Messages -------------------------------------------------------------
 */
char banner[]	=
	"SPPP Configurator. v1.00 (c) 1995-1996 Sangoma Technologies Inc."
;
char helptext[] =
	"Usage:\n"
	"  spppcfg {interface} {conf_file}   Configure interface\n"
	"  spppcfg {interface}               Get interface status\n"
	"\n"
	"Where:\n"
	"  {interface}    interface name, e.g. sppp0\n"
	"  {conf_file}    configuration file name\n"
	"\n"
	"Optional command line switches:\n"
	"  -h or -?       display help screen\n"
	"  -v             enable verbose output\n"
;

LookUp_t	err_table[] =		/* Error messages */
{
  { ERR_SWITCH,		"Invalid command line switch"		},
  { ERR_SYNTAX,		"Invalid command line syntax"		},
  { ERR_CFG_SYNTAX,	"Configuration file syntax error"	},
  { 0,			"Unknown error code"			}, /* EOT */
};

LookUp_t	drverr_table[] =	/* Driver error messages */
{
  { SPPPER_BADCMD,	"Invalid SPPP IOCTL command"		},
  { SPPPER_BADARG,	"Invalid SPPP IOCTL argument"		},
  { SPPPER_BADVER,	"Incompatible driver version"		},
  { SPPPER_BADSTATE,	"Invalid device state"			},
  { 0,			"SPPP IOCTL failure"			}, /* EOT */
};

/*
 *----- Miscellaneous --------------------------------------------------------
 */
char		progname[] = PROG_NAME;
int		flags;
int		syserr;			/* system error number */
int		drverr;			/* device error code */
struct ifreq	ifr;

spppCtl_t	sppp_cb =	/* SPPP control block */
{
  SPPP_SIGNATURE,	/* .signature */
  SPPP_VERSION,		/* .version */
  0,			/* .result */
  0,			/* .reserved */
  {			/* .v */
    { 0, }
  }
};

/*
 *----- Configuration Data ---------------------------------------------------
 */
LookUp_t	yesno_opt[] =
{
  { 1,	"Yes"	}, 
  { 0,	"No"	}, 
  { 0,	NULL	}, 
};

LookUp_t	onoff_opt[] =
{
  { 1,	"On"	}, 
  { 0,	"Off"	}, 
  { 0,	NULL	}, 
};

LookUp_t	port_opt[] =
{
  { 1,	"V35"	},
  { 0,	"RS232"	},
  { 0,	NULL	}, 
};

LookUp_t	clock_opt[] =
{
  { 1,	"Internal"	}, 
  { 0,	"External"	}, 
  { 0,	NULL		}, 
};

CfgRec_t	cfgtable[] =
{
  { "PortType",	ARGT_SYMNAME, offsetof(spppConf_t, interface), port_opt },
  { "Clocking",	ARGT_SYMNAME, offsetof(spppConf_t, clocking), clock_opt },
  { "BaudRate",	ARGT_NUMERIC, offsetof(spppConf_t, speed), 0 },
  { "RestartTimer", ARGT_NUMERIC, offsetof(spppConf_t, restart_tmr), 0 },
  { "AuthRestart", ARGT_NUMERIC, offsetof(spppConf_t, auth_rsrt_tmr), 0 },
  { "AuthWait", ARGT_NUMERIC, offsetof(spppConf_t, auth_wait_tmr), 0 },
  { "ModemTimeout", ARGT_NUMERIC, offsetof(spppConf_t, mdm_fail_tmr), 0 },
  { "DTRDropTimer", ARGT_NUMERIC, offsetof(spppConf_t, dtr_drop_tmr), 0	},
  { "ConnectTimeout", ARGT_NUMERIC, offsetof(spppConf_t, connect_tmout), 0 },
  { "ConfigureRetry", ARGT_NUMERIC, offsetof(spppConf_t, conf_retry), 0 },
  { "TerminateRetry", ARGT_NUMERIC, offsetof(spppConf_t, term_retry), 0 },
  { "MaxConfReject", ARGT_NUMERIC, offsetof(spppConf_t, fail_retry), 0 },
  { "AuthRetry", ARGT_NUMERIC, offsetof(spppConf_t, auth_retry), 0 },
  { NULL, 0, 0, 0 }
};

char	cfgspaces[]	= "\t =:";
char	cfgeol[]	= "\n\r#";

/****** Entry Point *********************************************************/

int main (int argc, char *argv[])
{
  int	err = 0;	/* return code */

  /*
   * Process command line switches or set help flag if no command
   * arguments are given.
   */
  if (argc == 1) flags |= FLAG_HELP;
  else while (!err && --argc && (**++argv == ARG_SWITCH))
    err = switch_proc(*argv)
  ;

  /*
   * Display banner.
   */
  if (flags & FLAG_VERBOSE) puts(banner);

  /*
   * Perform requested action or show help screen.
   */
  if (flags & FLAG_HELP) show_help();
  else if (!err)
  {
    switch (argc)
    {
      case 1:
        if (flags & FLAG_VERBOSE)
          printf("Reading status of the interface %s ...\n", argv[0])
        ;
        err = cmd_status(argv[0]);
        break;

      case 2:
        if (flags & FLAG_VERBOSE)
          printf("Configuring interface %s using configuration file %s ...\n",
                 argv[0], argv[1])
        ;
        err = cmd_config(argv[0], argv[1]);
        break;

      default:
        err = ERR_SYNTAX;
    }
  }
  if (err) show_error(err);	/* display error message */
  return err;
} /* main() */

/****** Command Handlers ****************************************************/

/*============================================================================
 * CMD_CONFIG Handler
 * ---------------------------------------------------------------------------
 */
int  cmd_config (char *ifname, char *filename)
{
  int	err = 0;		/* error code */
  int	sock;			/* SPPP socket */

  err = read_conf(&sppp_cb.v.conf, filename);
  if (!err)
  {
    /*
     * Create a socket and initialize its IOCTL structure.
     */
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock <= 0)
    {
      err = ERR_SYSTEM;
      syserr = errno;
    }
    else
    {
      strcpy(ifr.ifr_ifrn.ifrn_name, ifname);
      ifr.ifr_ifru.ifru_data = (void *)&sppp_cb;

      /*
       * Issue command to the device.
       */
      if (ioctl(sock, SPPPIO_CONF, &ifr) < 0)
      {
        if (errno == EIO)
        {
          err = ERR_DRIVER;
          drverr = sppp_cb.result;
        }
        else
        {
          err = ERR_SYSTEM;
          syserr = errno;
        }
      }
    }
    close(sock);
  }
  return err;
} /* cmd_config() */

/*============================================================================
 * CMD_STATUS Handler
 * ---------------------------------------------------------------------------
 */
int cmd_status (char *ifname)
{
  int	err = 0;
  int	sock;		/* SPPP socket */

  /*
   * Create a socket and initialize its IOCTL structure.
   */
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock <= 0)
  {
    err = ERR_SYSTEM;
    syserr = errno;
  }
  else
  {
    strcpy(ifr.ifr_ifrn.ifrn_name, ifname);
    ifr.ifr_ifru.ifru_data = (void *)&sppp_cb;

    /*
     * Issue command to the device.
     */
    if (ioctl(sock, SPPPIO_STAT, &ifr) < 0)
    {
      if (errno == EIO)
      {
        err = ERR_DRIVER;
        drverr = sppp_cb.result;
      }
      else
      {
        err = ERR_SYSTEM;
        syserr = errno;
      }
    }
    close(sock);

    if (!err)
    {
      /*
       * Display status.
       */
      printf("Link status ................. ");
      printf("DCD %s : ", (sppp_cb.v.stat.link_status & SPPP_MODEM_DCD) ?
             "UP" : "DOWN")
      ;
      printf("CTS %s : ", (sppp_cb.v.stat.link_status & SPPP_MODEM_CTS) ?
             "UP" : "DOWN")
      ;
      printf("LINK %s\n", (sppp_cb.v.stat.link_status & SPPP_LINK_UP) ?
             "UP" : "DOWN")
      ;
      printf("Receive interrupts .......... %d\n", sppp_cb.v.stat.rx_intr_cnt);
      printf("Transmit interrupts ......... %d\n", sppp_cb.v.stat.tx_intr_cnt);
      printf("Status interrupts ........... %d\n", sppp_cb.v.stat.ctl_intr_cnt);
      printf("Spurious interrupts ......... %d\n", sppp_cb.v.stat.spurious_cnt);
      printf("Total bytes received ........ %u\n", sppp_cb.v.stat.rx_byte_cnt);
      printf("Total frames received ....... %u\n", sppp_cb.v.stat.rx_frm_cnt);
      printf("Receiver overrun errors ..... %u\n", sppp_cb.v.stat.rx_overruns);
      printf("Aborted frames received ..... %u\n", sppp_cb.v.stat.rx_aborts);
      printf("CRC errors .................. %u\n", sppp_cb.v.stat.rx_crc_err);
      printf("Discarded frames ............ %u\n", sppp_cb.v.stat.rx_dropped);
      printf("Too long frames received .... %u\n", sppp_cb.v.stat.rx_bad_len);
      printf("Total bytes transmitted ..... %u\n", sppp_cb.v.stat.tx_byte_cnt);
      printf("Total frames transmitted .... %u\n", sppp_cb.v.stat.tx_frm_cnt);
      printf("Transmit underrun errors .... %u\n", sppp_cb.v.stat.tx_underruns);
      printf("Aborted frames transmitted .. %u\n", sppp_cb.v.stat.tx_aborts);
      printf("Re-transmitted frames ....... %u\n", sppp_cb.v.stat.tx_retries);
      printf("Number of DCD changes ....... %u\n", sppp_cb.v.stat.dcd_change);
      printf("Number of CTS changes ....... %u\n", sppp_cb.v.stat.cts_change);
      printf("\n");
    }
  }
  return err;
} /* cmd_status() */

/***** Helper Functions *****************************************************/

/*============================================================================
 * Process command line switches.
 * --------------------------------------------------------------------------
 * Return 0 if processed o.k., otherwise return error code.
 */
int switch_proc (char *str)
{
  int	err = 0;

  switch (*++str)
  {
    case SW_GETHELP:
    case SW_ALTHELP:
      flags |= FLAG_HELP;
      break;

    case SW_VERBOSE:
      flags |= FLAG_VERBOSE;
      break;

    default:
      err = ERR_SWITCH;
  }
  return err;
} /* switch_proc() */

/*============================================================================
 * Show help 
 * ---------------------------------------------------------------------------
 */
void show_help (void)
{
  puts(helptext);
} /* show_help() */

/*============================================================================
 * Show error message
 * ---------------------------------------------------------------------------
 */
void show_error (int err)
{
  char	*str;

  switch (err)
  {
    case ERR_SYSTEM:
      str = strerror(syserr);
      break;

    case ERR_DRIVER:
      str = lookup(drverr, drverr_table);
      break;

    default:
      str = lookup(err, err_table);
  }
  fprintf(stderr, "%s error %u: %s!\n", progname, err, str);
} /* show_error() */

/*============================================================================
 * Read Configuration File.
 * ---------------------------------------------------------------------------
 * Return zero if o.k., otherwise return:
 *	1 -  in case of syntax error
 */
int read_conf (spppConf_t *cfg_p, char *filename)
{
  int	err = 0;
  FILE	*cfgfile;		/* configuration file handle */
  char	cfgline[MAX_CFGLINE];	/* line buffer */

  /*
   * Open configuration file.
   */
  cfgfile = fopen(filename, "r");
  if (cfgfile == NULL)
  {
    err = ERR_SYSTEM;
    syserr = errno;
  }
  else
  {
    memset(cfg_p, 0, sizeof(spppConf_t));
    while (fgets(cfgline, MAX_CFGLINE, cfgfile))
    {
      if (parse_cfg(cfgline, cfgtable, cfg_p))
      {
        err = ERR_CFG_SYNTAX;
        break;
      }
    }
    fclose(cfgfile);
  }
  return err;
}

/*============================================================================
 * Parse Configuration Line.
 * ---------------------------------------------------------------------------
 * Return zero if o.k., otherwise return:
 *	1 -  in case of syntax error
 */
int parse_cfg (char *cfgline, CfgRec_t *cfgtable, void *cfgdata)
{
  int	err	= 0;
  int	toknum;
  char	*token[MAX_TOKENS];

  toknum = tokenize(cfgline, cfgspaces, cfgeol, token);
  if (toknum > 1)
  {
    for (;
         cfgtable->keyword && strcmp(token[0], cfgtable->keyword);
         ++cfgtable)
    ;
    if (!cfgtable->keyword) err = 1;
    else switch (cfgtable->argtype)
    {
      case ARGT_SYMNAME:
        *(int *)((char *)cfgdata + cfgtable->offs) =
          name2val(token[1], cfgtable->argvals)
        ;
        break;

      case ARGT_NUMERIC:
        *(int *)((char *)cfgdata + cfgtable->offs) =
          strtoul(token[1], NULL, 10)
        ;
        break;
    }
  }
  return err;
} /* parse_cfg() */

/*============================================================================
 * Tokenize string.
 * ---------------------------------------------------------------------------
 * Return number of tokens
 */
int tokenize (char *str, char *space, char *eol, char **tokens)
{
  int	toknum	= 0;
  int	toklen;

  str[strcspn(str, eol)] = '\0';		/* terminate string with 0 */
  while (*str && (toknum < MAX_TOKENS))
  {
    while (*str && strchr(space, *str)) ++str;	/* skip leading spaces */
    toklen = strcspn(str, space);
    if (toklen)
    {
      tokens[toknum] = str;
      ++toknum;
      str += toklen;
      if (*str) *str++ = '\0';
    }
  }
  return toknum;
} /* tokenize() */

/*============================================================================
 * Get a pointer from a look-up table 
 * ---------------------------------------------------------------------------
 */
void *lookup (int code, LookUp_t *table)
{
  while (table->code && (table->code != code)) table++;
  return table->ptr;
} /* lookup() */

/*============================================================================
 * Look up a symbolic name in name table.
 * ---------------------------------------------------------------------------
 * Return a numeric value associated with that name or zero, if name was not
 * found.
 */
int name2val (char *name, LookUp_t *table)
{
  while (table->ptr && strcmp(name, table->ptr)) table++;
  return table->code;
} /* name2val() */

/****** End *****************************************************************/

